(****** Soundness of CAP ******)

Require Arith.
Require Export cap.

(****** Operational Semantics ******)

Inductive Step : prog -> prog -> Prop :=
  (* pure rules *)
 | ev_add : (ch:codeheap; H:heap; R:rfile; rd,rs,rt:reg; b:iseq)
            (Step (ch, ((H, R), (seq (add rd rs rt) b)))
	      (ch, ((H, 
		    (updateR R rd (plus (lookupR R rs) (lookupR R rt)))), b)))
 | ev_addi : (ch:codeheap; H:heap; R:rfile; rd,rs:reg; i:int; b:iseq)
            (Step (ch, ((H, R), (seq (addi rd rs i) b)))
	      (ch, ((H, 
		    (updateR R rd (plus i (lookupR R rs)))), b)))
 | ev_sub : (ch:codeheap; H:heap; R:rfile; rd,rs,rt:reg; b:iseq)
            (Step (ch, ((H, R), (seq (sub rd rs rt) b)))
	      (ch, ((H, 
		    (updateR R rd (minus (lookupR R rs) (lookupR R rt)))), b)))
 | ev_subi : (ch:codeheap; H:heap; R:rfile; rd,rs:reg; i:int; b:iseq)
            (Step (ch, ((H, R), (seq (subi rd rs i) b)))
	      (ch, ((H, 
		    (updateR R rd (minus (lookupR R rs) i))), b)))
 | ev_mov : (ch:codeheap; H:heap; R:rfile; rd,rs:reg; b:iseq)
            (Step (ch, ((H, R), (seq (mov rd rs) b)))
	      (ch, ((H, 
		    (updateR R rd (lookupR R rs))), b)))
 | ev_movi : (ch:codeheap; H:heap; R:rfile; rd:reg; i:int; b:iseq)
            (Step (ch, ((H, R), (seq (movi rd i) b)))
	      (ch, ((H, 
		    (updateR R rd i)), b)))
 | ev_bgtF : (ch:codeheap; H:heap; R:rfile; rs,rt:reg; f:lab; b:iseq)
           (le (lookupR R rs) (lookupR R rt))
            -> (Step
             (ch, ((H,R), (seq (bgt rs rt f) b)))
             (ch, ((H,R), b)))
 | ev_bgtT : (ch:codeheap; H:heap; R:rfile; rs,rt:reg; f:lab; b,b':iseq)
           (gt (lookupR R rs) (lookupR R rt))
            -> (lookupC ch f b')
            -> (Step
             (ch, ((H,R), (seq (bgt rs rt f) b)))
             (ch, ((H,R), b')))
 | ev_bgtiF : (ch:codeheap; H:heap; R:rfile; rs:reg; i:int; f:lab; b:iseq)
           (le (lookupR R rs) i)
            -> (Step
             (ch, ((H,R), (seq (bgti rs i f) b)))
             (ch, ((H,R), b)))
 | ev_bgtiT : (ch:codeheap; H:heap; R:rfile; rs:reg; i:int; f:lab; b,b':iseq)
           (gt (lookupR R rs) i)
            -> (lookupC ch f b')
            -> (Step
             (ch, ((H,R), (seq (bgti rs i f) b)))
             (ch, ((H,R), b')))
 | ev_jd : (ch:codeheap; s:state; f:lab; b:iseq)
           (lookupC ch f b)
           -> (Step (ch, (s, (jd f))) (ch, (s, b)))
 | ev_jmp : (ch:codeheap; s:state; r:reg; b:iseq)
            (lookupC ch (getValinR s r) b)
            -> (Step (ch, (s, (jmp r))) (ch, (s, b)))
  (* impure rules *)
 | ev_alloc: (ch:codeheap; H:heap; R:rfile; rd,rs:reg; b:iseq)
           (l:addr; nl:(nat->int))
           (gt l O)
            -> ((i:nat) (gt (lookupR R rs) i) -> (hnotindom (plus i l) H))
            -> (Step
             (ch, ((H,R), (seq (alloc rd rs) b)))
             (ch, (((hwriteN (lookupR R rs) H l nl),(updateR R rd l)), b)))
 | ev_ld : (ch:codeheap; H:heap; R:rfile; rd,rs:reg; i,n:nat; b:iseq)
           ((hread H (plus i (lookupR R rs))) = (hvsome n))
            -> (Step
             (ch, ((H,R), (seq (ld rd rs i) b)))
             (ch, ((H, (updateR R rd n)), b)))
 | ev_st : (ch:codeheap; H:heap; R:rfile; rd,rs:reg; i:nat; b:iseq)
           (hindom (plus i (lookupR R rd)) H)
            -> (Step
             (ch, ((H,R), (seq (st rd i rs) b)))
             (ch, (((hwrite H (plus i (lookupR R rd)) (lookupR R rs)), R), b)))
 | ev_free : (ch:codeheap; H:heap; R:rfile; rs:reg; b:iseq)
           (hindom (lookupR R rs) H)
            -> (Step
             (ch, ((H,R), (seq (free rs) b)))
             (ch, (((hfree H (lookupR R rs)), R), b))).

(****** Soundness ******)
Lemma CT_CH_lookup_0 : (ct0,ct:codeheapty; ch:codeheap; f:lab; a:assert)
 (WF_CH_0 ct0 ct ch) -> (lookupCT ct f a) -> (EX I:iseq | (lookupC ch f I)).
Proof.
Induction ct; Intros.
Inversion H0.
Inversion H1.
Inversion_clear H0.
Rewrite H2; Exists b; Apply lookupC_eq.
Inversion_clear H0.
Cut (EX I:iseq | (lookupC ch0 f I)).
Intro EXI; Inversion_clear EXI.
Exists x; Apply lookupC_neq; Auto.
Apply H with a; Auto.
Save.

Lemma CT_CH_lookup : (ct:codeheapty; ch:codeheap; f:lab; a:assert)
 (WF_CH ct ch) -> (lookupCT ct f a) -> (EX I:iseq | (lookupC ch f I)).
Proof.
Intros.
Inversion_clear H.
Apply CT_CH_lookup_0 with ct ct a; Auto.
Save.

Lemma le_or_gt : (n,m:nat) (le n m) \/ (gt n m).
Proof.
Double Induction 1 2; Auto.
Left; Auto with arith.
Right; Auto with arith.
Intros.
Cut (le n1 n0)\/(gt n1 n0); Auto.
Intro OR; Inversion_clear OR; [Left | Right]; Auto with arith.
Save.

(* As assumed in the paper: we have infinite heap so that alloc never fails *)
Axiom InfiniteHeap : (H:heap; siz:nat)
 (EX l:addr | (gt l O)
  /\ ((i:nat) (gt siz i) -> (hnotindom (plus i l) H))).

Lemma Progress_alloc : (ch:codeheap; s:state; rd,rs:reg; I:iseq)
 (EX P:prog | (Step (ch, (s, (seq (alloc rd rs) I))) P)).
Proof.
Intros.
Induction s.
Cut (EX l:addr | (gt l O)
  /\ ((i:nat) (gt (lookupR b rs) i) -> (hnotindom (plus i l) a)));
 [Intro | Apply InfiniteHeap].
Inversion_clear H; SplitAllHyps.
Exists (ch, (((hwriteN (lookupR b rs) a x ([i:nat] O)), (updateR b rd x)), I)).
Apply ev_alloc; Auto.
Save.

Lemma Progress : (ct:codeheapty; g:assert; ch:codeheap; s:state; I:iseq)
 (InferP ct g (ch, (s, I))) -> (EX P:prog | (Step (ch, (s, I)) P)).
Proof.
Intros.
NewInduction H; Auto.
NewInduction b.
(* seq *)
Clear IHb.
NewInduction s0.
NewInduction c.
Exists (ch0, ((a0, (updateR b0 r (plus (lookupR b0 r0) (lookupR b0 r1)))), b));
  Apply ev_add.
Exists (ch0, ((a0, (updateR b0 r (plus n (lookupR b0 r0)))), b));
  Apply ev_addi.
Exists (ch0, ((a0,
       (updateR b0 r (minus (lookupR b0 r0) (lookupR b0 r1)))), b));
	Apply ev_sub.
Exists (ch0, ((a0, (updateR b0 r (minus (lookupR b0 r0) n))), b));
  Apply ev_subi.
Exists (ch0, ((a0, (updateR b0 r (lookupR b0 r0))), b)); Apply ev_mov.
Exists (ch0, ((a0, (updateR b0 r n)), b)); Apply ev_movi.
(* bgt *)
Inversion_clear H0.
Cut (le (lookupR b0 r) (lookupR b0 r0))
	\/ (gt (lookupR b0 r) (lookupR b0 r0)); [Intro OR | Apply le_or_gt].
Inversion_clear OR.
Exists (ch0, ((a0, b0), b)); Apply ev_bgtF; Auto.
Cut (EX I:iseq | (lookupC ch0 n I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with ct a2; Auto].
Exists (ch0, ((a0,b0), x)); Apply ev_bgtT; Auto.
(* bgti *)
Inversion_clear H0.
Cut (le (lookupR b0 r) n) \/ (gt (lookupR b0 r) n);
 [Intro OR | Apply le_or_gt].
Inversion_clear OR.
Exists (ch0, ((a0, b0), b)); Apply ev_bgtiF; Auto.
Cut (EX I:iseq | (lookupC ch0 n0 I));
 [Intro EXI; Inversion_clear EXI | Apply CT_CH_lookup with ct a2; Auto].
Exists (ch0, ((a0,b0), x)); Apply ev_bgtiT; Auto.
(* alloc *)
Apply Progress_alloc.
(* ld *)
Inversion_clear H0.
Cut (EX n0:nat | (hread a0 (plus n (lookupR b0 r0)))=(Some nat n0)
  /\ (a' (a0,(updateR b0 r n0)))); [Clear H2; Intro H2 | Auto].
Inversion_clear H2; SplitAllHyps.
Exists (ch0, ((a0, (updateR b0 r x)), b)); Apply ev_ld; Auto.
(* st *)
Inversion_clear H0.
Exists (ch0, (((hwrite a0 (plus n (lookupR b0 r)) (lookupR b0 r0)), b0), b)).
Cut (hindom (plus n (lookupR b0 r)) a0)
  /\(a' ((hwrite a0 (plus n (lookupR b0 r)) (lookupR b0 r0)),b0)); Auto.
Intro HH; Inversion_clear HH; Apply ev_st; Auto.
(* free *)
Inversion_clear H0.
Exists (ch0, (((hfree a0 (lookupR b0 r)), b0), b)).
Cut (hindom (lookupR b0 r) a0)
  /\(a' ((hfree a0 (lookupR b0 r)),b0)); Auto.
Intro HH; Inversion_clear HH; Apply ev_free; Auto.
(* jd *)
Inversion_clear H0.
Cut (EX I:iseq | (lookupC ch0 n I));
 [Intro | Apply CT_CH_lookup with ct a1; Auto].
Inversion_clear H0.
Exists (ch0, (s0,x)); Apply ev_jd; Auto.
(* jmp *)
Inversion_clear H0.
NewInduction s0.
Cut (EXT a:assert | (lookupCT ct (lookupR b r) a) /\ (a (a0,b)));
              [Clear H2; Intro H2 | Auto].
Inversion_clear H2; SplitAllHyps.
Cut (EX I:iseq | (lookupC ch0 (lookupR b r) I));
              [Intro | Apply CT_CH_lookup with ct x; Auto].
Inversion_clear H0.
Exists (ch0,((a0,b),x0)); Apply ev_jmp; Auto.
Save.

Lemma CT_CH_Infer_0 : (ct0,ct:codeheapty; ch:codeheap)
 (f:lab; a:assert; I:iseq)
 (WF_CH_0 ct0 ct ch)
  -> (lookupCT ct f a)
  -> (lookupC ch f I)
  -> (Infer ct0 a I).
Proof.
Intros.
Induction H.
Inversion H0.
Compare f0 f; Intro.
Rewrite e in H0; Rewrite e in H1.
Inversion H0; [Idtac | Absurd f=f; Auto].
Inversion H1; [Idtac | Absurd f=f; Auto].
Rewrite <- H7; Rewrite <- H12; Assumption.
Apply HrecH.
Inversion H0; [Absurd f=f; Auto | Auto].
Inversion H1; [Absurd f=f; Auto | Auto].
Save.

Lemma CT_CH_Infer : (ct:codeheapty; ch:codeheap)
 (f:lab; a:assert; I:iseq)
 (WF_CH ct ch)
  -> (lookupCT ct f a)
  -> (lookupC ch f I)
  -> (Infer ct a I).
Proof.
Intros.
Inversion H; Apply CT_CH_Infer_0 with ct ch f; Auto.
Save.

Lemma Preservation : (ct:codeheapty; a:assert)
 (ch:codeheap; s:state; I:iseq; P:prog)
 (InferP ct a (ch, (s, I))) -> (Step (ch, (s, I)) P)
  -> (EXT a':assert | (InferP ct a' P)).
Proof.
Intros.
NewInduction H.
NewInduction b.
(* seq *)
Clear IHb.
Induction s0.
NewInduction c.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
(* bgt *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Exists a2; Apply I_prog; Auto.
Apply CT_CH_Infer with ch0 n; Auto.
(* bgti *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Exists a2; Apply I_prog; Auto.
Apply CT_CH_Infer with ch0 n0; Auto.
(* alloc *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
(* ld *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Cut (EX n0:nat |
             (hread a0 (plus n (lookupR b0 r0)))=(Some nat n0)
             /\(a' (a0,(updateR b0 r n0)))); [Clear H3; Intro H3 | Auto].
Inversion_clear H3; SplitAllHyps.
Rewrite H5 in H1.
Inversion_clear H1; Auto.
(* st *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Cut (hindom (plus n (lookupR b0 r)) a0)
              /\(a' ((hwrite a0 (plus n (lookupR b0 r)) (lookupR b0 r0)),b0));
	   [Clear H3; Intro H3 | Auto].
Inversion_clear H3; Auto.
(* free *)
Inversion_clear H1; Inversion_clear H0.
Exists a'; Apply I_prog; Auto.
Cut (hindom (lookupR b0 r) a0)/\(a' ((hfree a0 (lookupR b0 r)),b0));
	   [Clear H3; Intro H3 | Auto].
Inversion_clear H3; Auto.
(* jd *)
Inversion_clear H1; Inversion_clear H0.
Exists a1; Apply I_prog; Auto.
Apply CT_CH_Infer with ch0 n; Auto.
(* jmp *)
Inversion_clear H1; Inversion_clear H0.
NewInduction s0.
Cut (EXT a1:heap*rfile->Prop |
               (lookupCT ct (lookupR b0 r) a1)/\(a1 (a0,b0)));
		[Clear H3; Intro H3 | Auto].
Inversion_clear H3; SplitAllHyps.
Exists x; Apply I_prog; Auto.
Apply CT_CH_Infer with ch0 (getValinR (a0,b0) r); Auto.
Save.
