(****** Certified Dynamic Storage Allocation - Definitions ******)
(******                              Sept.24 - Oct.7, 2002 ******)
Require Arith.
Require Export cap.

(****** Assertions ******)
(* p points to a block of unknown elements *)
Fixpoint ptoanylist[p:addr; n:nat] : assert :=
 Cases n of
  O => emp
 | (S m) => (star (ptoany p) (ptoanylist (S p) (m)))
 end.

(* p points to a memory block with link q and size s *)
Definition MBlk[p,q:addr; s:nat] : assert :=
 [st:state]
 (gt p (2)) /\ (gt s (2)) /\ (
   (star1 (minus p (2)) (q) (
    star1 (minus p (1)) (s) (
    ptoanylist p (minus s (2))
       ))) st).

(* p points to a list of memory blocks of n elements
   with ending pointer q *)
Inductive MBlkLst : nat -> addr -> addr -> assert :=
 | mbl_O : (s:state; p:addr)
           (emp s)
	   -> (MBlkLst O p p s)
 | mbl_S : (H,H1,H2:heap; R:rfile; n,s:nat; p,p',q:addr)
           ((lt p p') \/ (p'=pnil))
	   -> (hdisj H1 H2)
	   -> (heap_eq (happend H1 H2) H)
           -> (MBlk (plus (2) p) p' s (H1,R))
           -> (MBlkLst n p' q (H2,R))
	   -> (MBlkLst (S n) p q (H,R)).

(* fl points to a list of memory blocks
   with p pointing at the ending block whose ending pointer is q *)
Definition EndL[fl,p,q:addr] : assert :=
 [st:state]
 (p=pnil -> (MBlkLst O fl q st))
  /\
 (~p=pnil -> (((lt p q) \/ q=pnil)
  /\ (EX n | (EX s | ((star (MBlkLst n fl p) (MBlk (plus (2) p) q s)) st)))
     )).

(* fl points to a list of memory blocks
   with p pointing at a middle block with ending pointer q *)
Definition MidL[fl,p,q:addr] : assert :=
 [st:state]
 (EX n | ((star (EndL fl p q) (MBlkLst n q pnil)) st)).

(* fl points to a good free list *)
Definition Good[fl:addr] : assert :=
 [st:state]
 (EX n | (MBlkLst n fl pnil st)).

(* code heap labels - free *)
Syntactic Definition L_free := (1).
Syntactic Definition L_iter := (2).
Syntactic Definition L_next := (3).
Syntactic Definition L_tryhi := (4).
Syntactic Definition L_notjhi := (6).
Syntactic Definition L_trylo := (7).
Syntactic Definition L_lnklo := (8).
Syntactic Definition L_notjlo := (9).
(* code heap labels - malloc *)
Syntactic Definition L_malloc := (10).
Syntactic Definition L_init := (11).
Syntactic Definition L_miter := (12).
Syntactic Definition L_comp := (13).
Syntactic Definition L_retptr := (14).
Syntactic Definition L_adjust := (15).
Syntactic Definition L_split := (16).
Syntactic Definition L_mnext := (17).
Syntactic Definition L_lprv := (18).
Syntactic Definition L_moremem := (19).

(* variables as shorthands for registers *)
Syntactic Definition ret0 := r0.
Syntactic Definition ret1 := r1.
Syntactic Definition flist := r2.
Syntactic Definition fptr := r3.
Syntactic Definition hp := r4.
Syntactic Definition prev := r5.
Syntactic Definition p := r6.
Syntactic Definition rnil := r7.
Syntactic Definition cursize := r8.
Syntactic Definition curend := r9.
Syntactic Definition psize := r10.
Syntactic Definition newsize := r11.
Syntactic Definition plink := r12.
Syntactic Definition prevsize := r13.
Syntactic Definition prevend := r14.
Syntactic Definition curlink := r15.

Syntactic Definition nsize := r16.
Syntactic Definition bsize := r17.
Syntactic Definition NALLOC := r18.
Syntactic Definition ebsize := r19.
Syntactic Definition mptr := r20.
Syntactic Definition newp := r21.
Syntactic Definition bbsize := r22.

(* short hands for "fields" *)
Syntactic Definition link := (0).
Syntactic Definition size := (1).

(* code heap types - free *)
Definition T_free[P:assert; f:lab] : assert :=
 [s:state]
 ((EX lnk:addr | (EX siz:addr |
       (star
	 (star P
	   (MBlk (getValinR s fptr) lnk siz))
	 (Good (getValinR s flist))
	  s)))
	  /\ ((getValinR s ret0) = f)).
Definition T_iter[P:assert; f:lab] : assert :=
 [s:state]
 (((EX lnk:addr | (EX siz:addr |
	 (star
	   (star P 
	     (MBlk (plus (2) (getValinR s hp)) lnk siz))
	   (MidL (getValinR s flist) (getValinR s prev) (getValinR s p))
	    s)))
	    /\ ((getValinR s ret0) = f))
	    /\ ((lt (getValinR s prev) (getValinR s hp))
	    \/ ((getValinR s prev)=pnil))).
Definition T_next[P:assert; f:lab] : assert :=
 [s:state]
 (((EX lnk:addr | (EX siz:addr |
	 (star
	   (star P
	     (MBlk (plus (2) (getValinR s hp)) lnk siz))
	   (MidL (getValinR s flist) (getValinR s prev) (getValinR s p))
	    s)))
	    /\ ((getValinR s ret0) = f))
	    /\ ((lt (getValinR s prev) (getValinR s hp))
	    \/ ((getValinR s prev)=pnil)))
	    /\ (~(getValinR s p)=pnil).
Definition T_tryhi[P:assert; f:lab] : assert :=
 [s:state]
 ((((EX lnk:addr | (EX siz:addr |
	   (star
	     (star P
	       (MBlk (plus (2) (getValinR s hp)) lnk siz))
	     (MidL (getValinR s flist) (getValinR s prev) (getValinR s p))
	      s)))
	      /\ ((getValinR s ret0) = f))
	      /\ ((lt (getValinR s prev) (getValinR s hp))
	      \/ ((getValinR s prev)=pnil)))
	      /\ (~(getValinR s p)=pnil))
	      /\ (lt (getValinR s hp) (getValinR s p)).
Definition T_notjhi[P:assert; f:lab] : assert :=
 [s:state]
 (EX lnk:addr | (EX siz:addr |
     (((star
	   (star P
	     (MBlk (plus (2) (getValinR s hp)) lnk siz))
	   (MidL (getValinR s flist) (getValinR s prev) (getValinR s p))
	    s)
	    /\ ((getValinR s ret0) = f))
	    /\ ((lt (getValinR s prev) (getValinR s hp))
	    \/ ((getValinR s prev)=pnil)))
	    /\ (gt (getValinR s p) (plus (getValinR s hp) siz)))).
Definition T_trylo[P:assert; f:lab] : assert :=
 [s:state]
 (EX lnk:addr | (EX n:nat |
     ((star
	 (star P
	   (EndL (getValinR s flist) (getValinR s prev) lnk))
	 (MBlkLst (S n) (getValinR s hp) pnil)
	  s)
	  /\ ((getValinR s ret0) = f))
	  /\ ((lt (getValinR s prev) (getValinR s hp))
	  \/ ((getValinR s prev)=pnil)))).
Definition T_lnklo[P:assert; f:lab] : assert :=
 [s:state]
 (EX lnk:addr | (EX n:nat | (EX m:nat | (EX siz:nat |
	 ((star
	     (star P
	       (star (MBlkLst m (getValinR s flist) (getValinR s prev))
		 (MBlk (plus (2) (getValinR s prev)) lnk siz)))
	     (MBlkLst (S n) (getValinR s hp) pnil)
	      s)
	      /\ ((getValinR s ret0) = f))
	      /\ (lt (getValinR s prev) (getValinR s hp))
	      /\ (~(getValinR s prev)=pnil))))).
Definition T_notjlo[P:assert; f:lab] : assert :=
 [s:state]
 (EX lnk:addr | (EX n:nat | (EX m:nat |
       ((star
	   (star P
	     (star (MBlkLst m (getValinR s flist) (getValinR s prev))
	       (MBlk (plus (2) (getValinR s prev))
		lnk (getValinR s prevsize))))
	   (MBlkLst (S n) (getValinR s hp) pnil)
	    s)
	    /\ ((getValinR s ret0) = f))
	    /\ (lt (getValinR s prev) (getValinR s hp))
	    /\ (lt (plus (getValinR s prev) (getValinR s prevsize))
	 (getValinR s hp))))).
(* code heap types - malloc *)
Definition T_malloc[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (Good (getValinR s flist)) s) /\
 (((getValinR s nsize)=siz0) /\ (gt siz0 (0)) /\ ((getValinR s ret1)=f)).
Definition T_init[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (Good (getValinR s flist)) s) /\
 ((le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_miter[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (MidL (getValinR s flist) (getValinR s prev) (getValinR s p)) s) /\
 ((le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_comp[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (MidL (getValinR s flist) (getValinR s prev) (getValinR s p)) s) /\
 ((gt (getValinR s p) pnil)
  /\ (le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_retptr[P,P':assert; siz0:nat; f:lab] : assert :=
 [s:state] (EX lnk | (EX siz |
     (star
       (star P (Good (getValinR s flist)))
       (MBlk (plus (2) (getValinR s p)) lnk siz)
	s) /\ (le (plus (2) siz0) siz)))
  /\ ((getValinR s ret1)=f).
Definition T_adjust[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (Good (getValinR s flist)) s) /\
 ((gt (getValinR s NALLOC) (getValinR s bsize))
  /\ (le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_split[P:assert; siz0:nat; f:lab] : assert :=
 [s:state] (EX q | (EX n |
       (star P
	 (star
	   (EndL (getValinR s flist) (getValinR s prev) (getValinR s p))
	   (star (MBlk (plus (2) (getValinR s p)) q (getValinR s psize))
	     (MBlkLst n q pnil)))
	      s) /\ (lt (getValinR s bsize) (minus (getValinR s psize) (2)))
	         /\ ((lt (getValinR s p) q) \/ (q=pnil))))
  /\ ((le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_mnext[P:assert; siz0:nat; f:lab] : assert :=
 [s:state] (EX q | (EX siz1 | (EX n |
       (star P
	 (star
	   (EndL (getValinR s flist) (getValinR s prev) (getValinR s p))
	   (star (MBlk (plus (2) (getValinR s p)) q siz1)
	     (MBlkLst n q pnil)))
	      s) /\ ((lt (getValinR s p) q) \/ (q=pnil)))))
  /\ ((le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)).
Definition T_lprv[P:assert; siz0:nat; f:lab] : assert :=
 [s:state] (EX siz1 | (EX n |
       (star P
	 (star
	   (EndL (getValinR s flist) (getValinR s prev) (getValinR s p))
	   (star (MBlk (plus (2) (getValinR s p)) (getValinR s plink) siz1)
	     (MBlkLst n (getValinR s plink) pnil)))
	      s) /\ (le (plus (2) siz0) siz1)
              /\ ((lt (getValinR s p) (getValinR s plink))
	      \/ ((getValinR s plink)=pnil))))
  /\ ((gt (getValinR s prev) pnil)
  /\ ((getValinR s ret1)=f)).

Definition T_moremem[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 (star P (Good (getValinR s flist)) s) /\
 ((le (plus (2) siz0) (getValinR s bsize))
  /\ (gt siz0 (0))
  /\ ((getValinR s ret1)=f)
  /\ (le (getValinR s bsize) (getValinR s bbsize))).

(* Pred of "free" for "init" of "malloc" *)
Definition Pred_free_for_init[P:assert; siz0:nat; f:lab] : assert :=
 [s:state]
 ((P s) /\
   ((le (plus (2) siz0) (getValinR s bsize))
    /\ (gt siz0 (0))
    /\ ((getValinR s ret1)=f))).


(* Code Heap Type Template *)
Definition CT_dsa
 [P1:assert; f1:lab; P2,P2':assert; siz0:nat; f2:lab] : codeheapty :=
  (* types for free *)
 (consCT L_free (Aor (T_free P1 f1)
     (T_free (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_iter (Aor (T_iter P1 f1)
       (T_iter (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_next (Aor (T_next P1 f1)
	 (T_next (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_tryhi (Aor (T_tryhi P1 f1)
	     (T_tryhi (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_notjhi (Aor (T_notjhi P1 f1)
	       (T_notjhi (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_trylo (Aor (T_trylo P1 f1)
		 (T_trylo (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_lnklo (Aor (T_lnklo P1 f1)
		   (T_lnklo (Pred_free_for_init P2 siz0 f2) L_init)) (
  consCT L_notjlo (Aor (T_notjlo P1 f1)
		     (T_notjlo (Pred_free_for_init P2 siz0 f2) L_init)) (
  (* types for malloc *)
  consCT L_malloc (T_malloc P2 siz0 f2) (
  consCT L_init (T_init P2 siz0 f2) (
  consCT L_miter (T_miter P2 siz0 f2) (
  consCT L_comp (T_comp P2 siz0 f2) (
  consCT L_retptr (T_retptr P2 P2' siz0 f2) (
  consCT L_adjust (T_adjust P2 siz0 f2) (
  consCT L_split (T_split P2 siz0 f2) (
  consCT L_mnext (T_mnext P2 siz0 f2) (
  consCT L_lprv (T_lprv P2 siz0 f2) (
  consCT L_moremem (T_moremem P2 siz0 f2) (
  emptyCT))))))))))))))))))).

(* code - free *)
Definition C_free : iseq :=
 (seq (subi hp fptr (2)) (
  seq (movi prev pnil) (
  seq (mov p flist) (
  jd L_iter)))).
Definition C_iter : iseq :=
 (seq (bgti p pnil L_next) (
  seq (st hp link p) (
  jd L_trylo))).
Definition C_next : iseq :=
 (seq (bgt p hp L_tryhi) (
  seq (mov prev p) (
  seq (ld p p link) (
  jd L_iter)))).
Definition C_tryhi : iseq :=
 (seq (ld cursize hp size) (
  seq (add curend hp cursize) (
  seq (bgt p curend L_notjhi) (
  seq (ld psize p size) (
  seq (add newsize cursize psize) (
  seq (st hp size newsize) (
  seq (ld plink p link) (
  seq (st hp link plink) (
  jd L_trylo))))))))).
Definition C_notjhi : iseq :=
 (seq (st hp link p) (
    jd L_trylo)).
Definition C_trylo : iseq :=
 (seq (bgti prev pnil L_lnklo) (
  seq (mov flist hp) (
  jmp ret0))).
Definition C_lnklo : iseq :=
 (seq (ld prevsize prev size) (
  seq (add prevend prev prevsize) (
  seq (bgt hp prevend L_notjlo) (
  seq (ld cursize hp size) (
  seq (add newsize prevsize cursize) (
  seq (ld curlink hp link) (
  seq (st prev size newsize) (
  seq (st prev link curlink) (
  jmp ret0))))))))).
Definition C_notjlo : iseq :=
 (seq (st prev link hp) (
  jmp ret0)).
(* code - malloc *)
Definition C_malloc : iseq :=
 (seq (addi bsize nsize (2)) (
  jd L_init)).
Definition C_init : iseq :=
 (seq (movi prev pnil) (
  seq (mov p flist) (
  jd L_miter))).
Definition C_miter : iseq :=
 (seq (bgti p pnil L_comp) (
  seq (bgt NALLOC bsize L_adjust) (
  seq (mov bbsize bsize) (
  jd L_moremem)))).
Definition C_comp : iseq :=
 (seq (ld psize p size) (
  seq (addi ebsize bsize (2)) (
  seq (bgt psize ebsize L_split) (
  seq (bgt bsize psize L_mnext) (
  seq (ld plink p link) (
  seq (bgti prev pnil L_lprv) (
  seq (mov flist plink) (
  jd L_retptr)))))))).
Definition C_retptr : iseq :=
 (seq (addi mptr p (2)) (
  jmp ret1)).
Definition C_adjust : iseq :=
 (seq (mov bbsize NALLOC) (
  jd L_moremem)).
Definition C_split : iseq :=
 (seq (sub psize psize bsize) (
  seq (st p size psize) (
  seq (add p p psize) (
  seq (st p size bsize) (
  jd L_retptr))))).
Definition C_mnext : iseq :=
 (seq (mov prev p) (
  seq (ld p p link) (
  jd L_miter))).
Definition C_lprv : iseq :=
 (seq (st prev link plink) (
  jd L_retptr)).
Definition C_moremem : iseq :=
 (seq (alloc newp bbsize) (
  seq (st newp size bbsize) (
  seq (addi fptr newp (2)) (
  seq (movi ret0 L_init) (
  jd L_free))))).

(* Code Heap Template *)
Definition CH_dsa : codeheap :=
  (* code for free *)
 (consC L_free C_free (
  consC L_iter C_iter (
  consC L_next C_next (
  consC L_tryhi C_tryhi (
  consC L_notjhi C_notjhi (
  consC L_trylo C_trylo (
  consC L_lnklo C_lnklo (
  consC L_notjlo C_notjlo (
  (* code for malloc *)
  consC L_malloc C_malloc (
  consC L_init C_init (
  consC L_miter C_miter (
  consC L_comp C_comp (
  consC L_retptr C_retptr (
  consC L_adjust C_adjust (
  consC L_split C_split (
  consC L_mnext C_mnext (
  consC L_lprv C_lprv (
  consC L_moremem C_moremem (
  emptyC))))))))))))))))))).
