(** Proofs - linking copy with library **)
Require Arith.
Require cap.
Require cdsa.
Require derived.
Require property.
Require free.
Require property2.
Require malloc.
Require copydefs.
Require copyprop.
Require copyblks.

(* The complete copy program *)
Definition CH_Copy : codeheap :=
 (consC L_copy C_copy (
  consC L_test C_test (
  consC L_next0 C_next0 (
  consC L_next1 C_next1 (
  consC L_next2 C_next2 (
  consC L_lnkprv C_lnkprv (
  consC L_halt C_halt (
  CH_dsa)))))))).
Definition List_Copy : (list int) := (nil int).
Definition H_Copy : heap := hempty.
Definition R_Copy : rfile := emptyR.
Definition S_Copy : state := (H_Copy, R_Copy).
Definition I_Copy : iseq :=
 (seq (movi flist pnil) (
  seq (movi src pnil) (
  jd L_copy))).
Definition A_Copy : assert := emp.
Definition Prog_Copy : prog := (CH_Copy, (S_Copy, I_Copy)).

(* Certifying the initialization block based on the initial state.
 * If the initialization block (I_Copy) or the initial heap (H_Copy)
 * changes (e.g., to accomodate other source lists to be copied),
 * then only this lemma WF_MAIN needs to be changed. All the other
 * certifications are still valid.
 *)
Lemma WF_MAIN : (Infer (CT_Copy List_Copy) A_Copy I_Copy).
Proof.
Unfold A_Copy; Unfold I_Copy.
Apply I_movi with [s:state](emp s)/\(getValinR s flist)=pnil.
Intros; Unfold getValinR; Unfold updateR.
Unfold lookupR; Unfold beq_reg; Constructor; Auto.
Apply I_movi with [s:state](emp s)
  /\(getValinR s flist)=pnil/\(getValinR s src)=pnil.
Unfold getValinR; Intros; Unfold updateR.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H0.
Constructor; Auto.
Apply I_jd with (T_copy List_Copy).
CTLookupTac.
Induction s; Unfold T_copy; Unfold getValinR; Intros.
Exists List_Copy; Constructor; Auto.
Exists hempty; Exists hempty.
SplitAllHyps.
Constructor; FAuto.
Constructor.
Apply heap_eq_trans with hempty.
Apply append_emp.
Apply all_same.
Unfold emp in H0; Unfold getValinH in H0.
Unfold hread in H0; Intros.
Unfold hempty; Apply sym_eq; Auto.
Constructor.
Rewrite H2; Apply l_nil.
Unfold emp; Unfold hempty; Auto.
Rewrite H; Exists O; Apply mbl_O.
Unfold emp; Unfold hempty; Auto.
Save.

Lemma WF_START : (A_Copy S_Copy).
Proof.
Unfold A_Copy; Unfold S_Copy.
Unfold emp; Unfold H_Copy.
Unfold getValinH; Unfold hempty.
Unfold hread; Auto.
Save.

(** The following axioms are there for good reasons.
 ** In the paper we assume all the registers (variables)
 ** are caller-saved. Thus all the "Preds" only depend
 ** on the stack, not on the register file.
 ** However, due to the lack of a stack to handle
 ** activation records, we cannot yet conveniently
 ** handle that. So the registers are not saved at all
 ** when calling library routines. In steand, we use
 ** different set of registers for different programs.
 ** We are using these axioms to maintain the structure
 ** of the whole framework so that in the next step where
 ** we certify stacks and activation records, things can
 ** be easily adapted to what they are supposed to be.
 **)
Axiom Pred_free_RF_Insensitive : (l0:(list int); H:heap; R,R':rfile)
 (Pred_free l0 (H,R)) -> (Pred_free l0 (H,R')).
Axiom Pred_free_for_init_RF_Insensitive : (l0:(list int); H:heap; R,R':rfile)
 (Pred_free_for_init (Pred_malloc l0) size_of_pair L_next1 (H,R))
  -> (Pred_free_for_init (Pred_malloc l0) size_of_pair L_next1 (H,R')).
Axiom Pred_malloc_RF_Insensitive : (l0:(list int); H:heap; R,R':rfile)
 (Pred_malloc l0 (H,R)) -> (Pred_malloc l0 (H,R')).
Axiom Pred'_malloc_RF_Insensitive : (l0:(list int); H:heap; R,R':rfile)
 (Pred'_malloc l0 (H,R)) -> (Pred'_malloc l0 (H,R')).

Lemma Pred_Side : (l0:(list int); H:heap; R:rfile; v:int)
 (Pred_malloc l0 (H,R)) -> (Pred'_malloc l0 (H,(updateR R mptr v))).
Proof.
Unfold Pred'_malloc; Unfold Pred_malloc; Unfold getValinR.
Unfold updateR; Unfold lookupR; Fold lookupR; Unfold beq_reg; Intros.
Inversion_clear H0.
Constructor.
Intro; Cut (EX b:nat |
            (EX l:(list nat) |
               (EX i:nat |
                  (star (Pair (lookupR R r25) b i) (Slist l i (0))
                    (H,R))/\(cons b l)=l0)))/\(lookupR R r26)=(0); Auto.
Clear H1; Clear H2; Intros.
SplitAllHyps.
Inversion_clear H2.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Constructor; Auto.
Exists x; Exists x0; Exists x1; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply Pair_RF_Insensitive.
Intro; Apply Slist_RF_Insensitive.
Intro; Cut (EX a:nat |
            (EX la:(list nat) |
               (EX lb:(list nat) |
                  (EX b:nat |
                     (EX i:nat |
                        (star
                          (star
                            (star (Pair (lookupR R r25) a i)
                              (Slist la i (0)))
                            (Slist lb (lookupR R r26) (lookupR R r28)))
                          (Pair (lookupR R r28) b (lookupR R r25))
                          (H,R))/\(app lb (cons b (cons a la)))=l0))))); Auto.
Clear H1; Clear H2; Intros.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Inversion_clear H2.
Exists x; Exists x0; Exists x1; Exists x2; Exists x3; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Clear H1; Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply Pair_RF_Insensitive.
Intro; Apply Slist_RF_Insensitive.
Intro; Apply Slist_RF_Insensitive.
Intro; Apply Pair_RF_Insensitive.
Save.

(* next2 is good return pointer for free *)
Lemma free_ret_next2 : (l0:(list int); s:state)
 (Q_free (Pred_free l0) s) -> (T_next2 l0 s).
Proof.
Unfold Q_free; Unfold Pred_free; Unfold T_next2; Auto.
Save.

(* init is good return pointer for free *)
Lemma free_ret_init : (l0:(list int); s:state)
 (Q_free (Pred_free_for_init (Pred_malloc l0) size_of_pair L_next1) s)
  -> (TT_init l0 s).
Proof.
Unfold TT_init; Unfold Q_free; Unfold Pred_free_for_init; Unfold Pred_malloc.
Unfold T_init; Intros.
Induction s; Inversion_clear H.
Inversion_clear H0.
SplitAllHyps.
Constructor; Auto.
Exists x; Exists x0; Constructor; Auto.
Save.

(* next1 is good return pointer for malloc *)
Lemma malloc_ret_next1 : (l0:(list int); s:state)
 (Q_malloc (Pred_malloc l0) size_of_pair s) -> (T_next1 l0 s).
Proof.
Unfold Q_malloc; Unfold Pred_malloc; Unfold T_next1.
Intros.
Inversion_clear H.
Inversion_clear H0.
Exists x; Exists x0.
Inversion_clear H; Constructor; Auto.
Save.

Lemma Q_free_works_on_next2 : (l0:(list int); s:state)
 (Q_free (Pred_free l0) s)
  -> (EXT A:assert | (lookupCT (CT_Copy l0) L_next2 A)/\(A s)).
Proof.
Intros; Exists (T_next2 l0); Constructor.
CTLookupTac.
Apply free_ret_next2; Auto.
Save.

Lemma Q_free_works_on_init : (l0:(list int); s:state)
 (Q_free (Pred_free_for_init (Pred_malloc l0) size_of_pair L_next1) s)
  -> (EXT A:assert | (lookupCT (CT_Copy l0) L_init A)/\(A s)).
Proof.
Intros; Exists (TT_init l0); Constructor.
CTLookupTac.
Apply free_ret_init; Auto.
Save.

Lemma Q_malloc_works_on_next1 : (l0:(list int); s:state)
 (Q_malloc (Pred_malloc l0) size_of_pair s)
  -> (EXT A:assert | (lookupCT (CT_Copy l0) L_next1 A)/\(A s)).
Proof.
Intros; Exists (T_next1 l0); Constructor.
CTLookupTac.
Apply malloc_ret_next1; Auto.
Save.

Theorem WF_Prog_Copy : (InferP (CT_Copy List_Copy) A_Copy Prog_Copy).
Proof.
Unfold Prog_Copy; Apply I_prog.
Apply wf_c.
Unfold CT_Copy; Unfold CH_Copy.
Cut (WF_CH_0 (CT_Copy List_Copy)
     (consCT (25) (T_copy List_Copy)
       (consCT (26) (T_test List_Copy)
         (consCT (27) (T_next0 List_Copy)
           (consCT (28) (T_next1 List_Copy)
             (consCT (29) (T_next2 List_Copy)
               (consCT (30) (T_lnkprv List_Copy)
                 (consCT (31) (T_halt List_Copy)
                   (CT_dsa (Pred_free List_Copy) (29)
                     (Pred_malloc List_Copy) (Pred'_malloc List_Copy)
                     (2) (28)))))))))
     (consC (25) C_copy
       (consC (26) C_test
         (consC (27) C_next0
           (consC (28) C_next1
             (consC (29) C_next2
               (consC (30) C_lnkprv (consC (31) C_halt CH_dsa)))))))); Auto.
(* blocks of copy *)
Apply wf_c_cons.
Apply WF_copy.
Apply wf_c_cons.
Apply WF_test.
Apply wf_c_cons.
Apply WF_next0.
Apply wf_c_cons.
Apply WF_next1.
Apply wf_c_cons.
Apply WF_next2.
Apply wf_c_cons.
Apply WF_lnkprv.
Apply wf_c_cons.
Apply WF_halt.
Unfold CT_dsa; Unfold CH_dsa.
(** free library **)
(* free *)
Apply wf_c_cons.
Apply I_or.
Apply WF_free with (TT_iter List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_free with (TT_iter List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
(* iter *)
Apply wf_c_cons.
Apply I_or.
Apply WF_iter with (TT_next List_Copy) (TT_trylo List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_iter with (TT_next List_Copy) (TT_trylo List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Unfold Aor; Right; Auto].
Constructor; [CTLookupTac | Intros; Unfold Aor; Right; Auto].
(* next *)
Apply wf_c_cons.
Apply I_or.
Apply WF_next with (TT_tryhi List_Copy) (TT_iter List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_next with (TT_tryhi List_Copy) (TT_iter List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
Constructor; [CTLookupTac | Intros; Right; Auto].
(* tryhi *)
Apply wf_c_cons.
Apply I_or.
Apply WF_tryhi with (TT_notjhi List_Copy) (TT_trylo List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_tryhi with (TT_notjhi List_Copy) (TT_trylo List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
Constructor; [CTLookupTac | Intros; Right; Auto].
(* notjhi *)
Apply wf_c_cons.
Apply I_or.
Apply WF_notjhi with (TT_trylo List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_notjhi with (TT_trylo List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
(* trylo *)
Apply wf_c_cons.
Apply I_or.
Apply WF_trylo with (TT_lnklo List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_trylo with (TT_lnklo List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
(* lnklo *)
Apply wf_c_cons.
Apply I_or.
Apply WF_lnklo with (TT_notjlo List_Copy).
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Left; Auto].
Apply WF_lnklo with (TT_notjlo List_Copy).
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
(* notjlo *)
Apply wf_c_cons.
Apply I_or.
Apply WF_notjlo.
Apply Q_free_works_on_next2.
Apply Pred_free_RF_Insensitive.
Apply WF_notjlo.
Apply Q_free_works_on_init.
Apply Pred_free_for_init_RF_Insensitive.
(** malloc library **)
(* malloc *)
Apply wf_c_cons.
Apply WF_malloc with (Pred'_malloc List_Copy) (TT_init List_Copy).
Apply Pred_Side.
Apply Q_malloc_works_on_next1.
Apply Pred_malloc_RF_Insensitive.
Apply Pred'_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* init *)
Apply wf_c_cons.
Apply WF_init with (TT_miter List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* miter *)
Apply wf_c_cons.
Apply WF_miter with (TT_comp List_Copy) (TT_adjust List_Copy) (TT_moremem List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
Constructor; [CTLookupTac | Auto].
Constructor; [CTLookupTac | Auto].
(* comp *)
Apply wf_c_cons.
Apply WF_comp with (Pred'_malloc List_Copy) (TT_split List_Copy) (TT_mnext List_Copy) (TT_lprv List_Copy) (TT_retptr List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
Constructor; [CTLookupTac | Auto].
Constructor; [CTLookupTac | Auto].
Constructor; [CTLookupTac | Auto].
(* retptr *)
Apply wf_c_cons.
Apply WF_retptr.
Apply Pred_Side.
Apply Q_malloc_works_on_next1.
Apply Pred_malloc_RF_Insensitive.
Apply Pred'_malloc_RF_Insensitive.
(* adjust *)
Apply wf_c_cons.
Apply WF_adjust with (TT_moremem List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* split *)
Apply wf_c_cons.
Apply WF_split with (Pred'_malloc List_Copy) (TT_retptr List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* mnext *)
Apply wf_c_cons.
Apply WF_mnext with (TT_miter List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* lprv *)
Apply wf_c_cons.
Apply WF_lprv with (Pred'_malloc List_Copy) (TT_retptr List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Auto].
(* moremem *)
Apply wf_c_cons.
Apply WF_moremem with (TT_free List_Copy).
Apply Pred_malloc_RF_Insensitive.
Constructor; [CTLookupTac | Intros; Right; Auto].
(* empty code heap *)
Apply wf_c_emp.
(* Main Block *)
Apply WF_MAIN.
(* Start State *)
Apply WF_START.
Save.
