(****** Definitions for Copy Program ******)
Require Arith.
Require Export cap.
Require Export cdsa.

(****** Assertions ******)
(* Pair *)
Definition Pair[p:addr; x:hval; q:addr] : assert :=
[s:state] (EX lnk | (EX siz | (
    (((getValinH s p)=(hvsome x)) /\ ((getValinH s (plus (1) p))=(hvsome q)))
    /\ (MBlk p lnk siz s) /\ (le (4) siz)
))).

(* Singly Linked List *)
Inductive Slist : (list int) -> addr -> addr -> assert :=
  l_nil : (s:state; i:addr)
          (emp s)
          -> (Slist (nil int) i i s)
 | l_cons : (H,H1,H2:heap; R:rfile; p,p',q:addr; x:int; l:(list int))
	    (hdisj H1 H2)
	    -> (heap_eq (happend H1 H2) H)
	    -> (Pair p x p' (H1,R))
	    -> (Slist l p' q (H2,R))
            -> (Slist (cons x l) p q (H, R)).

(* Code heap labels - copy *)
Syntactic Definition L_copy := (25).
Syntactic Definition L_test := (26).
Syntactic Definition L_next0 := (27).
Syntactic Definition L_next1 := (28).
Syntactic Definition L_next2 := (29).
Syntactic Definition L_lnkprv := (30).
Syntactic Definition L_halt := (31).

(* Variables as shorthands for registers *)
Syntactic Definition src := r25.
Syntactic Definition tgt := r26.
Syntactic Definition temp := r27.
Syntactic Definition cprev := r28.

(* Code heap types - copy *)
Definition T_copy[l0:(list int)] : assert := [s:state] (EX l:(list int) |
   (star
     (Slist l (getValinR s src) pnil)
     (Good (getValinR s flist))
   s) /\ (l=l0)).
Definition T_test[l0:(list int)] : assert := [s:state]
 (star ([s:state]
     ((getValinR s cprev)=pnil -> (
	 (EX l | (Slist l (getValinR s src) pnil s)
	  /\ (l=l0))
	  /\((getValinR s tgt)=pnil)
       ))
	  /\
     (~(getValinR s cprev)=pnil -> (
	 (EX la | (EX lb | (EX b |
	       (star
		 (star
		   (Slist la (getValinR s src) pnil)
		   (Slist lb (getValinR s tgt) (getValinR s cprev)))
		 (Pair (getValinR s cprev) b (getValinR s src))
		  s ) /\ (app lb (cons b la))=l0
	     )))
       ))
   ) (Good (getValinR s flist)) s).
Definition T_next0[l0:(list int)] : assert := [s:state]
 (star ([s:state]
     ((getValinR s cprev)=pnil -> (
	 (EX b | (EX l | (EX i | (star (Pair (getValinR s src) b i)
	     (Slist l i pnil) s)
	  /\ ((cons b l)=l0))))
	  /\((getValinR s tgt)=pnil)
       ))
	  /\
     (~(getValinR s cprev)=pnil -> (
	 (EX a | (EX la | (EX lb | (EX b | (EX i |
	       (star
		 (star
		   (star (Pair (getValinR s src) a i)
		     (Slist la i pnil))
		   (Slist lb (getValinR s tgt) (getValinR s cprev)))
		 (Pair (getValinR s cprev) b (getValinR s src))
		  s ) /\ (app lb (cons b (cons a la)))=l0
	     )))))
       ))
   ) (Good (getValinR s flist)) s).
Definition T_next1[l0:(list int)] : assert :=[s:state] (EX lnk | (EX siz |
     (star
       (star ([s:state]
	   ((getValinR s cprev)=pnil -> (
	       (EX b | (EX l | (EX i | (star (Pair (getValinR s src) b i)
		       (Slist l i pnil) s)
			/\ ((cons b l)=l0))))
			/\((getValinR s tgt)=pnil)
	     ))
		/\
	   (~(getValinR s cprev)=pnil -> (
	       (EX a | (EX la | (EX lb | (EX b | (EX i |
			 (star
			   (star
			     (star (Pair (getValinR s src) a i)
			       (Slist la i pnil))
			     (Slist lb (getValinR s tgt) (getValinR s cprev)))
			   (Pair (getValinR s cprev) b (getValinR s src))
			    s ) /\ (app lb (cons b (cons a la)))=l0
		       )))))
	     ))
	 ) (Good (getValinR s flist)))
       (MBlk (getValinR s mptr) lnk siz)
	s) /\ (le (4) siz)
   )).
Definition T_next2[l0:(list int)] : assert := [s:state]
 (star ([s:state]
     ((getValinR s cprev)=pnil -> (
	 (EX b | (EX l | (star (Pair (getValinR s mptr) b (getValinR s src))
	       (Slist l (getValinR s src) pnil) s)
		/\ ((cons b l)=l0)))
		/\((getValinR s tgt)=pnil)
       ))
	  /\
     (~(getValinR s cprev)=pnil -> (
	 (EX a | (EX la | (EX lb | (EX b |
	       (star
		 (star
		   (star (Pair (getValinR s mptr) a (getValinR s src))
		     (Slist la (getValinR s src) pnil))
		   (Slist lb (getValinR s tgt) (getValinR s cprev)))
		 (Pair (getValinR s cprev) b (getValinR s fptr))
		  s ) /\ (app lb (cons b (cons a la)))=l0
	     ))))
       ))
   ) (Good (getValinR s flist)) s).
Definition T_lnkprv[l0:(list int)] : assert := [s:state]
 (star ([s:state]
     (~(getValinR s cprev)=pnil /\ (
	 (EX a | (EX la | (EX lb | (EX b |
	       (star
		 (star
		   (star (Pair (getValinR s mptr) a (getValinR s src))
		     (Slist la (getValinR s src) pnil))
		   (Slist lb (getValinR s tgt) (getValinR s cprev)))
		 (Pair (getValinR s cprev) b (getValinR s fptr))
		  s ) /\ (app lb (cons b (cons a la)))=l0
	     ))))
       ))
   ) (Good (getValinR s flist)) s).
Definition T_halt[l0:(list int)] : assert := [s:state] (EX l:(list int) |
   (star
     (Slist l (getValinR s tgt) pnil)
     (Good (getValinR s flist))
   s) /\ (l=l0)).

(* Shorthands for ease of memorization *)
Syntactic Definition size_of_pair := (2).
(* Pred of "free" for "copy" *)
Definition Pred_free[l0:(list int)] : assert := [s:state]
 ((getValinR s cprev)=pnil -> (
     (EX b | (EX l | (star (Pair (getValinR s mptr) b (getValinR s src))
	   (Slist l (getValinR s src) pnil) s)
	    /\ ((cons b l)=l0)))
	    /\((getValinR s tgt)=pnil)
   ))
      /\
 (~(getValinR s cprev)=pnil -> (
     (EX a | (EX la | (EX lb | (EX b |
	     (star
	       (star
		 (star (Pair (getValinR s mptr) a (getValinR s src))
		   (Slist la (getValinR s src) pnil))
		 (Slist lb (getValinR s tgt) (getValinR s cprev)))
	       (Pair (getValinR s cprev) b (getValinR s fptr))
		s ) /\ (app lb (cons b (cons a la)))=l0
	   ))))
   )).
(* Pred of "malloc" for "copy" *)
Definition Pred_malloc[l0:(list int)] : assert := [s:state]
 ((getValinR s cprev)=pnil -> (
     (EX b | (EX l | (EX i | (star (Pair (getValinR s src) b i)
	     (Slist l i pnil) s)
	      /\ ((cons b l)=l0))))
	      /\((getValinR s tgt)=pnil)
   ))
      /\
 (~(getValinR s cprev)=pnil -> (
     (EX a | (EX la | (EX lb | (EX b | (EX i |
	       (star
		 (star
		   (star (Pair (getValinR s src) a i)
		     (Slist la i pnil))
		   (Slist lb (getValinR s tgt) (getValinR s cprev)))
		 (Pair (getValinR s cprev) b (getValinR s src))
		  s ) /\ (app lb (cons b (cons a la)))=l0
	     )))))
   )).
(* Pred' of "malloc" for "copy" *)
Definition Pred'_malloc : (list int) -> assert := Pred_malloc.

Definition CT_Copy[l0:(list int)] : codeheapty :=
 (consCT L_copy (T_copy l0) (
  consCT L_test (T_test l0) (
  consCT L_next0 (T_next0 l0) (
  consCT L_next1 (T_next1 l0) (
  consCT L_next2 (T_next2 l0) (
  consCT L_lnkprv (T_lnkprv l0) (
  consCT L_halt (T_halt l0) (
  CT_dsa (Pred_free l0) L_next2
		 (Pred_malloc l0) (Pred'_malloc l0) size_of_pair L_next1
	       )))))))).

(* Code - copy *)
Definition C_copy : iseq :=
 (seq (movi tgt pnil) (
  seq (movi cprev pnil) (
  jd L_test))).
Definition C_test : iseq :=
 (seq (bgti src pnil L_next0) (
  jd L_halt)).
Definition C_next0 : iseq :=
 (seq (movi nsize (2)) (
  seq (movi ret1 L_next1) (
  jd L_malloc))).
Definition C_next1 : iseq :=
 (seq (ld temp src O) (
  seq (st mptr O temp) (
  seq (mov fptr src) (
  seq (ld src src (1)) (
  seq (st mptr (1) src) (
  seq (movi ret0 L_next2) (
  jd L_free))))))).
Definition C_next2 : iseq :=
 (seq (bgti cprev pnil L_lnkprv) (
  seq (mov tgt mptr) (
  seq (mov cprev tgt) (
  jd L_test)))).
Definition C_lnkprv : iseq :=
 (seq (st cprev (1) mptr) (
  seq (mov cprev mptr) (
  jd L_test))).
Definition C_halt : iseq :=
 (jd L_halt).

(* Shorthands for code heap types -- linked version *)
Definition Pred_free_for_init0[l0:(list int)] : assert :=
 (Pred_free_for_init (Pred_malloc l0) size_of_pair L_next1).
Definition TT_free[l0:(list int)] : assert :=
 (Aor (T_free (Pred_free l0) L_next2) (T_free (Pred_free_for_init0 l0) L_init)).
Definition TT_iter[l0:(list int)] : assert :=
 (Aor (T_iter (Pred_free l0) L_next2) (T_iter (Pred_free_for_init0 l0) L_init)).
Definition TT_next[l0:(list int)] : assert :=
 (Aor (T_next (Pred_free l0) L_next2) (T_next (Pred_free_for_init0 l0) L_init)).
Definition TT_tryhi[l0:(list int)] : assert :=
 (Aor (T_tryhi (Pred_free l0) L_next2) (T_tryhi (Pred_free_for_init0 l0) L_init)).
Definition TT_notjhi[l0:(list int)] : assert :=
 (Aor (T_notjhi (Pred_free l0) L_next2) (T_notjhi (Pred_free_for_init0 l0) L_init)).
Definition TT_trylo[l0:(list int)] : assert :=
 (Aor (T_trylo (Pred_free l0) L_next2) (T_trylo (Pred_free_for_init0 l0) L_init)).
Definition TT_lnklo[l0:(list int)] : assert :=
 (Aor (T_lnklo (Pred_free l0) L_next2) (T_lnklo (Pred_free_for_init0 l0) L_init)).
Definition TT_notjlo[l0:(list int)] : assert :=
 (Aor (T_notjlo (Pred_free l0) L_next2) (T_notjlo (Pred_free_for_init0 l0) L_init)).

Definition TT_malloc[l0:(list int)] : assert :=
 (T_malloc (Pred_malloc l0) size_of_pair L_next1).
Definition TT_init[l0:(list int)] : assert :=
 (T_init (Pred_malloc l0) size_of_pair L_next1).
Definition TT_miter[l0:(list int)] : assert :=
 (T_miter (Pred_malloc l0) size_of_pair L_next1).
Definition TT_comp[l0:(list int)] : assert :=
 (T_comp (Pred_malloc l0) size_of_pair L_next1).
Definition TT_retptr[l0:(list int)] : assert :=
 (T_retptr (Pred_malloc l0) (Pred'_malloc l0) size_of_pair L_next1).
Definition TT_adjust[l0:(list int)] : assert :=
 (T_adjust (Pred_malloc l0) size_of_pair L_next1).
Definition TT_split[l0:(list int)] : assert :=
 (T_split (Pred_malloc l0) size_of_pair L_next1).
Definition TT_mnext[l0:(list int)] : assert :=
 (T_mnext (Pred_malloc l0) size_of_pair L_next1).
Definition TT_lprv[l0:(list int)] : assert :=
 (T_lprv (Pred_malloc l0) size_of_pair L_next1).
Definition TT_moremem[l0:(list int)] : assert :=
 (T_moremem (Pred_malloc l0) size_of_pair L_next1).
