(* Properties for Copy *)
Require Arith.
Require cdsa.
Require property.
Require free.
Require property2.
Require malloc.
Require copydefs.

Tactic Definition CTLookupTac := Unfold CT_Copy; Unfold CT_dsa; Repeat
 (Apply lookupCT_eq) Orelse (Apply lookupCT_neq; [Auto | Discriminate]).

Lemma minus_le : (a,b:nat)
 (le (minus a b) a).
Proof.
Induction a.
Unfold minus; Auto.
Intros; Induction b.
Unfold minus; Fold minus; Auto.
Unfold minus; Fold minus.
Apply le_trans with n; Auto.
Save.

Lemma imply_cut : (a,b,c:Prop)
 (a -> b) -> (b -> c) -> (a -> c).
Proof.
Intros; Auto.
Save.

(* Manipulating list *)
Lemma list_prop_1 : (x1,x:(list nat); x2:nat)
 (app (app x1 (cons x2 (nil nat))) x) = (app x1 (cons x2 x)).
Proof.
Induction x1; Intros.
Unfold app; Auto.
Transitivity (app (cons a (app l (cons x2 (nil nat)))) x).
Replace (app (cons a l) (cons x2 (nil nat)))
  with (cons a (app l (cons x2 (nil nat)))); Auto.
Transitivity (cons a (app (app l (cons x2 (nil nat))) x)); Auto.
Transitivity (cons a (app l (cons x2 x))); Auto.
Replace (app (app l (cons x2 (nil nat))) x) with (app l (cons x2 x)); Auto.
Save.

(* Manipulating star *)
Lemma star_emp_rev : (s:state; a:assert)
 (a s) -> (star emp a s).
Proof.
Intros.
Induction s.
Exists hempty; Exists a0; Constructor; FAuto.
Constructor.
Apply append_emp.
Constructor; Auto.
Unfold hempty; Unfold emp; Trivial.
Save.

Lemma star_subst_diffR : (H:heap; R,R':rfile; a1,a2,a1',a2':assert)
 ((H0:heap)(a1 (H0,R)) -> (a1' (H0,R')))
  -> ((H0:heap)(a2 (H0,R)) -> (a2' (H0,R')))
  -> (star a1 a2 (H,R)) -> (star a1' a2' (H,R')).
Proof.
Unfold star; Intros.
Inversion_clear H2.
Inversion_clear H3.
SplitAllHyps.
Exists x; Exists x0; Constructor; Auto.
Save.

(* MBlk implies indom *)
Lemma MBlk_indom_0 : (H:heap; R:rfile; p,q:addr; siz:nat)
 (MBlk p q siz (H,R)) -> (hindom p H).
Proof.
Intros; Cut (gt p (2)).
Intro; Induction p.
Inversion H1.
Induction p.
Inversion H1.
Inversion H3.
Clear Hrecp; Clear Hrecp0; Cut (S (S p))=(plus (2) p); Auto.
Intro EQ; Rewrite EQ; Rewrite EQ in H0.
Apply MBlk_indom with R q siz; Auto.
Apply gt_lt.
Apply MBlk_siz_gt_2 with (H,R) (plus (2) p) q; Auto.
Apply MBlk_ptr_gt_2 with (H,R) q siz; Auto.
Save.

Lemma MBlk_indom_1 : (H:heap; R:rfile; p,q:addr; siz:nat)
 (MBlk p q siz (H,R)) -> (le (4) siz) -> (hindom (S p) H).
Proof.
Intros.
Cut (gt p (2)).
Intro; Induction p.
Inversion_clear H2.
Induction p.
Inversion H2.
Inversion H4.
Clear Hrecp; Clear Hrecp0; Cut (S (S p))=(plus (2) p); Auto.
Intro EQ; Rewrite EQ in H0.
Replace (S (S (S p))) with (plus (3) p); Auto.
Apply MBlk_indom with R q siz; Auto.
Apply MBlk_ptr_gt_2 with (H,R) q siz; Auto.
Save.

(* Star MBlk implies indom *)
Lemma Star_MBlk_indom_0 : (H:heap; R:rfile; p,q:addr; siz:nat; a:assert)
 (star (MBlk p q siz) a (H,R)) -> (hindom p H).
Proof.
Intros.
Cut (gt p (2)).
Intro; Induction p.
Inversion_clear H1.
Induction p.
Inversion H1.
Inversion H3.
Clear Hrecp; Clear Hrecp0; Cut (S (S p))=(plus (2) p); Auto.
Intro EQ; Rewrite EQ; Rewrite EQ in H0.
Apply Star_MBlk_indom with R q siz a; Auto.
Apply gt_lt.
Apply Star_MBlk_siz_gt_2 with (H,R) (plus (2) p) q a; Auto.
Apply Star_MBlk_ptr_gt_2 with (H,R) q siz a; Auto.
Save.

Lemma Star_MBlk_indom_1 : (H:heap; R:rfile; p,q:addr; siz:nat; a:assert)
 (star (MBlk p q siz) a (H,R)) -> (le (4) siz) -> (hindom (S p) H).
Proof.
Intros.
Cut (gt p (2)).
Intro; Induction p.
Inversion_clear H2.
Induction p.
Inversion H2.
Inversion H4.
Clear Hrecp; Clear Hrecp0; Cut (S (S p))=(plus (2) p); Auto.
Intro EQ; Rewrite EQ in H0.
Replace (S (S (S p))) with (plus (3) p); Auto.
Apply Star_MBlk_indom with R q siz a; Auto.
Apply Star_MBlk_ptr_gt_2 with (H,R) q siz a; Auto.
Save.

(* indom from star *)
Lemma indom_star : (H:heap; R:rfile; a1,a2:assert; p:addr)
 (star a1 a2 (H,R))
  -> ((H':heap) (a1 (H',R)) -> (hindom p H'))
  -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
SplitAllHyps.
Cut (hindom p x); Auto.
Intro; Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L; Auto.
Save.

(* Reading from star *)
Lemma Read_Star : (H:heap; R:rfile; a1,a2:assert; x:addr; y:hval)
 (star a1 a2 (H,R))
  -> ((H':heap) (a1 (H',R)) -> ((hread H' x) = (hvsome y)))
  -> (hread H x)=(hvsome y).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
SplitAllHyps.
Cut (hread x0 x) = (Some nat y).
Intro EQ; Rewrite <- EQ.
Replace (hread H x) with (hread (happend x0 x1) x).
Apply append_preserve_read_L.
Apply disj_found_notin with x0 y; Auto.
Apply h_subst_read; Auto.
Apply H1; Auto.
Save.

(* Writing to star *)
Lemma Write_Star : (H:heap; R:rfile; a1,a1',a2:assert; p:addr; q:hval)
 (star a1 a2 (H,R))
  -> ((H':heap) (a1 (H',R)) -> (hindom p H'))
  -> ((H':heap) (a1 (H',R)) -> (a1' ((hwrite H' p q),R)))
  -> (star a1' a2 ((hwrite H p q),R)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H3.
SplitAllHyps.
Exists (hwrite x p q); Exists x0.
Constructor.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_indom_notin with x; Auto.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) p q).
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Save.

(* Pair on eq heaps *)
Lemma h_subst_Pair : (H,H':heap; R:rfile; p,q:addr; x:hval)
 (heap_eq H H') -> (Pair p x q (H,R)) -> (Pair p x q (H',R)).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Exists x0; Exists x1.
Constructor; Constructor.
Rewrite <- H3; Apply sym_eq; Unfold getValinH; Apply h_subst_read; Auto.
Rewrite <- H5; Apply sym_eq; Unfold getValinH; Apply h_subst_read; Auto.
Apply h_subst_MBlk with H; Auto.
Assumption.
Save.

(* Slist on eq heaps *)
Lemma h_subst_Slist : (l:(list int); i,j:addr; H,H':heap; R:rfile)
 (heap_eq H H') -> (Slist l i j (H, R)) -> (Slist l i j (H', R)).
Proof.
Intros.
Inversion_clear H1.
Apply l_nil.
Apply h_subst_emp with H; Auto.
Apply l_cons with H3 H4 p'; Auto.
Apply heap_eq_trans with H; Auto.
Save.

(* Pair Insensitive on RFile *)
Lemma Pair_RF_Insensitive : (H:heap; R,R':rfile; p,q:addr; x:hval)
 (Pair p x q (H,R)) -> (Pair p x q (H,R')).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Exists x0; Exists x1; Constructor; Constructor; Auto.
Apply MBlk_RF_Insensitive with R; Auto.
Save.

(* Slist Insensitive on RFile *)
Lemma Slist_RF_Insensitive : (l:(list int); H:heap; R,R':rfile; i,j:addr)
 (Slist l i j (H,R)) -> (Slist l i j (H,R')).
Proof.
Induction l; Intros.
Inversion_clear H0.
Apply l_nil.
Unfold emp in H1; Unfold emp; Auto.
Inversion_clear H1.
Apply l_cons with H3 H4 p'; Auto.
Apply Pair_RF_Insensitive with R; Auto.
Apply H with R; Auto.
Save.

(* Pair pointers > 0 *)
Lemma Pair_ptr_gt_O : (s:state; p,q:addr; x:int)
 (Pair p x q s) -> (~p=O).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Intro EQO; Rewrite EQO in H.
Absurd (gt O (2)).
Intro FF; Inversion FF.
Apply MBlk_ptr_gt_2 with s x0 x1; Auto.
Save.

Lemma Star_Pair_ptr_gt_O : (s:state; p,q:addr; x:int; a:assert)
 (star (Pair p x q) a s) -> (~p=O).
Proof.
Induction s; Unfold star; Intros.
SplitAllExs.
SplitAllHyps.
Apply Pair_ptr_gt_O with (x0,b) q x; Auto.
Save.

(* Slist pointers > 0 *)
Lemma Slist_ptr_gt_O : (s:state; p,q:addr; x:int; l:(list int))
 (Slist (cons x l) p q s) -> (~p=O).
Proof.
Intros.
Inversion_clear H.
Apply Pair_ptr_gt_O with (H1,R) p' x; Auto.
Save.

Lemma Star_Slist_ptr_gt_O : (s:state; p,q:addr; x:int; l:(list int); a:assert)
 (star (Slist (cons x l) p q) a s) -> (~p=O).
Proof.
Induction s; Unfold star; Intros.
SplitAllExs.
SplitAllHyps.
Apply Slist_ptr_gt_O with (x0,b) q x l; Auto.
Save.

(* Slist nil means empty heap *)
Lemma Slist_nil_empty : (s:state; p,q:addr)
 (Slist (nil int) p q s) -> (emp s).
Proof.
Intros.
Inversion_clear H; Auto.
Save.

(* Slist O O means nil *)
Lemma Slist_O_O_nil : (s:state; x:(list int))
 (Slist x O O s) -> (x=(nil int)).
Proof.
Intros.
Inversion_clear H; Auto.
Absurd O=O; Auto.
Apply Pair_ptr_gt_O with (H1,R) p' x0; Auto.
Save.

Lemma Star_Slist_O_O_nil : (s:state; x:(list int); a:assert)
 (star (Slist x O O) a s) -> (x=(nil int)).
Proof.
Induction s; Unfold star; Intros.
SplitAllExs.
SplitAllHyps.
Apply Slist_O_O_nil with (x0,b); Auto.
Save.

(* MBlk to Pair *)
Lemma MBlk_Pair : (s:state; p,lnk:addr; siz:nat)
 (MBlk p lnk siz s) -> (le (4) siz) -> (EX x | (EX q | (Pair p x q s))).
Proof.
Induction s; Intros.
Unfold Pair.
Cut (EX lnk0:nat | (EX siz0:nat |
     (MBlk p lnk0 siz0 (a,b))/\(le (4) siz0)))
      /\
 (EX x:nat | (EX q:nat |
     ((getValinH (a,b) p)=(Some nat x)
      /\(getValinH (a,b) (plus (1) p))=(Some nat q)))).
Clear H; Intros.
Inversion_clear H.
SplitAllExs.
Exists x; Exists x0; Exists x1; Exists x2.
Constructor; Auto.
Constructor.
Exists lnk; Exists siz; Constructor; Auto.
Inversion_clear H.
SplitAllHyps.
Unfold star1 in H3.
Unfold getValinH in H3; Unfold removeS in H3.
SplitAllHyps.
Induction siz.
Inversion H.
Induction siz.
Inversion H.
Inversion H6.
Clear Hrecsiz; Clear Hrecsiz0.
Unfold minus in H5; Fold minus in H5.
Cut (minus siz O)=siz; Auto with arith.
Intro EQ; Rewrite EQ in H5.
Induction siz.
Inversion H0.
Inversion H6.
Inversion H8.
Induction siz.
Inversion H0.
Inversion H6.
Inversion H8.
Inversion H10.
Clear Hrecsiz; Clear Hrecsiz0.
Unfold ptoanylist in H5; Fold ptoanylist in H5.
Unfold star in H5.
SplitAllExs.
SplitAllHyps.
SplitAllExs.
SplitAllHyps.
Inversion_clear H6; Inversion_clear H9.
Exists x3; Exists x4.
Inversion_clear H10; Inversion_clear H6.
Cut (lt (minus p (1)) p).
Intro; Cut (lt (minus p (2)) p).
Intro.
Unfold getValinH in H9; Unfold getValinH in H10; Unfold getValinH; Constructor.
Rewrite <- H9.
Transitivity (hread (happend x x0) p).
Apply sym_eq.
Transitivity (hread (hfree (hfree a (minus p (2))) (minus p (1))) p).
Apply h_subst_read; Auto.
Transitivity (hread (hfree a (minus p (2))) p).
Apply local_free; Apply lt_neq; Auto.
Apply local_free; Apply lt_neq; Auto.
Apply append_preserve_read_L.
Apply disj_found_notin with x x3; Auto.
Rewrite <- H10.
Unfold plus; Cut (hread x0 (S p))=(hread x1 (S p)).
Intro; Transitivity (hread x0 (S p)); Auto.
Transitivity (hread (happend x x0) (S p)).
Apply sym_eq.
Transitivity (hread (hfree (hfree a (minus p (2))) (minus p (1))) (S p)).
Apply h_subst_read; Auto.
Transitivity (hread (hfree a (minus p (2))) (S p)).
Apply local_free; Apply lt_neq; Auto.
Apply local_free; Apply lt_neq; Auto.
Apply append_preserve_read_R.
Apply disj_found_notin with x0 x4.
Apply disj_sym; Auto.
Rewrite H15; Auto.
Transitivity (hread (happend x1 x2) (S p)).
Apply h_subst_read; Apply heap_eq_sym; Auto.
Apply append_preserve_read_L.
Apply disj_found_notin with x1 x4; Auto.
Induction p.
Inversion H1.
Clear Hrecp; Unfold minus; Fold minus.
Apply le_lt_trans with p; Auto.
Apply minus_le.
Induction p.
Inversion H1.
Clear Hrecp; Unfold minus; Fold minus.
Apply le_lt_trans with p; Auto.
Apply minus_le.
Save.

(* Pair to MBlk *)
Lemma Pair_MBlk : (s:state; p,q:addr; x:int)
 (Pair p x q s) -> (EX lnk | (EX siz | (MBlk p lnk siz s))).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Exists x0; Exists x1; Auto.
Save.

(* Reading from Pair *)
Lemma Read_Pair_val : (H:heap; R:rfile; p,q:addr; x:hval)
 (Pair p x q (H,R)) -> ((hread H p)=(hvsome x)).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Unfold getValinH in H2; Auto.
Save.

Lemma Read_Pair_lnk : (H:heap; R:rfile; p,q:addr; x:hval)
 (Pair p x q (H,R)) -> ((hread H (plus (1) p))=(hvsome q)).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Unfold getValinH in H4; Auto.
Save.

(* Writing to ptoanylist *)
Lemma Write_ptoanylist : (i,n:nat; H:heap; R:rfile; p:addr; y:hval)
 (ptoanylist p n (H,R)) -> (lt i n)
  -> (ptoanylist p n ((hwrite H (plus i p) y),R)).
Proof.
Double Induction 1 2; Intros.
Inversion H1.
Unfold plus; Unfold ptoanylist in H1; Fold ptoanylist in H1.
Unfold ptoanylist; Fold ptoanylist.
Apply Write_Star with (ptoany p); Auto.
Intros.
Inversion_clear H3.
Unfold pto in H4.
Inversion_clear H4; Unfold plus.
Unfold getValinH in H3; Apply indom_some with x.
Unfold hread in H3; Auto.
Intros; Exists y.
Inversion_clear H3; Unfold pto.
Unfold pto in H4; Inversion_clear H4; Constructor.
Unfold plus; Unfold getValinH.
Apply write_read.
Unfold getValinH in H5; Unfold getValinH.
Intros; Unfold plus.
Transitivity (hread H' j); Auto.
Apply local_write; Auto.
Inversion H2.
Unfold ptoanylist in H2; Fold ptoanylist in H2.
Unfold ptoanylist; Fold ptoanylist.
Apply star_sym.
Apply Write_Star with (ptoanylist (S p) n).
Apply star_sym; Auto.
Intros; Replace (plus (S n1) p) with (plus n1 (S p)).
Apply ptoanylist_indom with n R; Auto with arith.
Apply sym_eq; Apply plus_Snm_nSm.
Replace (plus (S n1) p) with (plus n1 (S p)).
Intros; Apply H0; Auto with arith.
Apply sym_eq; Apply plus_Snm_nSm.
Save.

(* Writing to MBlk *)
Lemma Write_MBlk_content : (i,siz:nat; H:heap; R:rfile; p,q:addr; y:hval)
 (MBlk p q siz (H,R)) -> (lt i (minus siz (2)))
  -> (MBlk p q siz ((hwrite H (plus i p) y),R)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Constructor; Auto.
Constructor; Auto.
Unfold star1 in H4; Unfold removeS in H4; Unfold getValinH in H4.
SplitAllHyps.
Unfold star1; Unfold removeS; Unfold getValinH.
Cut (x:nat)~(minus p (S x))=(plus i p).
Constructor.
Rewrite <- H3; Apply local_write.
Intros; Intro.
Absurd (minus p (2))=(plus i p); Auto.
Constructor.
Rewrite <- H4.
Cut ~(minus p (2))=(minus p (1)).
Transitivity (hread (hwrite H (plus i p) y) (minus p (1))).
Apply local_free; Auto.
Transitivity (hread H (minus p (1))).
Apply local_write.
Intros; Intro; Absurd (minus p (1))=(plus i p); Auto.
Apply sym_eq; Apply local_free; Auto.
Induction p.
Inversion H2.
Induction p.
Inversion H2.
Inversion H8.
Clear Hrecp; Clear Hrecp0.
Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Cut (heap_eq 
   (hwrite (hfree (hfree H(minus p (2))) (minus p (1))) (plus i p) y)
   (hfree (hfree (hwrite H (plus i p) y) (minus p (2)))
     (minus p (1)))).
Apply h_subst_ptoanylist.
Apply Write_ptoanylist; Auto.
Apply heap_eq_sym; Apply heap_eq_trans with
 (hfree (hwrite (hfree H (minus p (2))) (plus i p) y) (minus p (1))).
Apply h_subst_free.
Apply write_free_sym; Auto.
Apply write_free_sym; Auto.
Intro; Apply lt_neq; Apply lt_le_trans with p; Auto with arith.
Induction p.
Inversion H2.
Unfold minus; Fold minus; Apply le_lt_trans with p; Auto with arith.
Apply minus_le.
Save.

(* Writing to Pair *)
Lemma Write_Pair_val : (H:heap; R:rfile; p,q:addr; x,y:hval)
 (Pair p x q (H,R)) -> (Pair p y q ((hwrite H p y) ,R)).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Exists x0; Exists x1; Constructor; Constructor; Auto.
Unfold getValinH; Apply write_read.
Rewrite <- H4; Unfold getValinH; Apply local_write.
Unfold plus; Auto.
Cut (MBlk p x0 x1 ((hwrite H (plus O p) y),R)); Auto.
Apply Write_MBlk_content; Auto.
Induction x1.
Inversion H3.
Induction x1.
Inversion H3.
Inversion H5.
Clear Hrecx1; Clear Hrecx0; Induction x1.
Inversion H3.
Inversion H5.
Inversion H7.
Clear Hrecx1; Unfold minus.
Auto with arith.
Save.

Lemma Write_Pair_lnk : (H:heap; R:rfile; p,q,q':addr; x:hval)
 (Pair p x q (H,R)) -> (Pair p x q' ((hwrite H (plus (1) p) q') ,R)).
Proof.
Unfold Pair; Intros.
SplitAllExs.
SplitAllHyps.
Exists x0; Exists x1; Constructor; Constructor; Auto.
Rewrite <- H2; Unfold getValinH; Apply local_write.
Unfold plus; Auto with arith.
Unfold getValinH; Apply write_read.
Apply Write_MBlk_content; Auto.
Cut (lt (minus (3) (2)) (minus x1 (2))); Auto.
Apply lt_minus_bothside; Auto.
Save.

(* Unfold Slist *)
Lemma Unfold_Slist : (s:state; p:addr; l:(list int))
 (Slist l p pnil s) -> (~p=pnil)
  -> (EX x | (EX l' | (EX lnk |
     (star (Pair p x lnk) (Slist l' lnk pnil) s)
      /\ (cons x l')=l))).
Proof.
Intros.
Inversion H.
Absurd p=O; Auto.
Exists x; Exists l0; Exists p'; Constructor; Auto.
Exists H2; Exists H3; Constructor; Auto.
Save.

Lemma Star_Unfold_Slist : (s:state; p:addr; l:(list int); a:assert)
 (star (Slist l p pnil) a s) -> (~p=pnil)
  -> (EX x | (EX l' | (EX lnk | (star
     (star (Pair p x lnk) (Slist l' lnk pnil)) a s)
      /\ (cons x l')=l))).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
SplitAllHyps.
Cut (EX xx | (EX l' | (EX lnk |
     (star (Pair p xx lnk) (Slist l' lnk pnil) (x,b))
      /\ (cons xx l')=l))).
Clear H2; Intro.
SplitAllExs.
Exists x1; Exists x2; Exists x3.
Inversion_clear H3.
Constructor; Auto.
Exists x; Exists x0; Constructor; Auto.
Apply Unfold_Slist; Auto.
Save.

(* Compose Pair and Slist to Slist *)
Lemma Compose_Pair_Slist_to_Slist : (s:state; p,q:addr; x:int; l:(list int))
 (star (Pair p x q) (Slist l q pnil) s)
  ->(Slist (cons x l) p pnil s).
Proof.
Induction s; Unfold star; Intros.
SplitAllExs.
SplitAllHyps.
Apply l_cons with x0 x1 q; Auto.
Save.

(* Compose Slist and Pair to Slist *)
Lemma Compose_Slist_Pair_to_Slist : (l:(list int); s:state; p,p',q:addr; x:int)
 (star (Slist l p p') (Pair p' x q) s)
  ->(Slist (app l (cons x (nil int))) p q s).
Proof.
Induction l; Intros.
Induction s; Inversion_clear H.
Inversion_clear H0; SplitAllHyps.
Inversion_clear H1.
Unfold app; Fold app.
Apply l_cons with x1 x0 q; Auto.
Apply disj_sym; Auto.
Apply heap_eq_trans with (happend x0 x1); Auto.
Apply append_sym; Auto.
Apply disj_sym; Auto.
Apply l_nil; Auto.
Replace (app (cons a l0) (cons x (nil nat)))
  with (cons a (app l0 (cons x (nil nat)))); Auto.
Induction s; Inversion_clear H0.
Inversion_clear H1.
SplitAllHyps.
Inversion_clear H2.
Apply l_cons with H5 (happend H6 x1) p'0.
Apply append_preserve_disj; Auto.
Apply disj_sym; Apply disj_append_L with H6.
Apply disj_sym; Apply h_subst_disj with x0; Auto.
Apply heap_eq_sym; Auto.
Apply heap_eq_trans with (happend x0 x1); Auto.
Apply heap_eq_trans with (happend (happend H5 H6) x1).
Apply heap_eq_sym; Apply append_assoc_L.
Apply h_subst_append; Auto.
Auto.
Apply H with p'.
Exists H6; Exists x1; Constructor.
Apply disj_sym; Apply disj_append_R with H5.
Apply disj_sym; Apply h_subst_disj with x0; Auto.
Apply heap_eq_sym; Auto.
Constructor; Auto.
Apply heap_eq_refl.
Save.
