(****** Derived Rules for CAP ******)
Require Arith.
Require cap.

Definition liftb := [b:bool] if b then True else False.
Definition isEmptyCT := [ct:codeheapty] Cases ct of emptyCT => true | _ 
=> false end.

(* Discrimination *)
Lemma trick : (l:lab; a:assert; ct:codeheapty)
  ~(emptyCT == (consCT l a ct)).
Proof.
Intros l a ct H.
Discriminate H.
Save.

(* Looking up CT yields unique results *)
Lemma CT_unique : (a,a':assert; ct:codeheapty; f:lab)
 (lookupCT ct f a) -> (lookupCT ct f a')
  -> (a==a').
Proof.
Induction ct.
Intros.
Inversion_clear H.
Intros.
Inversion H0.
Inversion H1.
Transitivity P; Auto.
Absurd n=f; Auto.
Inversion H1.
Absurd n=f; Auto.
Apply H with f; Auto.
Save.

Lemma lem_aor : (s,s':state; a1,a2,a1',a2':assert)
 ((a1 s) -> (a1' s')) -> ((a2 s) -> (a2' s'))
  -> (Aor a1 a2 s) -> (Aor a1' a2' s').
Proof.
Intros.
Inversion_clear H1; [Left | Right]; Auto.
Save.

Lemma lem_aor_1 : (s,s':state; a1,a2,a1',a2':assert; p:Prop)
 ((a1 s) -> (p /\ (a1' s'))) -> ((a2 s) -> (p /\ (a2' s')))
  -> (Aor a1 a2 s) -> (p /\ (Aor a1' a2' s')).
Proof.
Intros.
Inversion_clear H1.
Cut p/\(a1' s'); Auto.
Intro HH; Inversion_clear HH; Constructor; [Auto | Left; Auto].
Cut p/\(a2' s'); Auto.
Intro HH; Inversion_clear HH; Constructor; [Auto | Right; Auto].
Save.

Tactic Definition InversInfSeq :=
 Repeat Match Context With [H:(Infer ? ? (seq ? ?)) |- ?] -> Inversion_clear H.

(* Derived Inference Rule for OR *)
Lemma I_or : (b:iseq; ct:codeheapty; a1,a2:assert)
 (Infer ct a1 b) -> (Infer ct a2 b) ->
 (Infer ct (Aor a1 a2) b).
Proof.
Induction b.
Induction c; Intros.
InversInfSeq; Apply I_add with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
InversInfSeq; Apply I_addi with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
InversInfSeq; Apply I_sub with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
InversInfSeq; Apply I_subi with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
InversInfSeq; Apply I_mov with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
InversInfSeq; Apply I_movi with (Aor a'0 a'); Auto.
Intros until R; Apply lem_aor; Auto.
(* bgt *)
InversInfSeq.
Cut (a3==a0).
Intro; Apply I_bgt with (Aor a'0 a') a3; Auto.
Intros until R; Intro; Apply lem_aor; Auto.
Intros.
Inversion_clear H11; Auto.
Rewrite H0; Auto.
Apply CT_unique with ct n; Auto.
(* bgti *)
InversInfSeq.
Cut (a3==a0).
Intro; Apply I_bgti with (Aor a'0 a') a3; Auto.
Intros until R; Intro; Apply lem_aor; Auto.
Intros.
Inversion_clear H11; Auto.
Rewrite H0; Auto.
Apply CT_unique with ct n0; Auto.
(* impure rules *)
(* alloc *)
InversInfSeq; Apply I_alloc with (Aor a'0 a'); Auto.
Intros; Apply lem_aor with (H0,R) a1 a2; Auto.
(* ld *)
InversInfSeq; Apply I_ld with (Aor a'0 a'); Auto.
Intros.
Inversion_clear H5.
Cut (EX n1:nat |
             (hread H0 (plus n (lookupR R r0)))=(Some nat n1)
             /\(a'0 (H0,(updateR R r n1)))); Auto.
Clear H2; Clear H1; Intro.
Inversion_clear H1; Exists x.
Inversion_clear H2; Constructor; [Auto | Left; Auto].
Cut (EX n1:nat |
             (hread H0 (plus n (lookupR R r0)))=(Some nat n1)
             /\(a' (H0,(updateR R r n1)))); Auto.
Clear H2; Clear H1; Intro.
Inversion_clear H1; Exists x.
Inversion_clear H2; Constructor; [Auto | Right; Auto].
(* st *)
InversInfSeq; Apply I_st with (Aor a'0 a'); Auto.
Intros; Apply lem_aor_1 with (H0,R) a1 a2; Auto.
(* free *)
InversInfSeq; Apply I_free with (Aor a'0 a'); Auto.
Intros; Apply lem_aor_1 with (H0,R) a1 a2; Auto.
(* jd *)
Intros.
Inversion_clear H; Inversion_clear H0.
Apply I_jd with a0; Auto.
Intros.
Inversion_clear H0; Auto.
Replace a0 with a3; Auto.
Apply CT_unique with ct n; Auto.
(* jmp *)
Intros.
Inversion_clear H; Inversion_clear H0.
Apply I_jmp.
Intros.
Inversion_clear H2; Auto.
Save.
