(* Proofs - malloc *)
Require Arith.
Require cdsa.
Require property.
Require free.
Require property2.

(* Postcondition - malloc;
   precondition is already defined as part of T_malloc *)
Definition Q_malloc[P:assert; siz0:nat] : assert :=
 [s:state] (EX lnk | (EX siz |
 (star
   (star P (Good (getValinR s flist)))
   (MBlk (getValinR s mptr) lnk siz)
    s) /\ (le (plus (2) siz0) siz))).

Theorem WF_malloc :
 (Pred,Pred':assert; siz0:nat; ct:codeheapty; f:lab; A_init:assert)
 ((H:heap; R:rfile; m:int) (Pred (H, R))
                        -> (Pred' (H, (updateR R mptr m))))
  ->((s:state)(Q_malloc Pred' siz0 s) ->
   (exT assert ([A:assert](lookupCT ct f A) /\ (A s))))
  -> ((H:heap; R,R':rfile) (Pred (H,R))
                        -> (Pred (H,R')))
  -> ((H:heap; R,R':rfile) (Pred' (H,R))
                        -> (Pred' (H,R')))
  -> ((lookupCT ct L_init A_init)
  /\ ((s:state)(T_init Pred siz0 f s) -> (A_init s)))
  -> (Infer ct (T_malloc Pred siz0 f) C_malloc).
Proof.
Intros.
Unfold T_malloc; Unfold C_malloc.
(* addi bsize nsize 2 *)
Apply I_addi with [s:state]
 (star Pred (Good (getValinR s flist)) s) /\
 (((getValinR s bsize)=(plus (2) siz0))
  /\ (gt siz0 (0)) /\ ((getValinR s ret1)=f)).
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H5.
Constructor.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H1.
Intro; Apply Good_RF_Insensitive.
Inversion_clear H7.
Constructor; Auto.
(* jd init *)
Inversion_clear H3.
Apply I_jd with A_init; Auto.
Intros; Apply H5.
Unfold T_init.
Inversion_clear H3; Constructor; Auto.
Inversion_clear H7; Constructor; Auto.
Rewrite H3; Auto.
Save.

Theorem WF_init : (Pred:assert; siz0:nat; ct:codeheapty; f:lab; A_miter:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
   -> ((lookupCT ct L_miter A_miter)
   /\ ((s:state)(T_miter Pred siz0 f s) -> (A_miter s)))
   -> (Infer ct (T_init Pred siz0 f) C_init).
Proof.
Intros; Unfold C_init.
(* movi prev nil *)
Apply I_movi with [s:state] (T_init Pred siz0 f s) /\ (getValinR s prev)=pnil.
Unfold T_init; Unfold updateR; Unfold getValinR.
Intros; Unfold lookupR; Fold lookupR; Unfold beq_reg.
Constructor; Auto.
Inversion_clear H2; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
(* mov p flist *)
Apply I_mov with [s:state] (T_init Pred siz0 f s)
/\ (getValinR s prev)=pnil /\ (getValinR s p)=(getValinR s flist).
Unfold T_init; Unfold updateR; Unfold getValinR.
Intros; Unfold lookupR; Fold lookupR; Unfold beq_reg.
Constructor; Auto.
Inversion_clear H2.
Inversion_clear H3; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
Inversion_clear H2; Constructor; Auto.
(* jd miter *)
Inversion_clear H0.
Apply I_jd with A_miter; Auto.
Unfold T_init; Intros.
Apply H2.
Unfold T_miter.
Inversion_clear H0.
Inversion_clear H3.
Constructor; Auto.
Apply star_subst with Pred (Good (getValinR s r2)); Auto.
Intros; Inversion_clear H4.
Rewrite H7; Inversion_clear H3.
Exists x.
Apply star_subst with emp (MBlkLst x (getValinR s r2) (0)); Auto.
Apply emp_star; Auto.
Intros; Unfold EndL.
Constructor.
Intro; Apply mbl_O; Auto.
Intro; Absurd ((getValinR s r5)=O); Auto.
Save.

Theorem WF_miter : (Pred:assert; siz0:nat; ct:codeheapty; f:lab)
 (A_comp,A_adjust,A_moremem:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
   -> ((lookupCT ct L_comp A_comp)
   /\ ((s:state)(T_comp Pred siz0 f s) -> (A_comp s)))
   -> ((lookupCT ct L_adjust A_adjust)
   /\ ((s:state)(T_adjust Pred siz0 f s) -> (A_adjust s)))
   -> ((lookupCT ct L_moremem A_moremem)
   /\ ((s:state)(T_moremem Pred siz0 f s) -> (A_moremem s)))
   -> (Infer ct (T_miter Pred siz0 f) C_miter).
Proof.
Intros; Unfold T_miter; Unfold C_miter.
(* bgti p nil comp *)
Apply I_bgti with
 [s:(heap*rfile)] (star Pred
   (Good (getValinR s r2)) s)
    /\(le (plus (2) siz0) (getValinR s r17))
    /\(gt siz0 (0))/\(getValinR s r1)=f
 A_comp.
Inversion_clear H0; Auto.
Unfold getValinR; Intros.
Inversion_clear H5; Constructor; Auto.
Apply star_subst with
  Pred (MidL (lookupR R r2) (lookupR R r5) (lookupR R r6)); Auto.
Intro; Apply MidL_Good.
Unfold getValinR; Intros.
Inversion_clear H0.
Apply H7.
Unfold T_comp; Unfold getValinR.
Inversion_clear H5; Constructor; Auto.
(* bgt NALLOC bsize adjust *)
Apply I_bgt with
 [s:(heap*rfile)] (star Pred
   (Good (getValinR s r2)) s)
    /\(le (plus (2) siz0) (getValinR s r17))
    /\(gt siz0 (0))/\(getValinR s r1)=f
 A_adjust.
Inversion_clear H1; Auto.
Intros; Auto.
Unfold getValinR; Intros.
Inversion_clear H1.
Apply H7.
Unfold T_adjust; Unfold getValinR.
Inversion_clear H5; Constructor; Auto.
(* mov bbsize bsize *)
Apply I_mov with [s:(heap*rfile)]
 (star Pred (Good (getValinR s r2)) s)
  /\(le (plus (2) siz0) (getValinR s r17))
  /\(gt siz0 (0))/\(getValinR s r1)=f
  /\(le (getValinR s bsize) (getValinR s bbsize)).
Unfold updateR; Unfold getValinR.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Clear H0; Clear H1; Clear H2; Intros; SplitAllHyps; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
(* jd moremem *)
Inversion_clear H2.
Apply I_jd with A_moremem; Auto.
Save.

Theorem WF_comp : (Pred,Pred':assert; siz0:nat; ct:codeheapty; f:lab)
 (A_split,A_mnext,A_lprv,A_retptr:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
  -> ((lookupCT ct L_split A_split)
   /\ ((s:state)(T_split Pred siz0 f s) -> (A_split s)))
  -> ((lookupCT ct L_mnext A_mnext)
   /\ ((s:state)(T_mnext Pred siz0 f s) -> (A_mnext s)))
  -> ((lookupCT ct L_lprv A_lprv)
   /\ ((s:state)(T_lprv Pred siz0 f s) -> (A_lprv s)))
  -> ((lookupCT ct L_retptr A_retptr)
   /\ ((s:state)(T_retptr Pred Pred' siz0 f s) -> (A_retptr s)))
  -> (Infer ct (T_comp Pred siz0 f) C_comp).
Proof.
Intros; Unfold T_comp; Unfold C_comp.
(* ld psize p(size) *)
Apply I_ld with
 [s:(heap*rfile)] (EX n | (EX q | (star Pred
       (star
	 (EndL (getValinR s r2) (getValinR s r5) (getValinR s r6))
	 (star
	   (MBlk (plus (2) (getValinR s r6)) q (getValinR s psize))
	   (MBlkLst n q pnil)))
	    s) /\ ((lt (getValinR s r6) q) \/ (q=pnil))))
	    /\(le (plus (2) siz0) (getValinR s r17))
	    /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg; Inversion_clear H5.
Cut (EX n0 | (EX n1 | (EX q | (star Pred
         (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
           (star (MBlk (plus (2) (lookupR R r6)) q n0)
             (MBlkLst n1 q (0)))) (H4,(consR R r10 n0)))
	      /\ ((lt (lookupR R r6) q) \/ (q=pnil))
))).
Clear H6; Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H5.
Inversion_clear H6.
Exists x; Constructor.
Apply Read_Star_MBlk_siz with (consR R r10 x) x1
(star Pred
  (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
    (MBlkLst x0 x1 (0)))).
Apply star_assoc_L.
Apply star_subst with
(star Pred (MBlk (plus (1) (plus (1) (lookupR R r6))) x1 x))
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (MBlkLst x0 x1 (0))); Auto.
Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (star (MBlk (plus (2) (lookupR R r6)) x1 x)
    (MBlkLst x0 x1 (0)))); Auto.
Intros; Apply star_assoc_L.
Apply star_subst with
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (MBlk (plus (1) (plus (1) (lookupR R r6))) x1 x))
(MBlkLst x0 x1 (0)); Auto.
Apply star_assoc_R; Auto.
Intro; Apply star_sym.
Intro; Apply star_sym.
Inversion_clear H7; Constructor; Auto.
Exists x0; Exists x1.
Constructor; Auto.
Cut (EX n0:nat | (EX n1:nat | (EX q:nat | (star Pred
        (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
          (star (MBlk (plus (2) (lookupR R r6)) q n0)
            (MBlkLst n1 q (0)))) (H4,R))
             /\((lt (lookupR R r6) q)\/q=(0))))).
Intro.
Inversion_clear H5.
Inversion_clear H8.
Inversion_clear H5.
Inversion_clear H8.
Exists x; Exists x0; Exists x1.
Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
Cut (EX n | (star Pred
    (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
      (MBlkLst (S n) (lookupR R r6) (0))) (H4,R))).
Clear H6; Clear H7; Intros.
Inversion_clear H5.
Cut (EX n0:nat | (EX n1:nat | (EX q:nat | (star 
        (star (MBlk (plus (2) (lookupR R r6)) q n0) (MBlkLst n1 q (0)))
        (star Pred (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6)))
	(H4,R)) /\((lt (lookupR R r6) q)\/q=(0))))).
Clear H6; Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H5.
Exists x0; Exists x1; Exists x2.
Inversion_clear H6; Constructor; Auto.
Apply star_assoc_L; Apply star_sym; Auto.
Cut (EX q:nat | (EX n0:nat | (star
        (star (MBlk (plus (2) (lookupR R r6)) q n0) (MBlkLst x q (0)))
        (star Pred (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6)))
        (H4,R))/\((lt (lookupR R r6) q)\/q=(0)))).
Clear H6; Intros.
Inversion_clear H5.
Inversion_clear H6.
Exists x1; Exists x; Exists x0; Auto.
Apply Star_Unfold_MBlkLst.
Apply star_sym; Apply star_assoc_R; Auto.
Cut (EX n0:nat | (star
      (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
	  (MBlkLst (S n0) (lookupR R r6) (0)))
	   Pred (H4,R))).
Clear H6; Clear H7; Intros.
Inversion_clear H5.
Exists x.
Apply star_sym; Auto.
Cut (EX n:nat | (star
        (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
          (MBlkLst n (lookupR R r6) (0))) Pred (H4,R))).
Intro.
Inversion_clear H5.
Cut (EX n0:nat |
      (star (MBlkLst (S n0) (lookupR R r6) (0))
        (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
         Pred) (H4,R))).
Clear H6; Clear H8; Intros.
Inversion_clear H5.
Exists x0.
Apply star_subst with
    (star (MBlkLst (S x0) (lookupR R r6) (0))
      (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6)))
       Pred; Auto.
Apply star_assoc_R; Auto.
Intro; Apply star_sym.
Apply Star_MBlkLst_neq_pnil_S with x.
Apply star_assoc_L.
Apply star_subst with
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
           (MBlkLst x (lookupR R r6) (0))) Pred; Auto.
Intro; Apply star_sym.
Inversion_clear H7; Intro.
Rewrite H7 in H5; Inversion H5.
Apply Star_Unfold_MidL.
Apply star_sym; Auto.
(* addi ebsize bsize 2 *)
Apply I_addi with [s:(heap*rfile)]
      (EX n:nat |
         (EX q:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5)
                  (getValinR s r6))
                (star
                  (MBlk (plus (2) (getValinR s r6)) q (getValinR s r10))
                  (MBlkLst n q (0)))) s)
            /\((lt (getValinR s r6) q)\/q=(0))))
		   /\((getValinR s ebsize)=(plus (2) (getValinR s bsize)))
		   /\(le (plus (2) siz0) (getValinR s r17))
		   /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H5.
Constructor; Auto.
Inversion_clear H6.
Inversion_clear H5.
Inversion_clear H6.
Exists x; Exists x0; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
(* bgt psize ebsize split *)
Apply I_bgt with [s:(heap*rfile)]
      (EX n:nat |
         (EX q:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5)
                  (getValinR s r6))
                (star
                  (MBlk (plus (2) (getValinR s r6)) q (getValinR s r10))
                  (MBlkLst n q (0)))) s)
            /\((lt (getValinR s r6) q)\/q=(0))))
		   /\(le (plus (2) siz0) (getValinR s r17))
		   /\(gt siz0 (0))/\(getValinR s r1)=f
       A_split.
Inversion_clear H0; Auto.
Unfold getValinR; Intros.
Inversion_clear H6.
Inversion_clear H8.
Constructor; Auto.
Unfold getValinR; Intros.
Inversion_clear H0.
Apply H8.
Unfold T_split; Unfold getValinR.
Inversion_clear H6.
Inversion_clear H9.
Constructor; Auto.
Inversion_clear H0.
Inversion_clear H9.
Inversion_clear H0.
Exists x0; Exists x.
Constructor; Auto.
Constructor; Auto.
Rewrite H6 in H5.
Apply lt_O_minus_lt.
Replace (minus (minus (lookupR R r10) (2)) (lookupR R r17))
	with (minus (lookupR R r10) (plus (2) (lookupR R r17))).
Apply gt_lt.
Apply gt_minus_gt_O; Auto.
Apply minus_plus_eq_minus_minus.
(* bgt bsize psize mnext *)
Apply I_bgt with [s:(heap*rfile)]
      (EX n:nat |
         (EX q:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5)
                  (getValinR s r6))
                (star
                  (MBlk (plus (2) (getValinR s r6)) q (getValinR s r10))
                  (MBlkLst n q (0)))) s)
            /\((lt (getValinR s r6) q)\/q=(0))))
		   /\(le (getValinR s bsize) (getValinR s psize))
		   /\(le (plus (2) siz0) (getValinR s r17))
		   /\(gt siz0 (0))/\(getValinR s r1)=f
       A_mnext.
Inversion_clear H1; Auto.
Unfold getValinR; Intros.
Inversion_clear H6.
Constructor; Auto.
Unfold getValinR; Intros; Inversion_clear H1.
Apply H8.
Unfold T_mnext; Unfold getValinR.
Intros.
Inversion_clear H6.
Inversion_clear H1.
Inversion_clear H6.
Constructor; Auto.
Exists x0; Exists (lookupR R r10); Exists x; Auto.
(* ld plink p(link) *)
Apply I_ld with [s:(heap*rfile)]
       (EX n:nat |
         (star Pred
           (star
             (EndL (getValinR s r2) (getValinR s r5)
               (getValinR s r6))
             (star
               (MBlk (plus (2) (getValinR s r6))
		 (getValinR s plink) (getValinR s r10))
               (MBlkLst n (getValinR s plink) (0)))) s)
		/\((lt (getValinR s r6) (getValinR s plink))
		\/(getValinR s plink)=(0)))
		/\(le (plus (2) siz0) (getValinR s r10))
		/\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H5.
Inversion_clear H6.
Exists x0.
Constructor.
Unfold plus.
Apply Read_Star_MBlk_lnk with R (lookupR R r10)
       (star Pred
	(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
	  (MBlkLst x x0 O))).
Apply star_assoc_L.
Apply star_subst with
      (star Pred (MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10)))
      (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
	(MBlkLst x x0 (0))); Auto.
Apply star_assoc_R.
Apply star_subst with Pred
      (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
        (star (MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10))
          (MBlkLst x x0 (0)))); Auto.
Clear H5; Intros.
Apply star_assoc_L.
Apply star_subst with
      (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
	(MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10)))
      (MBlkLst x x0 (0)); Auto.
Apply star_assoc_R; Auto.
Intro; Apply star_sym.
Intro; Apply star_sym.
Constructor.
Exists x; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
Inversion_clear H7.
Inversion_clear H9.
Constructor; Auto.
Apply le_trans with (lookupR R r17); Auto.
(* bgti prev nil lprv *)
Apply I_bgti with [s:(heap*rfile)]
      (EX n:nat |
         (star Pred
           (star
             (EndL (getValinR s r2) pnil (getValinR s r6))
             (star
               (MBlk (plus (2) (getValinR s r6)) (getValinR s r12)
                 (getValinR s r10)) (MBlkLst n (getValinR s r12) (0))))
           s)
         /\((lt (getValinR s r6) (getValinR s r12))
            \/(getValinR s r12)=(0)))
        /\(le (plus (2) siz0) (getValinR s r10))
          /\(gt siz0 (0))/\(getValinR s r1)=f
       A_lprv.
Inversion_clear H2; Auto.
Unfold getValinR; Intros until R.
Intro; Replace (lookupR R r5) with O; Auto with arith.
Unfold getValinR; Intros; Inversion_clear H2.
Apply H8.
Unfold T_lprv; Unfold getValinR.
Inversion_clear H6.
Inversion_clear H2.
Inversion_clear H6.
Inversion_clear H9.
Inversion_clear H11.
Constructor.
Exists (lookupR R r10); Exists x.
Constructor; Auto.
Constructor; Auto.
(* mov flist plink *)
Apply I_mov with [s:(heap*rfile)] (EX n:nat |
    (star Pred
      (star
        (MBlk (plus (2) (getValinR s r6)) (getValinR s r12) (getValinR s r10))
	(MBlkLst n (getValinR s flist) (0)))
	 s))
	 /\(le (plus (2) siz0) (getValinR s r10))
	 /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H5.
Constructor; Auto.
Exists x.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (0) (lookupR R r6))
           (star
             (MBlk (plus (2) (lookupR R r6)) (lookupR R r12)
               (lookupR R r10)) (MBlkLst x (lookupR R r12) (0)))); Auto.
Clear H6; Intros.
Apply star_emp.
Apply star_subst with (EndL (lookupR R r2) (0) (lookupR R r6))
	 (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12)
             (lookupR R r10)) (MBlkLst x (lookupR R r12) (0))); Auto.
Intros.
Unfold EndL in H6.
Inversion_clear H6.
Clear H10; Cut (MBlkLst (0) (lookupR R r2) (lookupR R r6) s1); Auto.
Clear H9; Intro H9; Inversion_clear H9; Auto.
Intros until R0; Apply h_subst_star.
(* jd retptr *)
Inversion_clear H3.
Apply I_jd with A_retptr; Auto.
Induction s; Unfold getValinR; Intros; Apply H5.
Unfold T_retptr; Unfold getValinR.
Inversion_clear H3.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Constructor; Auto.
Exists (lookupR b r12); Exists (lookupR b r10).
Constructor; Auto.
Apply star_assoc_R.
Apply star_subst with Pred 
	 (star (MBlk (plus (2) (lookupR b r6)) (lookupR b r12)
             (lookupR b r10)) (MBlkLst x (lookupR b r2) (0))); Auto.
Clear H3; Intros.
Apply star_sym.
Apply star_subst with (MBlk (plus (2) (lookupR b r6)) (lookupR b r12)
           (lookupR b r10)) (MBlkLst x (lookupR b r2) (0)); Auto.
Clear H3; Intros.
Exists x; Auto.
Save.

Theorem WF_retptr : (Pred,Pred':assert; siz0:nat; ct:codeheapty; f:lab)
 ((H:heap; R:rfile; m:int) (Pred (H, R))
                        -> (Pred' (H, (updateR R mptr m))))
  ->((s:state)(Q_malloc Pred' siz0 s) ->
   (exT assert ([A:assert](lookupCT ct f A) /\ (A s))))
  -> ((H:heap; R,R':rfile) (Pred (H,R))
                        -> (Pred (H,R')))
  -> ((H:heap; R,R':rfile) (Pred' (H,R))
                        -> (Pred' (H,R')))
  -> (Infer ct (T_retptr Pred Pred' siz0 f) C_retptr).
Proof.
Intros; Unfold T_retptr; Unfold C_retptr.
(* addi mptr p 2 *)
Apply I_addi with [s:(heap*rfile)] (EX lnk:nat | (EX siz:nat |
            (star (star Pred' (Good (getValinR s r2)))
              (MBlk (getValinR s mptr) lnk siz) s)
            /\(le (plus (2) siz0) siz)))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H4.
Constructor; Auto.
Inversion_clear H5.
Inversion_clear H4.
Inversion_clear H5.
Exists x; Exists x0; Constructor; Auto.
Replace (consR R r20 (plus (2) (lookupR R r6)))
	       with (updateR R r20 (plus (2) (lookupR R r6))); Auto.
Apply star_assoc_R.
Apply Star_RF_Frame with Pred; Auto.
Apply star_assoc_L; Auto.
Clear H4; Intros until m; Apply RF_Insensitive_Distrib_Star.
Intro; Apply Good_RF_Insensitive.
Intro; Apply MBlk_RF_Insensitive.
(* jmp ret1 *)
Apply I_jmp.
Unfold getValinR; Intros.
Inversion_clear H4.
Rewrite H6.
Apply H0; Auto.
Save.

Theorem WF_adjust : (Pred:assert; siz0:nat; ct:codeheapty; f:lab)
 (A_moremem:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
  -> ((lookupCT ct L_moremem A_moremem)
   /\ ((s:state)(T_moremem Pred siz0 f s) -> (A_moremem s)))
  -> (Infer ct (T_adjust Pred siz0 f) C_adjust).
Proof.
Intros; Unfold T_adjust; Unfold C_adjust.
(* mov bbsize NALLOC *)
Apply I_mov with [s:(heap*rfile)]
 (star Pred (Good (getValinR s r2)) s)
  /\(le (plus (2) siz0) (getValinR s r17))
  /\(gt siz0 (0))/\(getValinR s r1)=f
  /\(le (getValinR s bsize) (getValinR s bbsize)).
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2.
Constructor.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
Clear H0; SplitAllHyps.
Repeat (Constructor; Auto).
Apply lt_le_weak.
Apply gt_lt; Auto.
(* jd moremem *)
Inversion_clear H0.
Apply I_jd with A_moremem; Auto.
Save.

Theorem WF_split : (Pred,Pred':assert; siz0:nat; ct:codeheapty; f:lab)
 (A_retptr:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
   -> ((lookupCT ct L_retptr A_retptr)
   /\ ((s:state)(T_retptr Pred Pred' siz0 f s) -> (A_retptr s)))
   -> (Infer ct (T_split Pred siz0 f) C_split).
Proof.
Intros; Unfold T_split; Unfold C_split.
(* sub psize psize bsize *)
Apply I_sub with [s:(heap*rfile)] (EX siz |
      (EX q:nat |
         (EX n:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5)
                  (getValinR s r6))
                (star
                  (MBlk (plus (2) (getValinR s r6)) q siz)
                  (MBlkLst n q (0)))) s)
		   /\((getValinR s psize)=(minus siz (getValinR s bsize)))
		   /\(lt (getValinR s r17) (minus siz (2)))
		   /\((lt (getValinR s r6) q)\/q=(0)))))
		   /\(le (plus (2) siz0) (getValinR s r17))
		   /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2; Constructor; Auto.
Exists (lookupR R r10).
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Exists x; Exists x0.
Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
(* st p(size) psize *)
Apply I_st with [s:(heap*rfile)]
      (EX siz:nat |
        (EX q:nat |
          (EX n:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5) (getValinR s r6))
                (star
		  (star 
		    (MBlk (plus (2) (getValinR s r6)) q (getValinR s psize))
		    (ptoanylist (plus (getValinR s p) (getValinR s psize))
		      (getValinR s bsize)))
                  (MBlkLst n q (0)))) s)
		   /\(getValinR s r10)=(minus siz (getValinR s r17))
                   /\(lt (getValinR s r17) (minus siz (2)))
                   /\((lt (getValinR s r6) q)\/q=(0)))))
		   /\(le (plus (2) siz0) (getValinR s r17))
		   /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold getValinR; Intros.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Cut (star (MBlk (plus (2) (lookupR R r6)) x0 x)
       (star Pred
	 (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
	   (MBlkLst x1 x0 (0))))
       (H1,R)).
Clear H3; Intros.
Constructor.
Apply Star_MBlk_indom with R x0 x
     (star Pred (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
         (MBlkLst x1 x0 (0)))); Auto.
Apply lt_trans with (2); Auto.
Apply gt_lt.
Apply Star_MBlk_siz_gt_2 with (H1,R) (plus (2) (lookupR R r6)) x0
     (star Pred (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
         (MBlkLst x1 x0 (0)))); Auto.
Constructor; Auto.
Exists x; Exists x0; Exists x1.
Constructor; Auto.
Apply star_subst with Pred
(star
  (star (MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10))
    (ptoanylist (plus (lookupR R r6) (lookupR R r10)) (lookupR R r17)))
  (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
    (MBlkLst x1 x0 (0)))); Auto.
Apply star_assoc_L.
Apply star_subst with
(star (star (MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10))
         (ptoanylist (plus (lookupR R r6) (lookupR R r10))
           (lookupR R r17))) Pred)
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
       (MBlkLst x1 x0 (0))); Auto.
Apply star_assoc_R.
Inversion_clear H5.
Rewrite H3.
Apply Write_Star_MBlk_siz_Split; Auto.
Inversion_clear H6.
Apply lt_gt.
Apply lt_minus_sym; Auto.
Intro; Apply star_sym.
Clear H2; Intros.
Apply star_sym.
Apply star_assoc_R.
Apply star_subst with
     (star (MBlk (plus (2) (lookupR R r6)) x0 (lookupR R r10))
       (ptoanylist (plus (lookupR R r6) (lookupR R r10))
         (lookupR R r17)))
     (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
       (MBlkLst x1 x0 (0))); Auto.
Intro; Apply star_sym.
Apply star_assoc_L.
Apply star_subst with
(star Pred (MBlk (plus (2) (lookupR R r6)) x0 x))
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (MBlkLst x1 x0 (0))); Auto.
Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (star (MBlk (plus (2) (lookupR R r6)) x0 x) (MBlkLst x1 x0 (0)))); Auto.
Clear H3; Intros.
Apply star_assoc_L.
Apply star_subst with
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
  (MBlk (plus (2) (lookupR R r6)) x0 x))
(MBlkLst x1 x0 (0)); Auto.
Apply star_assoc_R; Auto.
Intro; Apply star_sym.
Intro; Apply star_sym.
(* add p p psize *)
Apply I_add with [s:(heap*rfile)] (EX p' | (EX siz | (EX q | (EX n |
        (star Pred
          (star
            (EndL (getValinR s r2) (getValinR s r5) p')
            (star
              (star
                (MBlk (plus (2) p') q (getValinR s r10))
                (ptoanylist (plus p' (getValinR s r10)) (getValinR s r17)))
	      (MBlkLst n q (0)))) s)
	       /\(getValinR s p)=(plus p' (getValinR s psize))
	       /\(getValinR s r10)=(minus siz (getValinR s r17))
               /\(lt (getValinR s r17) (minus siz (2)))
               /\((lt p' q)\/q=(0))))))
	       /\(le (plus (2) siz0) (getValinR s r17))
               /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2; Constructor; Auto.
Exists (lookupR R p).
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Exists x; Exists x0; Exists x1; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply ptoanylist_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
(* st p(size) bsize *)
Apply I_st with [s:(heap*rfile)] (EX p' | (EX siz | (EX q | (EX n | (EX lnk |
          (star Pred
            (star (EndL (getValinR s r2) (getValinR s r5) p')
              (star
                (star (MBlk (plus (2) p') q (getValinR s r10))
		  (MBlk (plus (2) (getValinR s p)) lnk (getValinR s bsize)))
		(MBlkLst n q (0)))) s)
                 /\(getValinR s r10)=(minus siz (getValinR s r17))
                 /\(lt (getValinR s r17) (minus siz (2)))
                 /\((lt p' q)\/q=(0)))))))
		 /\(le (plus (2) siz0) (getValinR s r17))
		 /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold getValinR; Intros.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H5.
Rewrite <- H3 in H2.
Cut (star (ptoanylist (lookupR R r6) (lookupR R r17))
  (star Pred
    (star (EndL (lookupR R r2) (lookupR R r5) x)
     (star (MBlk (plus (2) x) x1 (lookupR R r10))
       (MBlkLst x2 x1 (0))))) (H1,R)).
Clear H2; Intros.
Constructor.
Apply Star_ptoanylist_indom with (lookupR R r17) R
(star Pred
           (star (EndL (lookupR R r2) (lookupR R r5) x)
             (star (MBlk (plus (2) x) x1 (lookupR R r10))
               (MBlkLst x2 x1 (0))))); Auto.
Inversion_clear H4.
Apply lt_le_trans with (2); Auto.
Apply le_trans with (plus (2) siz0); Auto with arith.
Constructor; Auto.
Exists x; Exists x0; Exists x1; Exists x2.
Cut (EX lnk |
(star (MBlk (plus (2) (lookupR R r6)) lnk (lookupR R r17))
  (star Pred
    (star
      (star (EndL (lookupR R r2) (lookupR R r5) x)
	(MBlk (plus (2) x) x1 (lookupR R r10)))
      (MBlkLst x2 x1 (0))))
        ((hwrite H1 (plus (1) (lookupR R r6)) (lookupR R r17)),R))
      /\(lookupR R r10)=(minus x0 (lookupR R r17))
        /\(lt (lookupR R r17) (minus x0 (2)))/\((lt x x1)\/x1=(0))).
Clear H2; Clear H6; Intros.
Inversion_clear H2.
Exists x3.
Inversion_clear H5; Constructor; Auto.
Apply star_subst with Pred
(star (MBlk (plus (2) (lookupR R r6)) x3 (lookupR R r17))
           (star
             (star (EndL (lookupR R r2) (lookupR R r5) x)
               (MBlk (plus (2) x) x1 (lookupR R r10)))
             (MBlkLst x2 x1 (0)))); Auto.
Apply star_assoc_L.
Apply star_subst with
(star (MBlk (plus (2) (lookupR R r6)) x3 (lookupR R r17)) Pred)
(star (star (EndL (lookupR R r2) (lookupR R r5) x)
    (MBlk (plus (2) x) x1 (lookupR R r10))) (MBlkLst x2 x1 (0))); Auto.
Apply star_assoc_R; Auto.
Intro; Apply star_sym.
Clear H2; Intros.
Apply star_assoc_L.
Apply star_subst with
(star (MBlk (plus (2) (lookupR R r6)) x3 (lookupR R r17))
  (star (EndL (lookupR R r2) (lookupR R r5) x)
    (MBlk (plus (2) x) x1 (lookupR R r10))))
(MBlkLst x2 x1 (0)); Auto.
Apply star_assoc_R; Auto.
Clear H2; Intros.
Apply star_assoc_L; Apply star_sym; Auto.
Cut (EX lnk:nat |
      (star (MBlk (plus (2) (lookupR R r6)) lnk (lookupR R r17))
        (star Pred
          (star
            (star (EndL (lookupR R r2) (lookupR R r5) x)
              (MBlk (plus (2) x) x1 (lookupR R r10)))
            (MBlkLst x2 x1 (0))))
        ((hwrite H1 (plus (1) (lookupR R r6)) (lookupR R r17)),R)))
      /\(lookupR R r10)=(minus x0 (lookupR R r17))
        /\(lt (lookupR R r17) (minus x0 (2)))/\((lt x x1)\/x1=(0)).
Clear H2; Intros.
Inversion_clear H2.
Inversion_clear H5.
Exists x3; Constructor; Auto.
Constructor; Auto.
Apply Star_ptoanylist_MBlk.
Apply star_subst with (ptoanylist (lookupR R r6) (lookupR R r17))
(star Pred
           (star (EndL (lookupR R r2) (lookupR R r5) x)
             (star (MBlk (plus (2) x) x1 (lookupR R r10))
               (MBlkLst x2 x1 (0))))); Auto.
Clear H2; Intros.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) x)
           (star (MBlk (plus (2) x) x1 (lookupR R r10))
             (MBlkLst x2 x1 (0)))); Auto.
Intro; Apply star_assoc_R.
Rewrite H3.
Apply le_gt_trans with x; Auto with arith.
Apply simpl_gt_plus_l with (2).
Replace (plus (2) O) with (2); Auto with arith.
Apply Star_MBlk_ptr_gt_2 with (H1,R) x1 (lookupR R r10)
(star (ptoanylist (lookupR R r6) (lookupR R r17))
  (star Pred
    (star (EndL (lookupR R r2) (lookupR R r5) x) (MBlkLst x2 x1 (0))))).
Apply star_sym.
Apply star_assoc_R.
Apply star_subst with (ptoanylist (lookupR R r6) (lookupR R r17))
(star Pred
  (star (EndL (lookupR R r2) (lookupR R r5) x)
    (star (MBlk (plus (2) x) x1 (lookupR R r10))
      (MBlkLst x2 x1 (0))))); Auto.
Clear H2; Intros.
Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) x)
  (star (MBlk (plus (2) x) x1 (lookupR R r10))
    (MBlkLst x2 x1 (0)))); Auto.
Clear H2; Intros; Apply star_assoc_R.
Apply star_subst with (EndL (lookupR R r2) (lookupR R r5) x)
         (star (MBlk (plus (2) x) x1 (lookupR R r10))
           (MBlkLst x2 x1 (0))); Auto.
Intro; Apply star_sym.
Inversion_clear H4.
Replace (2) with (plus (2) O); Auto.
Apply le_gt_trans with (plus (2) siz0); Auto with arith.
Apply gt_reg_l.
Inversion_clear H7; Auto.
Apply star_sym; Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) x)
           (star
             (star (MBlk (plus (2) x) x1 (lookupR R r10))
               (ptoanylist (lookupR R r6) (lookupR R r17)))
             (MBlkLst x2 x1 (0)))); Auto.
Clear H2; Intros.
Apply star_assoc_R.
Apply star_subst with(EndL (lookupR R r2) (lookupR R r5) x)
         (star
           (star (MBlk (plus (2) x) x1 (lookupR R r10))
             (ptoanylist (lookupR R r6) (lookupR R r17)))
           (MBlkLst x2 x1 (0))); Auto.
Clear H2; Intros.
Apply star_sym; Apply star_assoc_L.
Apply star_subst with 
(star (MBlk (plus (2) x) x1 (lookupR R r10))
  (ptoanylist (lookupR R r6) (lookupR R r17)))
(MBlkLst x2 x1 (0)); Auto.
Intro; Apply star_sym.
(* jd retptr *)
Inversion_clear H0.
Apply I_jd with A_retptr; Auto.
Intros; Apply H2.
Unfold T_retptr.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H4.
Inversion_clear H8.
Constructor; Auto.
Exists x3; Exists (getValinR s r17).
Constructor; Auto.
Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (getValinR s r2) (getValinR s r5) x)
  (star
    (star (MBlk (plus (2) x) x1 (getValinR s r10))
      (MBlk (plus (2) (getValinR s r6)) x3 (getValinR s r17)))
    (MBlkLst x2 x1 (0)))); Auto.
Clear H3; Intros.
Apply star_subst with
(star (EndL (getValinR s r2) (getValinR s r5) x)
  (star (MBlk (plus (2) x) x1 (getValinR s r10))
    (MBlkLst x2 x1 (0))))
(MBlk (plus (2) (getValinR s r6)) x3 (getValinR s r17)); Auto.
Apply star_assoc_R.
Apply star_subst with (EndL (getValinR s r2) (getValinR s r5) x)
         (star
           (star (MBlk (plus (2) x) x1 (getValinR s r10))
             (MBlk (plus (2) (getValinR s r6)) x3 (getValinR s r17)))
           (MBlkLst x2 x1 (0))); Auto.
Clear H3; Intros.
Apply star_assoc_R.
Apply star_subst with 
(MBlk (plus (2) x) x1 (getValinR s r10))
(star 
  (MBlk (plus (2) (getValinR s r6)) x3 (getValinR s r17))
  (MBlkLst x2 x1 (0))); Auto.
Apply star_assoc_L; Auto.
Clear H3; Intro; Apply star_sym.
Clear H3; Intros.
Apply MidL_Good with (getValinR s r5) x.
Exists (S x2).
Apply star_subst with (EndL (getValinR s r2) (getValinR s r5) x)
         (star (MBlk (plus (2) x) x1 (getValinR s r10))
           (MBlkLst x2 x1 (0))); Auto.
Clear H3; Intros; Apply Compose_MBlk_MBlkLst_to_MBlkLst
	    with x1 (getValinR s r10); Auto.
Save.

Theorem WF_mnext : (Pred:assert; siz0:nat; ct:codeheapty; f:lab)
 (A_miter:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
  -> ((lookupCT ct L_miter A_miter)
   /\ ((s:state)(T_miter Pred siz0 f s) -> (A_miter s)))
  -> (Infer ct (T_mnext Pred siz0 f) C_mnext).
Proof.
Intros; Unfold T_mnext; Unfold C_mnext.
(* mov prev p *)
Apply I_mov with [s:(heap*rfile)] (EX p' | (EX q | (EX siz1 | (EX n |
         (star Pred
           (star
             (EndL (getValinR s r2) p' (getValinR s r6))
             (star (MBlk (plus (2) (getValinR s p)) q siz1)
               (MBlkLst n q (0)))) s)
		/\((lt (getValinR s prev) q)\/q=(0))))))
		/\(getValinR s prev)=(getValinR s p)
		/\(le (plus (2) siz0) (getValinR s r17))
		/\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR.
Intros; Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2; Constructor; Auto.
Exists (lookupR R prev).
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Exists x; Exists x0; Exists x1; Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
(* ld p p(link) *)
Apply I_ld with [s:(heap*rfile)]
 (EX n:nat |
   (star Pred
     (star (EndL (getValinR s r2) (getValinR s r5) (getValinR s p))
       (MBlkLst n (getValinR s p) (0))) s))
	/\(le (plus (2) siz0) (getValinR s r17))
        /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Exists x0.
Constructor; Auto.
Unfold plus; Apply Read_Star_MBlk_lnk with R x1
 (star Pred
   (star (EndL (lookupR R r2) x (lookupR R r6))
     (MBlkLst x2 x0 (0)))).
Apply star_sym.
Apply star_assoc_R.
Apply star_subst with Pred
 (star (EndL (lookupR R r2) x (lookupR R r6))
   (star (MBlk (plus (2) (lookupR R r6)) x0 x1)
     (MBlkLst x2 x0 (0)))); Auto.
Clear H2; Intros.
Apply star_assoc_R.
Apply star_subst with (EndL (lookupR R r2) x (lookupR R r6))
         (star (MBlk (plus (2) (lookupR R r6)) x0 x1)
           (MBlkLst x2 x0 (0))); Auto.
Intro; Apply star_sym.
Constructor.
Exists x2.
Cut (star Pred
     (star (EndL (lookupR R r2) (lookupR R r5) x0) (MBlkLst x2 x0 (0)))
     (H1,R)).
Clear H2; Intros.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply H.
Clear H2; Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
Apply star_subst with Pred
(star (EndL (lookupR R r2) x (lookupR R r6))
           (star (MBlk (plus (2) (lookupR R r6)) x0 x1)
             (MBlkLst x2 x0 (0)))); Auto.
Clear H2; Intros.
Apply star_subst with
(star (EndL (lookupR R r2) x (lookupR R r6))
  (MBlk (plus (2) (lookupR R r6)) x0 x1))
(MBlkLst x2 x0 (0)); Auto.
Apply star_assoc_R; Auto.
Clear H2; Intros.
Induction s1; Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H6.
Inversion_clear H7.
Apply h_subst_EndL with (happend x3 x4); Auto.
Inversion_clear H4.
Rewrite H7.
Apply Compose_EndL_MBlk_to_EndL with x x1; Auto.
Rewrite <- H7; Auto.
Inversion_clear H4; Auto.
(* jd miter *)
Inversion_clear H0.
Apply I_jd with A_miter; Auto.
Intros; Apply H2.
Unfold T_miter.
Inversion_clear H0; Constructor; Auto.
Inversion_clear H3.
Apply star_subst with Pred (star
           (EndL (getValinR s r2) (getValinR s r5) (getValinR s r6))
           (MBlkLst x (getValinR s r6) (0))); Auto.
Clear H0; Intros.
Exists x; Auto.
Save.

Theorem WF_lprv : (Pred,Pred':assert; siz0:nat; ct:codeheapty; f:lab)
 (A_retptr:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
  -> ((lookupCT ct L_retptr A_retptr)
   /\ ((s:state)(T_retptr Pred Pred' siz0 f s) -> (A_retptr s)))
  -> (Infer ct (T_lprv Pred siz0 f) C_lprv).
Proof.
Intros; Unfold T_lprv; Unfold C_lprv.
(* st prev(link) plink *)
Apply I_st with  [s:(heap*rfile)]
      (EX siz1:nat |
         (EX n:nat |
            (star Pred
              (star
                (EndL (getValinR s r2) (getValinR s r5)
                  (getValinR s plink))
                (star
                  (MBlk (plus (2) (getValinR s r6)) (getValinR s r12)
                    siz1) (MBlkLst n (getValinR s r12) (0)))) s)
		   /\(le (plus (2) siz0) siz1)))
	     /\(getValinR s r1)=f.
Unfold getValinR; Intros.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Inversion_clear H3.
Cut (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
         (star Pred
           (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
             (MBlkLst x0 (lookupR R r12) (0)))) (H1,R)).
Clear H2; Intros.
Constructor.
Unfold plus.
Apply Star_EndL_indom_P with R (lookupR R r2) (lookupR R r6)
(star Pred
           (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
             (MBlkLst x0 (lookupR R r12) (0)))); Auto.
Inversion_clear H4.
Intro EQO; Rewrite EQO in H3; Inversion_clear H3.
Constructor; Auto.
Exists x; Exists x0.
Constructor; Auto.
Apply star_assoc_L.
Apply star_subst with
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r12)) Pred)
(star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
  (MBlkLst x0 (lookupR R r12) (0))); Auto.
Apply star_assoc_R.
Replace (plus (0) (lookupR R r5)) with (lookupR R r5); Auto.
Inversion_clear H5.
Apply Write_Star_EndL_lnk with (lookupR R r6); Auto.
Inversion_clear H4; Intro EQO.
Rewrite EQO in H5; Inversion H5.
Inversion_clear H4.
Inversion_clear H6; Auto.
Left.
Apply lt_trans with (lookupR R r6); Auto.
Cut ((lt (lookupR R r5) (lookupR R r6)) \/ ((lookupR R r6)=pnil)).
Intro OR; Inversion OR; Auto.
Cut (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
         (star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
           (star Pred
             (MBlkLst x0 (lookupR R r12) (0)))) (H1,R)).
Clear H2; Intro.
Rewrite H6 in H2.
Absurd (gt (2) (2)); Auto with arith.
Apply Star_MBlk_ptr_gt_2 with (H1,R) (lookupR R r12) x
(star (EndL (lookupR R r2) (lookupR R r5) (0))
           (star Pred (MBlkLst x0 (lookupR R r12) (0)))); Auto.
Apply star_sym.
Apply star_assoc_R.
Apply star_subst with (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
         (star Pred
           (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
             (MBlkLst x0 (lookupR R r12) (0)))); Auto.
Clear H2; Intros.
Apply star_assoc_R.
Apply star_subst with Pred
         (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
           (MBlkLst x0 (lookupR R r12) (0))); Auto.
Intro; Apply star_sym.
Apply proj2 with (EX n | (EX siz | (star (
	    star (MBlkLst n (lookupR R r2) (lookupR R r5))
		 (MBlk (plus (2) (lookupR R r5)) (lookupR R r6) siz))
	       (star Pred
		 (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
		   (MBlkLst x0 (lookupR R r12) (0))))
	       (H1,R)))).
Apply Star_Unfold_EndL; Auto.
Intro EQO; Rewrite EQO in H5; Inversion H5.
Intro; Apply star_sym.
Inversion_clear H5; Auto.
Inversion_clear H4; Auto.
Apply star_sym; Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR R r2) (lookupR R r5) (lookupR R r6))
           (star (MBlk (plus (2) (lookupR R r6)) (lookupR R r12) x)
             (MBlkLst x0 (lookupR R r12) (0)))); Auto.
Clear H2; Intro; Apply star_sym.
(* jd retptr *)
Inversion_clear H0.
Apply I_jd with A_retptr; Auto.
Induction s; Unfold getValinR; Intros.
Apply H2.
Unfold T_retptr; Unfold getValinR.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H3.
Constructor; Auto.
Exists (lookupR b r12); Exists x.
Constructor; Auto.
Apply star_assoc_R.
Apply star_subst with Pred
(star (EndL (lookupR b r2) (lookupR b r5) (lookupR b r12))
           (star (MBlk (plus (2) (lookupR b r6)) (lookupR b r12) x)
             (MBlkLst x0 (lookupR b r12) (0)))); Auto.
Clear H0; Intros.
Apply star_subst with
	 (star (EndL (lookupR b r2) (lookupR b r5) (lookupR b r12))
	   (MBlkLst x0 (lookupR b r12) (0)))
	 (MBlk (plus (2) (lookupR b r6)) (lookupR b r12) x); Auto.
Apply star_assoc_R.
Apply star_subst with 
	 (EndL (lookupR b r2) (lookupR b r5) (lookupR b r12))
         (star (MBlk (plus (2) (lookupR b r6)) (lookupR b r12) x)
           (MBlkLst x0 (lookupR b r12) (0))); Auto.
Intro; Apply star_sym.
Clear H0; Intros.
Apply MidL_Good with (lookupR b r5) (lookupR b r12).
Exists x0; Auto.
Save.

Theorem WF_moremem : (Pred:assert; siz0:nat; ct:codeheapty; f:lab)
 (A_free:assert)
  ((H:heap; R,R':rfile) (Pred (H,R))
                     -> (Pred (H,R')))
  -> ((lookupCT ct L_free A_free)
   /\ ((s:state)(T_free (Pred_free_for_init Pred siz0 f) L_init s) -> (A_free s)))
  -> (Infer ct (T_moremem Pred siz0 f) C_moremem).
Proof.
Intros; Unfold T_moremem; Unfold C_moremem.
(* alloc newp[bbsize] *)
Apply I_alloc with [s:(heap*rfile)]
 (star (ptoanylist (getValinR s newp) (getValinR s bbsize))
   (star Pred (Good (getValinR s r2))) s)
    /\ (gt (getValinR s newp) O)
    /\(le (plus (2) siz0) (getValinR s r17))
    /\(gt siz0 (0))/\(getValinR s r1)=f
    /\(le (getValinR s bsize) (getValinR s bbsize)).
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
Inversion_clear H2.
Constructor; Auto.
Apply RF_Insensitive_Distrib_Star with R.
Intro; Apply ptoanylist_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
Apply star_alloc; Auto.
(* st newp(size) bbsize *)
Apply I_st with  [s:(heap*rfile)] (EX lnk |
      (star (MBlk (plus (2) (getValinR s r21)) lnk (getValinR s bbsize))
        (star Pred (Good (getValinR s r2))) s))
      /\(le (plus (2) siz0) (getValinR s r17))
        /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold getValinR; Intros.
SplitAllHyps.
Constructor.
Apply Star_ptoanylist_indom with (lookupR R r22) R
    (star Pred (Good (lookupR R r2))); Auto.
Apply lt_S_n.
Apply le_lt_n_Sm.
Apply le_trans with (lookupR R r17); Auto.
Apply le_trans with (plus (2) siz0); Auto with arith.
Constructor; Auto.
Apply Star_ptoanylist_MBlk; Auto.
Apply le_gt_trans with (lookupR R r17); Auto.
Apply le_gt_trans with (plus (2) siz0); Auto.
Unfold plus; Apply gt_n_S.
Apply gt_n_S; Auto.
(* addi fptr newp 2 *)
Apply I_addi with [s:(heap*rfile)]
      (EX lnk:nat |
         (star
           (MBlk (getValinR s fptr) lnk (getValinR s r22))
           (star Pred (Good (getValinR s r2))) s))
      /\(le (plus (2) siz0) (getValinR s r17))
        /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
SplitAllHyps.
Constructor; Auto.
Inversion_clear H3.
Exists x.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
(* movi ret0 init *)
Apply I_movi with [s:(heap*rfile)]
      (EX lnk:nat |
         (star (MBlk (getValinR s r3) lnk (getValinR s r22))
           (star Pred (Good (getValinR s r2))) s))
      /\(getValinR s ret0)=L_init
      /\(le (plus (2) siz0) (getValinR s r17))
        /\(gt siz0 (0))/\(getValinR s r1)=f.
Unfold updateR; Unfold getValinR; Intros.
Unfold lookupR; Fold lookupR; Unfold beq_reg.
SplitAllHyps; Constructor; Auto.
Inversion_clear H3.
Exists x; Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply MBlk_RF_Insensitive.
Intro; Apply RF_Insensitive_Distrib_Star.
Intro; Apply H.
Intro; Apply Good_RF_Insensitive.
(* jd free *)
Inversion_clear H0.
Apply I_jd with A_free; Auto.
Induction s; Unfold getValinR; Intros.
Apply H2.
Unfold T_free; Unfold getValinR.
SplitAllHyps.
Inversion_clear H3.
Constructor; Auto.
Exists x; Exists (lookupR b r22).
Apply star_subst_sameR with (star (MBlk (lookupR b r3) x (lookupR b r22)) Pred)
       (Good (lookupR b r2)); Auto.
Apply star_assoc_R; Auto.
Intros; Apply star_sym.
Apply star_subst_sameR with
       (MBlk (lookupR b r3) x (lookupR b r22)) Pred; Auto.
Intros; Unfold Pred_free_for_init.
Constructor; Auto.
Save.
