(****** Lemmas on Various Asserts ******)
Require Arith.
Require Export cdsa.

(***** Basic ******)

Lemma gt_lt : (p,q:nat)
 (gt p q) -> (lt q p).
Proof.
Intros; Induction p.
Inversion H.
Apply le_lt_n_Sm.
Apply gt_S_le.
Assumption.
Save.

Lemma relation_p_p'_s : (p,p',s:nat)
 (le p p') -> (lt p' (plus p s)) -> (lt (minus p' p) s).
Proof.
Induction p; Intros.
Replace (minus p' O) with p'; Auto with arith.
Induction p'.
Inversion_clear H0.
Clear Hrecp'.
Unfold minus; Fold minus.
Apply H; Auto with arith.
Save.

Lemma disj_emp_1 : (H,H':heap; R:rfile)
 (emp (H,R)) -> (hdisj H H').
Proof.
Intros.
Unfold emp in H0.
Apply disj_notboth.
Intro.
Left.
Apply notindom_none.
Unfold getValinH in H0.
Unfold hread in H0.
Apply H0.
Save.

Lemma append_emp_1 : (H,H':heap; R:rfile)
 (emp (H,R)) -> (heap_eq (happend H H') H').
Proof.
Unfold emp; Unfold getValinH; Unfold hread; Unfold happend; Intros.
Apply all_same.
Intro; Replace (H i) with (None nat); Auto.
Save.

Lemma pto_singleton : (R:rfile; i:addr; n:int)
 (pto i n ((hsingleton i n), R)).
Proof.
Intros.
Unfold pto.
Unfold getValinH; Constructor.
Unfold hsingleton; Unfold hread.
Replace (beq_nat i i) with true; Auto.
Apply beq_nat_refl.
Intros; Unfold hsingleton; Unfold hread.
Replace (beq_nat i j) with false; Auto.
Apply beq_nat_neq; Auto.
Save.

Lemma MBlkLst_pnil_pnil_emp : (n:nat; s:state)
 (MBlkLst n pnil pnil s) -> (emp s).
Proof.
Intros.
Inversion_clear H; Auto.
Inversion_clear H6.
Unfold plus in H.
Inversion_clear H.
Inversion_clear H6.
Inversion_clear H.
Save.

Lemma pto_dom : (H:heap; R:rfile; i,j:addr; x:int)
 (pto i x (H,R)) -> (hindom j H) -> (i=j).
Proof.
Unfold pto; Intros.
Inversion_clear H0.
Compare i j; Auto.
Intro; Inversion_clear H1.
Unfold getValinH in H3; Unfold hread in H3.
Cut (H j)=(None nat); Auto.
Intro.
Rewrite H0 in H1; Inversion H1.
Save.

Lemma ptoanylist_dom : (n:nat; H:heap; R:rfile; p:addr; i:nat)
 (ptoanylist p n (H,R)) -> (hindom i H) -> (le p i)/\(lt i (plus n p)).
Proof.
Induction n; Unfold ptoanylist; Fold ptoanylist; Intros.
Unfold emp in H0; Unfold getValinH in H0.
Absurd (hindom i H); Auto.
Apply notin_not_indom.
Apply notindom_none.
Unfold hread in H0; Apply H0.
Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H5.
Cut (hindom i x)\/(hnotindom i x).
Intro OR; Inversion_clear OR.
Inversion_clear H4.
Replace i with p.
Constructor; Auto.
Apply le_lt_trans with (plus n0 p); Auto with arith.
Apply pto_dom with x R x1; Auto.
Cut (le (S p) i)/\(lt i (plus n0 (S p))).
Intro HH; Inversion_clear HH; Constructor.
Apply le_trans with (S p); Auto with arith.
Replace (plus (S n0) p) with (plus n0 (S p)); Auto with arith.
Apply sym_eq; Apply plus_Snm_nSm.
Apply H with x0 R; Auto.
Cut (hindom i x0)\/(hnotindom i x0).
Intro OR; Case OR; Auto.
Intro; Absurd (hindom i H0); Auto.
Apply notin_not_indom.
Apply h_subst_notin with (happend x x0); Auto.
Apply append_preserve_notin; Auto.
Apply indom_or_notin.
Apply indom_or_notin.
Save.

Lemma MBlk_ptoanylist : (s:state; p,q:addr; siz:nat)
 (MBlk (plus (2) p) q siz s) -> (ptoanylist p siz s).
Proof.
Intros.
Inversion_clear H.
Inversion_clear H1.
Induction siz.
Inversion H.
Induction siz.
Inversion H.
Inversion H3.
Clear Hrecsiz; Clear Hrecsiz0.
Inversion_clear H2.
Inversion_clear H3.
Unfold ptoanylist; Fold ptoanylist.
Induction s; Unfold star.
Unfold removeS in H1; Unfold removeS in H2.
Unfold getValinH in H1; Unfold getValinH in H2.
Exists (hsingleton p q); Exists (hfree a p).
Constructor.
Apply disj_singleton_free.
Constructor.
Apply append_singleton_free.
Rewrite <- H1.
Replace (minus (plus (2) p) (2)) with p; Auto with arith.
Constructor.
Unfold ptoany.
Exists q; Apply pto_singleton.
Exists (hsingleton (S p) (S (S siz))); Exists (hfree (hfree a p) (S p)).
Constructor.
Apply disj_singleton_free.
Constructor.
Apply append_singleton_free.
Rewrite <- H2.
Replace (minus (plus (2) p) (1)) with (S p); Auto with arith.
Replace (minus (plus (2) p) (2)) with (p); Auto with arith.
Constructor.
Unfold ptoany; Exists (S (S siz)).
Apply pto_singleton.
Unfold removeS in H4.
Unfold plus in H4.
Unfold minus in H4; Fold minus in H4.
Cut (p=(minus p O)); Auto with arith.
Intro EQ; Rewrite <- EQ in H4.
Cut (siz=(minus siz O)); Auto with arith.
Clear EQ; Intro EQ; Rewrite <- EQ in H4; Auto.
Save.

(* Star Properties *)

Lemma star_emp : (s:state; a:assert)
 (star emp a s)
  -> ((H,H':heap;R:rfile)(a (H,R)) -> (heap_eq H H') -> (a (H',R)))
  -> (a s).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Apply H0 with x0; Auto.
Apply heap_eq_trans with (happend x x0); Auto.
Apply heap_eq_sym; Apply append_emp_1 with b; Auto.
Save.

Lemma emp_star : (s:state; a:assert)
 (a s) -> (star emp a s).
Intros; Induction s; Exists hempty.
Exists a0; Constructor.
Apply disj_emp.
Constructor.
Apply append_emp.
Constructor; Auto.
Unfold hempty; Unfold emp; Auto.
Save.

Lemma star_sym : (a1,a2:assert; s:state)
 ((star a1 a2) s) -> ((star a2 a1) s).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Exists x0; Exists x.
Constructor.
Apply disj_sym; Auto.
Constructor.
Apply heap_eq_trans with (happend x x0); Auto.
Apply append_sym; Apply disj_sym; Auto.
Constructor; Auto.
Save.

Lemma star_assoc_L : (a1,a2,a3:assert; s:state)
 ((star (star a1 a2) a3) s) -> ((star a1 (star a2 a3)) s).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H5.
Exists x1.
Exists (happend x2 x0).
Constructor.
Apply append_preserve_disj; Auto.
Apply disj_sym; Apply disj_append_L with x2.
Apply disj_sym; Auto.
Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Constructor.
Apply heap_eq_trans with (happend x x0); Auto.
Apply heap_eq_trans with (happend (happend x1 x2) x0).
Apply append_assoc_R.
Apply h_subst_append; Auto.
Constructor; Auto.
Exists x2; Exists x0.
Constructor.
Apply disj_sym; Apply disj_append_R with x1.
Apply disj_sym; Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Constructor; Auto.
Apply heap_eq_refl.
Save.

Lemma star_assoc_R : (a1,a2,a3:assert; s:state)
 ((star a1 (star a2 a3)) s) -> ((star (star a1 a2) a3) s).
Proof.
Intros; Apply star_sym.
Apply star_assoc_L.
Apply star_sym.
Apply star_assoc_L.
Apply star_sym; Auto.
Save.

Lemma star_subst : (s:state; a1,a2,a1',a2':assert)
 (star a1 a2 s)
  -> ((s0:state)(a1 s0) -> (a1' s0))
  -> ((s0:state)(a2 s0) -> (a2' s0))
  -> (star a1' a2' s).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H.
Inversion_clear H3.
Inversion_clear H4.
Exists x; Exists x0.
Constructor; Auto.
Save.

(****** Substitution with Equal Heaps ******)

(* emp on eq heaps *)
Lemma h_subst_emp : (H,H':heap; R:rfile)
 (emp (H, R)) -> (heap_eq H H') -> (emp (H', R)).
Proof.
Unfold emp; Unfold getValinH.
Intros; Replace (hread H' i) with (hread H i); Auto.
Apply h_subst_read; Auto.
Save.

(* star1 on eq heaps *)
Lemma h_subst_star1 : (H,H':heap; R:rfile; i:addr; n:int; a:assert)
 (star1 i n a (H,R)) -> (heap_eq H H')
  -> ((H,H':heap; R:rfile) (a (H,R)) -> (heap_eq H H') -> (a (H',R)))
  -> (star1 i n a (H',R)).
Proof.
Unfold star1; Unfold getValinH; Intros.
Inversion_clear H0; Constructor.
Rewrite <- H3; Apply h_subst_read; Apply heap_eq_sym; Auto.
Unfold removeS in H4; Unfold removeS.
Apply H2 with (hfree H i); Auto.
Apply h_subst_free; Auto.
Save.

(* star on eq heaps *)
Lemma h_subst_star : (H,H':heap; R:rfile; a1,a2:assert)
 (star a1 a2 (H, R)) ->
 (heap_eq H H') ->
 (star a1 a2 (H', R)).
Proof.
Intros.
Unfold star in H0.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Unfold star.
Exists x; Exists x0.
Constructor; Auto.
Constructor.
Apply heap_eq_trans with H; Auto.
Auto.
Save.

(* ptoanylist on eq heaps *)
Lemma h_subst_ptoanylist : (n:nat; H,H':heap; R:rfile; p:addr)
 (ptoanylist p n (H, R)) -> (heap_eq H H') -> (ptoanylist p n (H', R)).
Proof.
Induction n.
Intros until p; Unfold ptoanylist.
Unfold emp; Unfold getValinH.
Intros; Replace (hread H' i) with (hread H i); Auto.
Apply h_subst_read; Auto.
Intros until p.
Unfold ptoanylist; Fold ptoanylist.
Apply h_subst_star.
Save.

(* MBlk on eq heaps *)
Lemma h_subst_MBlk : (H,H':heap; R:rfile; p,q:addr; siz:nat)
 (MBlk p q siz (H, R)) -> (heap_eq H H') -> (MBlk p q siz (H', R)).
Proof.
Unfold MBlk; Unfold star1.
Unfold removeS; Unfold getValinH.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H5.
Constructor; Auto.
Constructor; Auto.
Constructor.
Replace (hread H' (minus p (2))) with (hread H (minus p (2))); Auto.
Apply h_subst_read; Auto.
Constructor.
Replace  (hread (hfree H' (minus p (2))) (minus p (1)))
  with (hread (hfree H (minus p (2))) (minus p (1))); Auto.
Apply h_subst_read; Apply h_subst_free; Auto.
Apply h_subst_ptoanylist
  with (hfree (hfree H (minus p (2))) (minus p (1))); Auto.
Apply h_subst_free; Apply h_subst_free; Auto.
Save.

(* MBlkLst on eq heaps *)
Lemma h_subst_MBlkLst : (n:nat; H,H':heap; R:rfile; p,q:addr)
 (MBlkLst n p q (H, R)) -> (heap_eq H H') -> (MBlkLst n p q (H', R)).
Proof.
Induction n; Intros.
Inversion H0; Apply mbl_O.
Apply h_subst_emp with H; Auto.
Inversion_clear H1.
Apply mbl_S with H4 H5 s p'; Auto.
Apply heap_eq_trans with H0; Auto.
Save.

(* EndL on eq heaps *)
Lemma h_subst_EndL : (H,H':heap; R:rfile; fl,p,q:addr)
 (EndL fl p q (H, R)) -> (heap_eq H H') -> (EndL fl p q (H', R)).
Proof.
Unfold EndL; Intros.
Inversion_clear H0.
Constructor.
Intro; Apply h_subst_MBlkLst with H; Auto.
Intro.
Cut ((lt p q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl p) (MBlk (plus (2) p) q s) (H,R)))); Auto.
Intro HH; Inversion_clear HH.
Constructor; Auto.
Inversion_clear H5.
Inversion_clear H6.
Exists x; Exists x0.
Apply h_subst_star with H; Auto.
Save.

(* MidL on eq heaps *)
Lemma h_subst_MidL : (H,H':heap; R:rfile; fl,p,q:addr)
 (MidL fl p q (H, R)) -> (heap_eq H H') -> (MidL fl p q (H', R)).
Proof.
Unfold MidL; Intros.
Inversion_clear H0.
Exists x.
Apply h_subst_star with H; Auto.
Save.

(* Good on eq heaps *)
Lemma h_subst_Good : (H,H':heap; R:rfile; fl:addr)
 (Good fl (H, R)) -> (heap_eq H H') -> (Good fl (H', R)).
Proof.
Unfold Good; Intros.
Inversion_clear H0.
Exists x.
Apply h_subst_MBlkLst with H; Auto.
Save.

(****** Insensitivity on Register Files ******)

(* register insensitivity is distributive on star *)
Lemma RF_Insensitive_Distrib_Star : (H:heap; R,R':rfile; a1,a2:assert)
 ((H':heap) (a1 (H',R)) -> (a1 (H',R')))
  -> ((H':heap) (a2 (H',R)) -> (a2 (H',R')))
  -> ((star a1 a2) (H,R)) -> ((star a1 a2) (H,R')).
Proof.
Unfold star; Intros.
Inversion_clear H2.
Inversion_clear H3.
Exists x; Exists x0.
Inversion_clear H2.
Inversion_clear H4.
Inversion_clear H5.
Constructor; Auto.
Save.

(* ptoanylist is insensitive on register files *)
Lemma ptoanylist_RF_Insensitive : (n:nat; H:heap; R,R':rfile; p:addr)
 (ptoanylist p n (H, R)) -> (ptoanylist p n (H, R')).
Proof.
Induction n.
Intros until p; Unfold ptoanylist.
Unfold emp.
Unfold getValinH; Trivial.
Intros until p.
Unfold ptoanylist; Fold ptoanylist.
Apply RF_Insensitive_Distrib_Star; Auto.
Intro; Apply H.
Save.

(* MBlk is insensitive on register files *)
Lemma MBlk_RF_Insensitive : (H:heap; R,R':rfile; p,q:addr; siz:nat)
 (MBlk p q siz (H, R)) -> (MBlk p q siz (H, R')).
Proof.
Unfold MBlk; Unfold star1.
Unfold removeS; Unfold getValinH.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H4.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Apply ptoanylist_RF_Insensitive with R; Auto.
Save.

(* MBlkLst is insensitive on register files *)
Lemma MBlkLst_RF_Insensitive : (n:nat; H:heap; R,R':rfile; p,q:addr)
 (MBlkLst n p q (H, R)) -> (MBlkLst n p q (H, R')).
Proof.
Induction n; Intros.
Inversion_clear H0; Apply mbl_O; Auto.
Inversion_clear H1.
Apply mbl_S with H3 H4 s p'; Auto.
Apply MBlk_RF_Insensitive with R; Auto.
Apply H with R; Auto.
Save.

(* EndL is insensitive on register files *)
Lemma EndL_RF_Insensitive : (H:heap; R,R':rfile; fl,p,q:addr)
 (EndL fl p q (H, R)) -> (EndL fl p q (H, R')).
Proof.
Unfold EndL.
Intros.
Inversion_clear H0.
Constructor.
Intro; Apply MBlkLst_RF_Insensitive with R; Auto.
Intro.
Cut ((lt p q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl p) (MBlk (plus (2) p) q s) (H,R)))); Auto.
Intro HH; Inversion_clear HH.
Constructor; Auto.
Inversion_clear H4.
Inversion_clear H5.
Exists x; Exists x0.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply MBlkLst_RF_Insensitive.
Intro; Apply MBlk_RF_Insensitive.
Save.

(* MidL is insensitive on register files *)
Lemma MidL_RF_Insensitive : (H:heap; R,R':rfile; fl,p,q:addr)
 (MidL fl p q (H, R)) -> (MidL fl p q (H, R')).
Proof.
Unfold MidL; Intros.
Inversion_clear H0.
Exists x.
Apply RF_Insensitive_Distrib_Star with R; Auto.
Intro; Apply EndL_RF_Insensitive.
Intro; Apply MBlkLst_RF_Insensitive.
Save.

(* Good is insensitive on register files *)
Lemma Good_RF_Insensitive : (H:heap; R,R':rfile; fl:addr)
 (Good fl (H, R)) -> (Good fl (H, R')).
Proof.
Unfold Good; Intros.
Inversion_clear H0.
Exists x.
Apply MBlkLst_RF_Insensitive with R; Auto.
Save.

(* Asserts Properties : about ptoanylist *)

Lemma ptoanylist_disj : (H1,H2:heap; R:rfile; p,p':addr; siz,siz':nat)
 (ptoanylist p siz (H1,R)) -> (ptoanylist p' siz' (H2,R))
  -> (le (plus p siz) p')
  -> (hdisj H1 H2).
Proof.
Intros.
Apply disj_notboth.
Intro.
Cut (hindom i H1)\/(hnotindom i H1).
Cut (hindom i H2)\/(hnotindom i H2).
Intros.
Inversion_clear H4; Auto.
Inversion_clear H5; Auto.
Absurd (lt i i); Auto with arith.
Apply lt_le_trans with p'.
Apply lt_le_trans with (plus siz p).
Apply proj2 with (le p i).
Apply ptoanylist_dom with H1 R; Auto.
Replace (plus siz p) with (plus p siz); Auto with arith.
Apply proj1 with (lt i (plus siz' p')).
Apply ptoanylist_dom with H2 R; Auto.
Apply indom_or_notin.
Apply indom_or_notin.
Save.

(* For coalescing *)
Lemma Compose_ptoanylist : (siz,siz':nat; H1,H2:heap; R:rfile; p:addr)
 (ptoanylist p siz (H1,R)) -> (ptoanylist (plus p siz) siz' (H2,R))
  -> (ptoanylist p (plus siz siz') ((happend H1 H2),R)).
Proof.
Induction siz; Intros.
Unfold ptoanylist in H; Unfold plus.
Apply h_subst_ptoanylist with H2.
Replace p with (plus p O); Auto with arith.
Apply heap_eq_sym; Apply append_emp_1 with R; Auto.
Cut (hdisj H1 H2).
Intro; Unfold ptoanylist in H0; Fold ptoanylist in H0.
Replace (plus (S n) siz') with (S (plus n siz')); Auto with arith.
Unfold ptoanylist; Fold ptoanylist.
Inversion_clear H0.
Inversion_clear H5.
Inversion_clear H0.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H6.
Exists x.
Exists (happend x0 H2).
Constructor.
Apply append_preserve_disj; Auto.
Apply disj_sym; Apply disj_append_L with x0.
Apply disj_sym; Apply h_subst_disj with H1; Auto.
Apply heap_eq_sym; Auto.
Constructor.
Apply heap_eq_trans with (happend (happend x x0) H2).
Apply append_assoc_R.
Apply h_subst_append; Auto.
Constructor.
Exists x1; Auto.
Apply H; Auto.
Replace (plus (S p) n) with (plus p (S n)); Auto.
Apply sym_eq; Apply plus_Snm_nSm.
Apply ptoanylist_disj with R p (plus p (S n)) (S n) siz'; Auto.
Save.

(* Asserts Properties : pointers > 0, siz > 2 *)

Lemma MBlk_ptr_gt_2 : (s:state; p,q:addr; siz:nat)
 (MBlk p q siz s) -> (gt p (2)).
Proof.
Unfold MBlk; Intros.
Inversion_clear H; Auto.
Save.

Lemma Star_MBlk_ptr_gt_2 : (s:state; p,q:addr; siz:nat; a:assert)
 (star (MBlk p q siz) a s) -> (gt p (2)).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Apply MBlk_ptr_gt_2 with (x,b) q siz; Assumption.
Save.

Lemma MBlk_siz_gt_2 : (s:state; p,q:addr; siz:nat)
 (MBlk p q siz s) -> (gt siz (2)).
Proof.
Unfold MBlk; Intros.
Inversion_clear H.
Inversion_clear H1.
Assumption.
Save.

Lemma Star_MBlk_siz_gt_2 : (s:state; p,q:addr; siz:nat; a:assert)
 (star (MBlk p q siz) a s) -> (gt siz (2)).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Apply MBlk_siz_gt_2 with (x,b) p q; Assumption.
Save.

Lemma MBlkLst_S_ptr_gt_O : (s:state; n:nat; p,q:addr)
 (MBlkLst (S n) p q s) -> (gt p O).
Proof.
Intros.
Inversion_clear H.
Apply gt_S_n.
Apply gt_S_n.
Replace (S (S p)) with (plus (2) p); Auto with arith.
Apply MBlk_ptr_gt_2 with (H1,R) p' s0; Assumption.
Save.

Lemma Star_MBlkLst_S_ptr_gt_O : (s:state; n:nat; p,q:addr; a:assert)
 (star (MBlkLst (S n) p q) a s) -> (gt p O).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Apply MBlkLst_S_ptr_gt_O with (x,b) n q; Assumption.
Save.

(* Deriving indom from asserts *)

Lemma ptoanylist_indom :(n,i:nat; H:heap; R:rfile; p:addr)
 (ptoanylist p n (H,R)) -> (lt i n) -> (hindom (plus i p) H).
Proof.
Induction n.
Intros.
Inversion H1.
Intros.
Unfold ptoanylist in H1; Fold ptoanylist in H1.
Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H5.
Induction i.
Unfold plus.
Inversion_clear H4.
Inversion_clear H5.
Cut (hindom p x).
Intro; Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L; Assumption.
Unfold getValinH in H4; Unfold hread in H4.
Apply indom_some with x1; Assumption.
Replace (plus (S i) p) with (plus i (S p)).
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_R; Auto.
Apply H with R; Auto with arith.
Apply sym_eq; Apply plus_Snm_nSm.
Save.

Lemma Star_ptoanylist_indom :(n,i:nat; H:heap; R:rfile; p:addr; a:assert)
 (star (ptoanylist p n) a (H,R)) -> (lt i n) -> (hindom (plus i p) H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply ptoanylist_indom with n R; Auto.
Save.

Lemma MBlk_indom : (H:heap; R:rfile; p,q:addr; siz,i:nat)
 (MBlk (plus (2) p) q siz (H,R))
  -> (lt i siz) -> (hindom (plus i p) H).
Proof.
Unfold MBlk; Unfold star1.
Intros until i.
Replace (minus (plus (2) p) (2)) with p; Auto with arith.
Replace (minus (plus (2) p) (1)) with (S p).
Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H5.
Induction i.
Unfold plus.
Unfold getValinH in H3; Unfold hread in H3.
Apply indom_some with q; Auto.
Induction i.
Unfold plus.
Unfold removeS in H4; Unfold getValinH in H4; Unfold hread in H4.
Apply indom_some with siz.
Rewrite <- H4.
Unfold hfree.
Replace (beq_nat p (S p)) with false; Auto.
Replace false with (beq_nat (S p) p).
Apply beq_nat_sym.
Apply not_beq_Sn_n.
Clear Hreci; Clear Hreci0; Clear H3; Clear H4.
Unfold removeS in H6.
Cut (hindom (plus i (S (S p))) (hfree (hfree H p) (S p))).
Intro HH; Inversion HH.
Apply indom_some with n.
Replace (plus (S (S i)) p) with (plus i (S (S p))).
Rewrite <- H3.
Unfold hfree.
Replace (beq_nat (S p) (plus i (S (S p)))) with false.
Replace (beq_nat p (plus i (S (S p)))) with false; Trivial.
Apply sym_eq; Apply not_beq_nat_small_big.
Apply lt_le_trans with (S (S p)); Auto with arith.
Apply sym_eq; Apply not_beq_nat_small_big.
Apply lt_le_trans with (S (S p)); Auto with arith.
Apply sym_eq; Apply trans_eq with (plus (S i) (S p)); Apply plus_Snm_nSm.
Apply ptoanylist_indom with (minus siz (2)) R; Auto.
Induction siz.
Inversion H0.
Induction siz.
Inversion H0.
Inversion H4.
Clear Hrecsiz; Clear Hrecsiz0; Clear H6; Clear H2.
Unfold minus; Fold minus.
Replace (minus siz (0)) with siz; Auto with arith.
Apply lt_S_n; Auto with arith.
Unfold plus; Unfold minus; Trivial.
Save.

Lemma Star_MBlk_indom : (H:heap; R:rfile; p,q:addr; siz,i:nat; a:assert)
 (star (MBlk (plus (2) p) q siz) a (H,R))
  -> (lt i siz) -> (hindom (plus i p) H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply MBlk_indom with R q siz; Auto.
Save.

Lemma MBlkLst_indom : (H:heap; R:rfile; n:nat; p,q:addr)
 (MBlkLst n p q (H,R)) -> (~p=q) -> (hindom p H).
Proof.
Intros.
Inversion H0.
Absurd p=q; Assumption.
Replace p with (plus O p); Auto.
Apply h_subst_indom with (happend H3 H4); Auto.
Apply append_preserve_indom_L.
Apply MBlk_indom with R p' s; Auto.
Apply lt_trans with (2); Auto.
Apply gt_lt.
Apply MBlk_siz_gt_2 with (H3,R) (plus (2) p) p'; Auto.
Save.

Lemma Star_MBlkLst_indom : (H:heap; R:rfile; n:nat; p,q:addr; a:assert)
 (star (MBlkLst n p q) a (H,R)) -> (~p=q) -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply MBlkLst_indom with R n q; Auto.
Save.

Lemma EndL_indom_P : (H:heap; R:rfile; fl,p,q:addr)
 (EndL fl p q (H,R)) -> (~p=pnil) -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Cut ((lt p q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl p) (MBlk (plus (2) p) q s) (H,R)))); Auto.
Clear H2; Clear H3; Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Replace p with (plus O p); Auto.
Apply Star_MBlk_indom with R q x0 (MBlkLst x fl p).
Apply star_sym; Auto.
Apply lt_trans with (2); Auto.
Apply Star_MBlk_siz_gt_2 with (H,R) (plus (2) p) q (MBlkLst x fl p).
Apply star_sym; Auto.
Save.

Lemma Star_EndL_indom_P : (H:heap; R:rfile; fl,p,q:addr; a:assert)
 (star (EndL fl p q) a (H,R)) -> (~p=pnil) -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply EndL_indom_P with R fl q; Auto.
Save.

Lemma MidL_indom_Q : (H:heap; R:rfile; fl,p,q:addr)
 (MidL fl p q (H,R)) -> (~q=pnil) -> (hindom q H).
Proof.
Intros.
Inversion_clear H0.
Apply Star_MBlkLst_indom with R x O (EndL fl p q); Auto.
Apply star_sym; Auto.
Save.

Lemma Star_MidL_indom_Q : (H:heap; R:rfile; fl,p,q:addr; a:assert)
 (star (MidL fl p q) a (H,R)) -> (~q=pnil) -> (hindom q H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply MidL_indom_Q with R fl p; Auto.
Save.

Lemma MidL_indom_P : (H:heap; R:rfile; fl,p,q:addr)
 (MidL fl p q (H,R)) -> (~p=pnil) -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Apply Star_EndL_indom_P with R fl q (MBlkLst x q (0)); Auto.
Save.

Lemma Star_MidL_indom_P : (H:heap; R:rfile; fl,p,q:addr; a:assert)
 (star (MidL fl p q) a (H,R)) -> (~p=pnil) -> (hindom p H).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Apply h_subst_indom with (happend x x0); Auto.
Apply append_preserve_indom_L.
Apply MidL_indom_P with R fl q; Auto.
Save.

(* ld & st on various asserts *)

(* ld & st on MBlk *)
Lemma Read_MBlk_lnk : (H:heap; R:rfile; p,q:addr; siz:nat)
 (MBlk (plus (2) p) q siz (H,R)) -> ((hread H p) = (hvsome q)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Unfold getValinH in H2.
Replace p with (minus (plus (2) p) (2)); Auto with arith.
Save.

Lemma Write_MBlk_lnk : (H:heap; R:rfile; p,q,q':addr; siz:nat)
 (MBlk (plus (2) p) q siz (H,R))
  -> (MBlk (plus (2) p) q' siz ((hwrite H p q'),R)).
Proof.
Unfold MBlk; Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Unfold getValinH in H2; Unfold getValinH.
Unfold hwrite; Unfold hread; Unfold hread in H2.
Replace (minus (plus (2) p) (2)) with p; Auto with arith.
Replace (beq_nat p p) with true; Auto with arith.
Apply beq_nat_refl.
Unfold removeS; Unfold removeS in H4.
Apply h_subst_star1 with (hfree H (minus (plus (2) p) (2))); Auto.
Replace (minus (plus (2) p) (2)) with p; Auto with arith.
Apply heap_eq_sym; Apply write_free.
Intros until R0; Apply h_subst_ptoanylist.
Save.

Lemma Read_MBlk_siz : (H:heap; R:rfile; p,q:addr; siz:nat)
 (MBlk (plus (1) p) q siz (H,R)) -> ((hread H p) = (hvsome siz)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H4.
Unfold removeS in H3; Unfold getValinH in H3.
Rewrite <- H3.
Replace (minus (plus (1) p) (1)) with p; Auto with arith.
Apply sym_eq; Apply local_free.
Unfold plus; Unfold minus; Fold minus.
Induction p.
Unfold plus in H1; Inversion_clear H1.
Inversion_clear H4.
Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Save.

(* For coalescing *)
Lemma Write_MBlk_siz : (H:heap; R:rfile; p,q:addr; siz,siz':nat)
 (star (MBlk (plus (2) p) q siz) (ptoanylist (plus p siz) siz') (H,R))
  -> (MBlk (plus (2) p) q (plus siz siz')
   ((hwrite H (plus (1) p) (plus siz siz')),R)).
Proof.
Intros.
Cut (gt siz (2)).
Unfold MBlk in H0; Unfold MBlk; Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H3.
Inversion_clear H6.
Constructor; Auto.
Constructor.
Apply le_gt_trans with siz; Auto with arith.
Unfold star1 in H7; Unfold star1.
Unfold removeS in H7; Unfold removeS.
Unfold getValinH in H7; Unfold getValinH.
Inversion_clear H7.
Inversion_clear H8.
Constructor.
Apply trans_eq with (hread H (minus (plus (2) p) (2))).
Apply local_write.
Unfold plus; Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Rewrite <- H6.
Apply trans_eq with (hread (happend x x0) (minus (plus (2) p) (2))).
Apply h_subst_read; Apply heap_eq_sym; Auto.
Apply append_preserve_read_L.
Apply disj_found_notin with x q; Auto.
Constructor.
Apply trans_eq with (hread (hwrite H (plus (1) p) (plus siz siz'))
   (minus (plus (2) p) (1))).
Apply local_free.
Unfold plus; Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Unfold plus; Fold plus; Unfold minus; Fold minus.
Apply write_read.
Replace (minus (plus siz siz') (2)) with (plus (minus siz (2)) siz').
Apply h_subst_ptoanylist with
 (happend (hfree (hfree x (minus (plus (2) p) (2)))
            (minus (plus (2) p) (1))) x0).
Apply Compose_ptoanylist; Auto.
Replace (plus (plus (2) p) (minus siz (2))) with (plus p siz); Auto.
Induction siz.
Inversion H1.
Induction siz.
Inversion H1.
Inversion H10.
Clear Hrecsiz; Clear Hrecsiz0.
Unfold minus; Fold minus.
Replace (minus siz O) with siz; Auto with arith.
Unfold plus; Fold plus.
Apply trans_eq with (plus (S (S p)) siz); Auto with arith.
Apply sym_eq; Apply trans_eq with (plus (S p) (S siz)); Apply plus_Snm_nSm.
Apply heap_eq_trans with
     (hfree
       (hfree (happend x x0)
         (minus (plus (2) p) (2))) (minus (plus (2) p) (1))).
Apply heap_eq_trans with
     (hfree
       (happend (hfree x
         (minus (plus (2) p) (2))) x0) (minus (plus (2) p) (1))).
Apply append_free_sym_L.
Apply disj_found_notin with x siz; Auto.
Rewrite <- H7.
Apply sym_eq; Apply local_free.
Unfold plus; Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Apply h_subst_free.
Apply append_free_sym_L.
Apply disj_found_notin with x q; Auto.
Apply heap_eq_trans with
     (hfree (hfree H (minus (plus (2) p) (2)))
       (minus (plus (2) p) (1))).
Apply h_subst_free.
Apply h_subst_free; Auto.
Apply heap_eq_trans with
     (hfree
       (hwrite
	 (hfree H (minus (plus (2) p) (2)))
	 (plus (1) p)
	 (plus siz siz'))
       (minus (plus (2) p) (1))).
Apply heap_eq_sym; Apply write_free.
Apply h_subst_free.
Apply heap_eq_sym; Apply write_free_sym.
Unfold plus; Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Induction siz.
Inversion H1.
Induction siz.
Inversion H1.
Inversion H10.
Clear Hrecsiz; Clear Hrecsiz0; Clear H6; Clear H7; Clear H9.
Unfold plus; Unfold minus; Fold minus; Fold plus.
Replace (minus siz O) with siz; Auto with arith.
Apply Star_MBlk_siz_gt_2 with (H,R) (plus (2) p) q
     (ptoanylist (plus p siz) siz'); Auto.
Save.

(* ld & st on Star MBlk *)
Lemma Read_Star_MBlk_lnk : (H:heap; R:rfile; p,q:addr; siz:nat; a:assert)
 (star (MBlk (plus (2) p) q siz) a (H,R)) -> ((hread H p) = (hvsome q)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Cut (hread x p) = (Some nat q).
Intro EQ; Rewrite <- EQ.
Replace (hread H p) with (hread (happend x x0) p).
Apply append_preserve_read_L.
Apply disj_found_notin with x q; Auto.
Apply h_subst_read; Auto.
Apply Read_MBlk_lnk with R siz; Auto.
Save.

Lemma Write_Star_MBlk_lnk :
 (H:heap; R:rfile; p,q,q':addr; siz:nat; a:assert)
 (star (MBlk (plus (2) p) q siz) a (H,R))
  -> (star (MBlk (plus (2) p) q' siz) a ((hwrite H p q'),R)).
Proof.
Unfold star; Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Exists (hwrite x p q'); Exists x0.
Cut (hread x p) = (Some nat q).
Constructor.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_found_notin with x q; Auto.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) p q').
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Apply Write_MBlk_lnk with q; Auto.
Apply Read_MBlk_lnk with R siz; Auto.
Save.

Lemma Read_Star_MBlk_siz : (H:heap; R:rfile; p,q:addr; siz:nat; a:assert)
 (star (MBlk (plus (1) p) q siz) a (H,R)) -> ((hread H p) = (hvsome siz)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Cut (hread x p) = (Some nat siz).
Intro EQ; Rewrite <- EQ.
Replace (hread H p) with (hread (happend x x0) p).
Apply append_preserve_read_L.
Apply disj_found_notin with x siz; Auto.
Apply h_subst_read; Auto.
Apply Read_MBlk_siz with R q; Auto.
Save.

(* For coalescing *)
Lemma Write_Star_MBlk_siz :
 (H:heap; R:rfile; p,q:addr; siz,siz':nat; a:assert)
 (star (star (MBlk (plus (2) p) q siz) (ptoanylist (plus p siz) siz'))
  a (H,R))
  -> (star (MBlk (plus (2) p) q (plus siz siz'))
  a ((hwrite H (plus (1) p) (plus siz siz')),R)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Exists (hwrite x (plus (1) p) (plus siz siz')); Exists x0.
Cut (hread x (plus (1) p)) = (Some nat siz).
Constructor.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_found_notin with x siz; Auto.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) (plus (1) p) (plus siz siz')).
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Apply Write_MBlk_siz; Auto.
Apply Read_Star_MBlk_siz with R q (ptoanylist (plus p siz) siz'); Auto.
Save.

(* ld & st on EndL *)
Lemma Read_EndL_lnk : (H:heap; R:rfile; fl,p,q:addr)
 (EndL fl p q (H,R)) -> (~p=pnil) -> ((hread H p) = (hvsome q)).
Proof.
Intros.
Inversion_clear H0.
Cut ((lt p q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl p) (MBlk (plus (2) p) q s) (H,R)))); Auto.
Intros.
Inversion_clear H0.
Inversion_clear H5.
Inversion_clear H0.
Apply Read_Star_MBlk_lnk with R x0 (MBlkLst x fl p).
Apply star_sym; Auto.
Save.

Lemma Write_EndL_lnk : (H:heap; R:rfile; fl,p,q,q':addr)
 (EndL fl p q (H,R)) -> (~p=pnil) -> ((lt p q') \/ (q'=pnil))
  -> (EndL fl p q' ((hwrite H p q'),R)).
Unfold EndL; Intros.
Constructor.
Intro; Absurd p=O; Auto.
Intro; Constructor; Auto.
Inversion_clear H0.
Cut ((lt p q)\/q=(0))
            /\(EX n:nat |
                 (EX s:nat |
                    (star (MBlkLst n fl p) (MBlk (plus (2) p) q s)
                      (H,R)))); Auto.
Intro.
Clear H4; Clear H5; Inversion_clear H0.
Inversion_clear H5.
Inversion_clear H0.
Exists x; Exists x0.
Apply star_sym; Apply Write_Star_MBlk_lnk with q.
Apply star_sym; Auto.
Save.

(* ld & st on Star EndL *)
Lemma Read_Star_EndL_lnk : (H:heap; R:rfile; fl,p,q:addr; a:assert)
 (star (EndL fl p q) a (H,R)) -> (~p=pnil) -> ((hread H p) = (hvsome q)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Cut (hread x p) = (Some nat q).
Intro EQ; Rewrite <- EQ.
Replace (hread H p) with (hread (happend x x0) p).
Apply append_preserve_read_L.
Apply disj_found_notin with x q; Auto.
Apply h_subst_read; Auto.
Apply Read_EndL_lnk with R fl; Auto.
Save.

Lemma Write_Star_EndL_lnk : (H:heap; R:rfile; fl,p,q,q':addr; a:assert)
 (star (EndL fl p q) a (H,R)) -> (~p=pnil) -> ((lt p q') \/ (q'=pnil))
  -> (star (EndL fl p q') a ((hwrite H p q'),R)).
Proof.
Unfold star; Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H4.
Inversion_clear H5.
Exists (hwrite x p q'); Exists x0.
Cut (hread x p) = (Some nat q).
Constructor.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_found_notin with x q; Auto.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) p q').
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Apply Write_EndL_lnk with q; Auto.
Apply Read_EndL_lnk with R fl; Auto.
Save.

(* Composing Blocks *)

Lemma Compose_MBlk_MBlkLst_to_MBlkLst : (s:state; p,p',q:addr; n,siz:nat)
 (star (MBlk (plus (2) p) p' siz) (MBlkLst n p' q) s)
  -> ((lt p p') \/ (p'=pnil))
  -> (MBlkLst (S n) p q s).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Apply mbl_S with x x0 siz p'; Auto.
Save.

Lemma Compose_MBlkLst_MBlk_to_MBlkLst : (n:nat; s:state; fl,p,q:addr; siz:nat)
 (star (MBlkLst n fl p) (MBlk (plus (2) p) q siz) s)
  -> ((lt p q) \/ (q=pnil))
  -> (MBlkLst (S n) fl q s).
Proof.
Induction n; Induction s; Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Apply mbl_S with x0 x siz q; Auto.
Apply disj_sym; Auto.
Apply heap_eq_trans with (happend x x0); Auto.
Apply append_sym; Apply disj_sym; Auto.
Apply mbl_O; Auto.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H3.
Apply mbl_S with H6 (happend H7 x0) s0 p'; Auto.
Apply append_preserve_disj; Auto.
Apply disj_sym.
Apply disj_append_L with H7.
Apply disj_sym; Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Apply heap_eq_trans with (happend (happend H6 H7) x0).
Apply append_assoc_R.
Apply heap_eq_trans with (happend x x0); Auto.
Apply h_subst_append; Auto.
Apply H with fl:=p' p:=p q:=q siz:=siz; Auto.
Unfold star.
Exists H7; Exists x0.
Constructor.
Apply disj_sym; Apply disj_append_R with H6.
Apply disj_sym; Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Constructor.
Apply heap_eq_refl.
Constructor; Auto.
Save.

Lemma Compose_EndL_MBlk_to_EndL :
 (H1,H2:heap; R:rfile; fl,p,q,q':addr; siz:nat)
 (EndL fl p q (H1,R)) -> (MBlk (plus (2) q) q' siz (H2,R))
  -> ((lt q q')\/(q'=pnil))
  -> (hdisj H1 H2)
  -> (EndL fl q q' ((happend H1 H2),R)).
Proof.
Unfold EndL.
Intros.
Constructor.
Cut (gt (plus (2) q) (2)); Intros.
Rewrite H6 in H5; Inversion H5.
Inversion H8.
Inversion H10.
Apply MBlk_ptr_gt_2 with (H2,R) q' siz; Auto.
Intro; Constructor; Auto.
Inversion_clear H.
Induction p.
Exists O; Exists siz.
Unfold star.
Cut (MBlkLst (0) fl q (H1,R)).
Intro EMP; Inversion_clear EMP.
Exists H1; Exists H2.
Constructor.
Apply disj_emp_1 with R; Auto.
Constructor.
Apply heap_eq_refl.
Constructor; Auto.
Apply mbl_O; Auto.
Apply H6; Auto.
Clear Hrecp; Clear H6.
Cut ((lt (S p) q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl (S p)) (MBlk (plus (2) (S p)) q s)
                   (H1,R)))); Auto.
Clear H7; Intros.
Inversion_clear H.
Inversion_clear H7.
Inversion_clear H.
Exists (S x); Exists siz.
Unfold star.
Exists H1; Exists H2.
Constructor; Auto.
Constructor.
Apply heap_eq_refl.
Constructor; Auto.
Apply Compose_MBlkLst_MBlk_to_MBlkLst with (S p) x0; Auto.
Save.

(* Linking "list + block + list" to be a list *)
Lemma Compose_Blk : (n1,n2:nat; s:state; fl,p,q:addr; siz:nat)
 ((star
     (star
       (MBlkLst n1 fl p)
       (MBlk (plus (2) p) q siz)
     )
     (MBlkLst n2 q pnil)
   ) s)
 -> ((lt p q) \/ (q=pnil))
 -> ((MBlkLst (plus n1 (S n2)) fl pnil) s).
Proof.
Induction n1; Intros.
Cut ((star (MBlkLst (0) fl p)
     (star (MBlk (plus (2) p) q siz) (MBlkLst n2 q (0)))) s).
Clear H; Intro.
Induction s; Unfold star in H.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H5.
Inversion_clear H6.
Unfold plus.
Inversion_clear H2.
Apply mbl_S with x1 x2 siz q; Auto.
Apply heap_eq_trans with x0; Auto.
Apply heap_eq_trans with (happend x x0); Auto.
Apply heap_eq_sym; Apply append_emp_1 with b; Auto.
Apply star_assoc_L; Auto.
Cut (star (MBlkLst (S n) fl p) (
   star (MBlk (plus (2) p) q siz) (MBlkLst n2 q (0))) s).
Intro; Clear H0.
Replace (plus (S n) (S n2)) with (S (plus n (S n2))); Auto with arith.
Induction s; Unfold star in H2.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H3.
Apply mbl_S with H6 (happend H7 x0) s p'; Auto.
Apply append_preserve_disj; Auto.
Apply disj_sym; Apply disj_append_L with H7.
Apply disj_sym; Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Apply heap_eq_trans with (happend (happend H6 H7) x0).
Apply append_assoc_R.
Apply heap_eq_trans with (happend x x0); Auto.
Apply h_subst_append; Auto.
Apply H with p q siz; Auto.
Unfold star.
Inversion_clear H5.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H5.
Inversion_clear H13.
Exists (happend x1 H7); Exists x2.
Constructor.
Apply disj_sym; Apply append_preserve_disj.
Apply disj_sym; Auto.
Apply disj_append_R with H6.
Apply disj_sym; Apply disj_append_R with x1.
Apply h_subst_disj with x.
Apply heap_eq_sym; Auto.
Apply disj_sym; Apply h_subst_disj with x0.
Apply heap_eq_sym; Auto.
Apply disj_sym; Auto.
Constructor.
Apply heap_eq_trans with  (happend (happend H7 x1) x2).
Apply h_subst_append; Apply append_sym.
Apply disj_append_R with H6.
Apply disj_sym; Apply disj_append_L with x2.
Apply h_subst_disj with x.
Apply heap_eq_sym; Auto.
Apply disj_sym; Apply h_subst_disj with x0.
Apply heap_eq_sym; Auto.
Apply disj_sym; Auto.
Apply heap_eq_trans with (happend H7 (happend x1 x2)).
Apply append_assoc_L.
Apply h_subst_append_R; Auto.
Constructor; Auto.
Exists H7; Exists x1.
Cut (hdisj H7 x1).
Intro; Constructor; Auto.
Constructor.
Apply append_sym; Auto.
Constructor; Auto.
Apply disj_append_L with x2.
Apply disj_sym; Apply disj_append_R with H6.
Apply h_subst_disj with x0.
Apply heap_eq_sym; Auto.
Apply disj_sym; Apply h_subst_disj with x; Auto.
Apply heap_eq_sym; Auto.
Apply star_assoc_L; Auto.
Save.

(* Frame rule for star for local reasoning *)
Lemma star_frame : (H:heap; R:rfile; a1,a2,a1':assert; i:addr; n:int)
 (star a1 a2 (H,R))
  -> ((H:heap;R:rfile)(a1 (H,R)) -> ((hindom i H) /\ (a1' ((hwrite H i n),R))))
  -> (star a1' a2 ((hwrite H i n),R)).
Proof.
Unfold star; Intros.
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H4.
Exists (hwrite x i n); Exists x0.
Cut (hindom i x)/\(a1' ((hwrite x i n),R)); Auto.
Clear H1; Intro H1; Inversion_clear H1.
Constructor.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_indom_notin with x; Auto.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) i n).
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Save.

(****** Others ******)

(* MidL pnil equals EndL pnil *)
Lemma MidL_pnil_EndL : (s:state; fl,p:addr)
 (MidL fl p pnil s) -> (EndL fl p pnil s).
Proof.
Intros.
Inversion_clear H.
Induction s; Inversion_clear H0.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Apply h_subst_EndL with x0; Auto.
Apply heap_eq_trans with (happend x1 x0).
Apply heap_eq_sym; Apply append_emp_1 with b.
Apply MBlkLst_pnil_pnil_emp with x; Auto.
Apply heap_eq_trans with (happend x0 x1); Auto.
Apply append_sym.
Apply disj_sym; Auto.
Save.

(* EndL pnil equals MidL pnil *)
Lemma EndL_pnil_MidL : (s:state; fl,p:addr)
 (EndL fl p pnil s) -> (MidL fl p pnil s).
Proof.
Intros; Unfold MidL.
Exists O.
Induction s; Unfold star.
Exists a; Exists hempty.
Constructor.
Apply disj_sym; Apply disj_emp.
Constructor.
Apply heap_eq_trans with (happend hempty a).
Apply append_sym.
Apply disj_sym; Apply disj_emp.
Apply append_emp.
Constructor; Auto.
Apply mbl_O.
Unfold emp; Unfold getValinH; Unfold hempty; Unfold hread; Trivial.
Save.

(* Next element in list *)
Lemma MidL_next : (s:state; fl,p,q,q':addr)
 (MidL fl p q s) -> (~q=pnil)
  -> (EX q' | ((getValinH s q)=(hvsome q')) /\ (MidL fl q q' s)).
Proof.
Unfold MidL; Intros.
Inversion_clear H.
Induction s; Inversion_clear H1.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H3.
Inversion H4.
Absurd q=O; Assumption.
Exists p'; Constructor.
Unfold getValinH.
Cut (hread H5 q) = (Some nat p').
Intro.
Cut (hread x1 q) = (Some nat p').
Intro.
Replace (hread a q) with (hread (happend x0 x1) q).
Replace (hread (happend x0 x1) q) with (hread x1 q); Auto.
Apply sym_eq; Apply append_preserve_read_R.
Apply disj_found_notin with x1 p'; Auto.
Apply disj_sym; Auto.
Apply h_subst_read; Auto.
Replace (hread x1 q) with (hread (happend H5 H6) q).
Rewrite <- H17.
Apply append_preserve_read_L.
Apply disj_found_notin with H5 p'; Auto.
Apply h_subst_read; Auto.
Apply Read_MBlk_lnk with b s; Assumption.
Exists n.
Clear H4; Unfold star.
Exists (happend x0 H5); Exists H6.
Constructor.
Apply disj_sym; Apply append_preserve_disj.
Apply disj_sym; Apply disj_append_R with H5.
Apply disj_sym; Apply h_subst_disj with x1.
Apply heap_eq_sym; Assumption.
Apply disj_sym; Assumption.
Apply disj_sym; Assumption.
Constructor.
Apply heap_eq_trans with (happend x0 (happend H5 H6)).
Apply append_assoc_L.
Apply heap_eq_trans with (happend x0 x1); Auto.
Apply h_subst_append_R; Auto.
Constructor; Auto.
Apply Compose_EndL_MBlk_to_EndL with p s; Auto.
Apply disj_append_L with H6.
Apply disj_sym.
Apply h_subst_disj with x1.
Apply heap_eq_sym; Auto.
Apply disj_sym; Auto.
Save.

Lemma Star_MidL_next : (s:state; fl,p,q,q':addr; a:assert)
 (star (MidL fl p q) a s) -> (~q=pnil)
  -> (EX q' | ((getValinH s q)=(hvsome q')) /\ (star (MidL fl q q') a s)).
Proof.
Induction s; Unfold star; Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Cut (EX q' | (getValinH (x,b) q)=(Some nat q') /\ (MidL fl q q' (x,b))).
Intro EE; Inversion_clear EE.
Exists x1.
Inversion_clear H3; Constructor.
Unfold getValinH; Unfold getValinH in H5.
Rewrite <- H5.
Replace (hread a q) with (hread (happend x x0) q).
Apply append_preserve_read_L.
Apply disj_found_notin with x x1; Auto.
Apply h_subst_read; Auto.
Exists x; Exists x0.
Constructor; Auto.
Apply MidL_next with p; Auto.
Save.

(* Unfold MBlkLst *)
Lemma Star_Unfold_MBlkLst : (s:state; n:nat; p,q:addr; a:assert)
 (star (MBlkLst (S n) p q) a s)
  -> (EX p' | (EX siz |
     (star (star (MBlk (plus (2) p) p' siz) (MBlkLst n p' q)) a s)
      /\ ((lt p p') \/ (p'=pnil)))).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Exists p'; Exists s.
Constructor; Auto.
Exists x; Exists x0.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Exists H4; Exists H5.
Constructor; Auto.
Save.

(* Unfold EndL *)
Lemma Star_Unfold_EndL : (s:state; fl,p,q:addr; a:assert)
 (star (EndL fl p q) a s)
  -> (~p=pnil)
  -> (EX n | (EX siz |
     (star (
	star (MBlkLst n fl p) (MBlk (plus (2) p) q siz)
       ) a s)))
	  /\ ((lt p q)\/(q=pnil)).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Inversion_clear H2.
Cut ((lt p q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (star (MBlkLst n fl p) (MBlk (plus (2) p) q s) (x,b)))); Auto.
Clear H3; Clear H5; Intros.
Inversion_clear H2.
Inversion_clear H5.
Inversion_clear H2.
Constructor; Auto.
Exists x1; Exists x2.
Exists x; Exists x0.
Constructor; Auto.
Save.

(* Unfolding MidL *)
Lemma Star_Unfold_MidL : (s:state; fl,p,q:addr; a:assert)
 (star (MidL fl p q) a s)
  -> (EX n | (star (star (EndL fl p q) (MBlkLst n q pnil)) a s)).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Exists x1.
Exists x; Exists x0.
Constructor; Auto.
Save.

(* MidL implies Good *)
Lemma MidL_Good : (s:state; fl,p,q:addr)
 (MidL fl p q s) -> (Good fl s).
Proof.
Unfold MidL; Unfold Good; Intros.
Inversion_clear H.
Induction s; Unfold EndL in H0; Unfold star in H0.
Inversion_clear H0.
Inversion_clear H.
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H1.
Induction p.
Exists x.
Cut (MBlkLst O fl q (x0,b)); Auto.
Intro LO; Inversion_clear LO.
Apply h_subst_MBlkLst with x1; Auto.
Apply heap_eq_trans with (happend x0 x1); Auto.
Apply heap_eq_sym; Apply append_emp_1 with b; Auto.
Clear H2; Clear Hrecp.
Cut ((lt (S p) q)\/q=(0))
         /\(EX n:nat |
              (EX s:nat |
                 (EX H1:heap |
                    (EX H2:heap |
                       (hdisj H1 H2)
                       /\(heap_eq (happend H1 H2) x0)
                         /\(MBlkLst n fl (S p) (H1,b))
                           /\(MBlk (plus (2) (S p)) q s (H2,b)))))); Auto.
Intro; Clear H4.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H5.
Inversion_clear H6.
Exists (plus x2 (S x)).
Apply Compose_Blk with (S p) q x3; Auto.
Unfold star.
Exists x0; Exists x1.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Exists x4; Exists x5.
Constructor; Auto.
Save.

(* (MBlk (p+2) _ s) means from p to p+s-1 all occupied *)
Lemma MBlk_Occupied : (H:heap; R:rfile; p,p',q:addr; siz:nat)
 (MBlk (plus (2) p) q siz (H,R))
 -> ((le p p') /\ (lt p' (plus p siz)))
  -> (hindom p' H).
Proof.
Intros.
Inversion_clear H1.
Replace p' with (plus (minus p' p) p).
Apply MBlk_indom with R q siz; Auto.
Apply relation_p_p'_s; Auto.
Apply trans_eq with (plus p (minus p' p)); Auto with arith.
Save.

Lemma Star_MBlk_Occupied : (s:state; p,p',q:addr; siz:nat; a:assert)
 (star (MBlk (plus (2) p) q siz) a s)
  -> ((H:heap; R:rfile) (a (H,R)) -> (hindom p' H))
  -> (~((le p p') /\ (lt p' (plus p siz)))).
Proof.
Intros.
Intro.
Inversion_clear H1.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H4.
Inversion_clear H5.
Absurd (hindom p' x).
Intro; Cut (hindom p' x0).
Intro; Absurd (p'=p'); Auto.
Apply disj_indom_neq with x x0; Auto.
Apply H0 with b; Auto.
Apply MBlk_Occupied with b p q siz; Auto.
Save.
