(****** More Lemmas on Various Asserts ******)
Require Arith.
Require Export cdsa.
Require property.

Lemma star_subst_sameR : (H:heap; R:rfile; a1,a2,a1',a2':assert)
 (star a1 a2 (H,R))
  -> ((H0:heap)(a1 (H0,R)) -> (a1' (H0,R)))
  -> ((H0:heap)(a2 (H0,R)) -> (a2' (H0,R)))
  -> (star a1' a2' (H,R)).
Proof.
Unfold star; Intros.
Inversion_clear H0.
Inversion_clear H3.
SplitAllHyps.
Exists x; Exists x0.
Constructor; Auto.
Save.

Lemma lt_neq : (a,b:nat)
 (lt a b) -> (~a=b).
Proof.
Double Induction 1 2; Auto.
Intro OO; Inversion OO.
Intros.
Intro.
Absurd n0=n.
Apply H0.
Apply lt_S_n; Auto.
Inversion H2; Auto.
Save.

Lemma lt_neq_rev : (a,b:nat)
 (lt a b) -> (~b=a).
Proof.
Unfold not; Intros.
Absurd a=b.
Apply lt_neq; Auto.
Rewrite H0; Trivial.
Save.

Lemma gt_or_le : (a,b:nat)
 (gt a b) \/ (le a b).
Proof.
Double Induction 1 2; Auto with arith.
Intros.
Cut (gt n0 n) \/ (le n0 n); Auto.
Intro OR; Inversion_clear OR.
Left; Apply gt_n_S; Auto.
Right; Auto with arith.
Save.

Lemma le_minus_eq_O : (a,b:nat)
 (le a b) -> (minus a b) = O.
Proof.
Induction a.
Intros; Unfold minus; Auto.
Induction b.
Intro LEO; Inversion LEO.
Intros.
Unfold minus; Fold minus.
Apply H.
Auto with arith.
Save.

Lemma lt_gt : (a,b:nat)
 (lt a b) -> (gt b a).
Proof.
Intros.
Inversion_clear H; Auto.
Save.

Lemma gt_minus_gt_O : (a,b:nat)
 (gt a b) -> (gt (minus a b) O).
Proof.
Induction a.
Intros.
Inversion H.
Intros; Induction b.
Unfold minus; Auto.
Unfold minus; Fold minus.
Apply H.
Auto with arith.
Save.

Lemma minus_plus_eq_minus_minus : (a,b,c:nat)
 (minus a (plus b c)) = (minus (minus a b) c).
Proof.
Induction a.
Unfold minus; Auto.
Intros; Induction b.
Unfold plus; Replace (minus (S n) O) with (S n); Auto.
Unfold plus; Fold plus.
Unfold minus; Fold minus.
Apply H.
Save.

Lemma plus_minus_eq_minus_plus : (a,b,c:nat)
 (gt a b) -> (plus (minus a b) c) = (minus (plus a c) b).
Proof.
Double Induction 1 2; Auto with arith.
Intros.
Inversion H0.
Intros.
Unfold plus; Fold plus; Unfold minus; Fold minus.
Apply H0.
Auto with arith.
Save.

Lemma lt_minus_sym : (a,b,c:nat)
 (lt a (minus b c)) -> (lt c (minus b a)).
Proof.
Induction a.
Intros.
Replace (minus b O) with b; Auto with arith.
Intros.
Induction b.
Unfold minus in H0; Inversion H0.
Unfold minus; Fold minus.
Apply H.
Apply lt_S_n.
Cut (gt b c)\/(le b c).
Intro OR; Inversion_clear OR.
Replace (S (minus b c)) with (minus (S b) c); Auto.
Replace (S b) with (plus b (1)).
Replace (S (minus b c)) with (plus (minus b c) (1)).
Apply sym_eq; Apply plus_minus_eq_minus_plus; Auto.
Apply trans_eq with (plus (1) (minus b c)); Auto with arith.
Apply trans_eq with (plus (1) b); Auto with arith.
Inversion H1.
Rewrite H2 in H0.
Cut (lt (S n) (minus (plus c (1)) (plus c (0)))).
Replace (minus (plus c (1)) (plus c (0))) with (minus (1) O); Auto with arith.
Replace (minus c c) with O; Auto with arith.
Replace (plus c (1)) with (S c).
Replace (plus c O) with c; Auto with arith.
Apply trans_eq with (plus (1) c); Auto with arith.
Rewrite <- H3 in H0.
Cut (minus (S b) (S m)) = O.
Intro EQO; Rewrite EQO in H0; Inversion H0.
Apply le_minus_eq_O.
Auto with arith.
Apply gt_or_le.
Save.

Lemma lt_minus_bothside : (a,b,c:nat)
 (le a b) -> (lt b c) -> (lt (minus b a) (minus c a)).
Proof.
Induction a.
Intros.
Replace (minus b O) with b; Auto with arith.
Replace (minus c O) with c; Auto with arith.
Intros.
Induction b.
Inversion H0.
Induction c.
Inversion H1.
Clear Hrecb; Clear Hrecc.
Unfold minus; Fold minus.
Apply H; Auto with arith.
Save.

Lemma MBlkLst_O_eq : (s:state; p,q:addr)
 (MBlkLst O p q s) -> (p=q).
Proof.
Intros.
Inversion_clear H; Auto.
Save.

Lemma MBlkLst_neq_pnil_S : (s:state; p:addr; n:nat)
 (MBlkLst n p pnil s) -> (~p=pnil) -> (EX m | (MBlkLst (S m) p pnil s)).
Proof.
Intros.
Inversion H.
Absurd p=O; Auto.
Exists n0; Apply mbl_S with H2 H3 s0 p'; Auto.
Save.

Lemma Star_MBlkLst_neq_pnil_S : (s:state; p:addr; n:nat; a:assert)
 (star (MBlkLst n p pnil) a s)
  -> (~p=pnil)
  -> (EX m | (star (MBlkLst (S m) p pnil) a s)).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Cut (EX m | (MBlkLst (S m) p O (x,b))).
Intro.
Inversion_clear H3.
Exists x1.
Exists x; Exists x0; Auto.
Apply MBlkLst_neq_pnil_S with n; Auto.
Save.

Lemma Star_RF_Frame : (H:heap; R:rfile; Pred,Pred',a:assert; r:reg; n:int)
 (star Pred a (H,R))
  -> ((H:heap; R:rfile; m:int) (Pred (H,R)) -> (Pred' (H, (updateR R r m))))
  -> ((H:heap; R:rfile; m:int) (a (H,R)) -> (a (H, (updateR R r m))))
  -> (star Pred' a (H,(updateR R r n))).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H4.
Inversion_clear H5.
Exists x; Exists x0.
Constructor; Auto.
Save.

Lemma star1_star : (s:state; p:addr; n:nat; a:assert)
 (star1 p n a s) -> (star (pto p n) a s).
Proof.
Intros.
Inversion_clear H.
Induction s.
Unfold removeS in H1.
Exists (hsingleton p n); Exists (hfree a0 p).
Constructor.
Apply disj_notboth.
Intro.
Cut (hindom i (hsingleton p n))\/(hnotindom i (hsingleton p n)).
Intro OR; Inversion_clear OR; Auto.
Right.
Inversion_clear H.
Unfold hsingleton in H2.
Replace i with p.
Apply free_notin.
Apply beq_nat_eq.
Cut (false=(beq_nat p i) \/ true=(beq_nat p i)).
Intro OR; Inversion_clear OR; Auto.
Rewrite <- H in H2; Inversion H2.
Generalize (beq_nat p i); Induction b0; Auto.
Apply indom_or_notin.
Constructor.
Apply append_singleton_free.
Unfold getValinH in H0; Auto.
Constructor; Auto.
Apply pto_singleton.
Save.

Lemma star_star1 : (s:state; p:addr; n:nat; a:assert)
 (star (pto p n) a s)
  -> ((H,H':heap;R:rfile) (a (H,R)) -> (heap_eq H H') -> (a (H',R)))
  -> (star1 p n a s).
Proof.
Intros.
Induction s; Inversion_clear H.
Inversion_clear H1.
Inversion_clear H.
Inversion_clear H2.
Inversion_clear H3.
Unfold star1.
Constructor.
Unfold getValinH.
Apply trans_eq with (hread (happend x x0) p); Auto.
Apply h_subst_read; Apply heap_eq_sym; Auto.
Apply trans_eq with (hread x p).
Apply append_preserve_read_L.
Apply disj_indom_notin with x; Auto.
Inversion_clear H2.
Unfold getValinH in H3.
Apply indom_some with n; Auto.
Inversion_clear H2.
Unfold getValinH in H3; Auto.
Unfold removeS.
Apply H0 with x0; Auto.
Apply all_same.
Intro.
Unfold hfree.
Cut (false=(beq_nat p i))\/(true=(beq_nat p i)).
Intro OR; Inversion_clear OR.
Rewrite <- H3.
Replace (x0 i) with (hread x0 i); Auto.
Replace (a0 i) with (hread a0 i); Auto.
Apply trans_eq with (hread (happend x x0) i); Auto.
Apply sym_eq; Apply append_preserve_read_R.
Inversion_clear H2.
Apply notindom_none.
Unfold getValinH in H6; Unfold hread in H6.
Apply H6.
Intro.
Rewrite H2 in H3; Cut true=(beq_nat i i).
Intro.
Rewrite <- H7 in H3; Inversion H3.
Apply beq_nat_refl.
Apply h_subst_read; Auto.
Rewrite <- H3.
Cut (hnotindom i x0).
Intro.
Inversion_clear H5; Auto.
Apply disj_indom_notin with x; Auto.
Inversion_clear H2.
Apply indom_some with n.
Unfold getValinH in H5; Unfold hread in H5; Replace i with p; Auto.
Apply beq_nat_eq; Auto.
Generalize (beq_nat p i); Induction b0; Auto.
Save.

Lemma ptoanylist_MBlk : (H:heap; R:rfile; p:addr; siz:nat)
 (ptoanylist p siz (H,R)) -> (gt p O) -> (gt siz (2))
  -> (EX q |
   (MBlk (plus (2) p) q siz ((hwrite H (plus (1) p) siz),R))).
Proof.
Intros.
Induction siz.
Inversion H2.
Induction siz.
Inversion H2.
Inversion H4.
Clear Hrecsiz; Clear Hrecsiz0.
Unfold ptoanylist in H0; Fold ptoanylist in H0.
Unfold plus.
Cut (EX lnk | (EX y | (star (pto p lnk)
         (star (pto (S p) y) (ptoanylist (S (S p)) siz)) (H,R)))).
Clear H0; Intros.
Inversion_clear H0.
Inversion_clear H3.
Exists x.
Constructor; Auto with arith.
Constructor; Auto.
Apply star_star1.
Unfold minus; Fold minus.
Replace (minus p O) with p; Auto with arith.
Apply star_sym; Apply star_frame with
   (star (pto (S p) x0) (ptoanylist (S (S p)) siz)).
Apply star_sym; Auto.
Clear H0; Intros.
Constructor.
Apply indom_some with x0.
Inversion_clear H3.
Inversion_clear H4.
Inversion_clear H3.
Inversion_clear H5.
Inversion_clear H6.
Cut (hnotindom (S p) x2).
Intros; Replace (H0 (S p)) with (hread H0 (S p)); Auto.
Apply trans_eq with (hread (happend x1 x2) (S p)).
Apply sym_eq; Apply h_subst_read; Auto.
Apply trans_eq with (hread x1 (S p)).
Apply append_preserve_read_L; Auto.
Inversion_clear H5.
Unfold getValinH in H8; Auto.
Apply disj_found_notin with x1 x0; Auto.
Inversion_clear H5.
Unfold getValinH in H8; Auto.
Replace (minus siz O) with siz; Auto with arith.
Apply star_star1.
Apply star_frame with (pto (S p) x0); Auto.
Clear H3; Intros.
Inversion_clear H4.
Unfold getValinH in H5.
Constructor.
Apply indom_some with x0; Unfold hread in H5; Auto.
Unfold pto; Constructor.
Unfold getValinH.
Apply write_read.
Intros.
Unfold getValinH in H6; Unfold getValinH.
Apply trans_eq with (hread H3 j); Auto.
Apply local_write; Auto.
Intros until R1; Apply h_subst_ptoanylist.
Intros.
Inversion_clear H4.
Unfold star1; Constructor.
Rewrite <- H6.
Unfold getValinH; Apply h_subst_read; Apply heap_eq_sym; Auto.
Unfold removeS; Unfold removeS in H7.
Apply h_subst_ptoanylist with (hfree H3 (minus (S (S p)) (1))); Auto.
Apply h_subst_free; Auto.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H4.
Inversion_clear H5.
Inversion_clear H4.
Exists x1.
Inversion_clear H6.
Inversion_clear H4.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H7.
Exists  x4.
Exists x; Exists x0.
Constructor; Auto.
Constructor; Auto.
Constructor; Auto.
Exists x2; Exists x3.
Constructor; Auto.
Save.

Lemma Star_ptoanylist_MBlk: (H:heap; R:rfile; p:addr; siz:nat; a:assert)
 (star (ptoanylist p siz) a (H,R)) -> (gt p O) -> (gt siz (2))
  -> (EX q |
   (star (MBlk (plus (2) p) q siz) a ((hwrite H (plus (1) p) siz),R))).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Inversion_clear H0.
Inversion_clear H4.
Inversion_clear H5.
Cut (EX q | (MBlk (plus (2) p) q siz ((hwrite x (plus (1) p) siz),R))).
Intros.
Inversion_clear H5.
Exists x1.
Exists (hwrite x (plus (1) p) siz); Exists x0.
Constructor; Auto.
Apply notin_write_preserve_disj_L; Auto.
Apply disj_indom_notin with x; Auto.
Apply ptoanylist_indom with siz R; Auto.
Apply gt_lt; Apply gt_trans with (2); Auto with arith.
Constructor.
Apply heap_eq_trans with (hwrite (happend x x0) (plus (1) p) siz).
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Constructor; Auto.
Apply ptoanylist_MBlk; Auto.
Save.

(* For splitting *)
Lemma Unfold_ptoanylist : (siz,siz':nat; s:state; p:addr)
 (ptoanylist p (plus siz siz') s)
 -> (star (ptoanylist p siz) (ptoanylist (plus p siz) siz') s).
Proof.
Induction siz.
Intros.
Apply star_subst with emp (ptoanylist (plus p (0)) siz'); Auto.
Apply emp_star.
Replace (plus p O) with p; Auto with arith.
Intros.
Cut (plus (S n) siz') = (S (plus n siz')); Auto.
Intro EQ; Rewrite EQ in H0.
Unfold ptoanylist in H0; Fold ptoanylist in H0.
Apply star_subst with
(star (ptoany p) (ptoanylist (S p) n))
(ptoanylist (plus p (S n)) siz'); Auto.
Apply star_assoc_R.
Apply star_subst with (ptoany p) (ptoanylist (S p) (plus n siz')); Auto.
Intro; Replace (plus p (S n)) with (plus (S p) n); Auto with arith.
Apply plus_Snm_nSm.
Save.

(* For splitting *)
Lemma Write_MBlk_siz_Split : (H:heap; R:rfile; p,q:addr; siz,siz':nat)
 (MBlk (plus (2) p) q siz (H,R))
  -> (gt (minus siz siz') (2))
  -> (star (MBlk (plus (2) p) q (minus siz siz'))
   (ptoanylist (plus p (minus siz siz')) siz')
   ((hwrite H (plus (1) p) (minus siz siz')),R)).
Proof.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Cut (star (pto (minus (plus (2) p) (2)) q)
         (star (pto (minus (plus (2) p) (1)) siz)
           (ptoanylist (plus (2) p) (minus siz (2)))) (H,R)).
Clear H4; Intros.
Cut (star (ptoanylist (plus (2) p) (minus siz (2)))
	 (star (pto (minus (plus (2) p) (2)) q)
	   (pto (minus (plus (2) p) (1)) siz)) (H,R)).
Clear H3; Intros.
Cut (star
	 (star
	   (ptoanylist (plus (2) p) (minus (minus siz siz') (2)))
	   (ptoanylist (plus p (minus siz siz')) siz'))
         (star (pto (minus (plus (2) p) (2)) q)
           (pto (minus (plus (2) p) (1)) siz)) (H,R)).
Clear H3; Intros.
Apply star_subst with
      (star
	(star (pto (minus (plus (2) p) (2)) q)
          (pto (minus (plus (2) p) (1)) (minus siz siz')))
	(ptoanylist (plus (2) p) (minus (minus siz siz') (2))))
       (ptoanylist (plus p (minus siz siz')) siz'); Auto.
Apply star_assoc_R.
Apply star_subst with
(star (pto (minus (plus (2) p) (1)) (minus siz siz'))
  (pto (minus (plus (2) p) (2)) q))
(star (ptoanylist (plus (2) p) (minus (minus siz siz') (2)))
  (ptoanylist (plus p (minus siz siz')) siz')); Auto.
Apply star_assoc_R.
Apply star_frame with (pto (minus (plus (2) p) (1)) siz).
Apply star_assoc_L.
Apply star_sym.
Apply star_subst with
         (star (ptoanylist (plus (2) p) (minus (minus siz siz') (2)))
           (ptoanylist (plus p (minus siz siz')) siz'))
         (star (pto (minus (plus (2) p) (2)) q)
           (pto (minus (plus (2) p) (1)) siz)); Auto.
Intro; Apply star_sym.
Clear H3; Intros.
Inversion_clear H4.
Unfold getValinH in H5; Constructor.
Apply indom_some with siz.
Replace (plus (1) p) with (minus (plus (2) p) (1)); Auto with arith.
Unfold pto; Constructor.
Unfold getValinH; Apply write_read.
Intros; Unfold getValinH; Unfold getValinH in H6.
Apply trans_eq with (hread H3 j); Auto.
Apply local_write; Auto.
Intro; Apply star_sym.
Clear H3; Intros.
Unfold MBlk.
Constructor; Auto.
Constructor; Auto.
Apply star_star1.
Apply star_subst with (pto (minus (plus (2) p) (2)) q)
	 (star (pto (minus (plus (2) p) (1)) (minus siz siz'))
	   (ptoanylist (plus (2) p) (minus (minus siz siz') (2)))); Auto.
Apply star_assoc_L; Auto.
Intros; Apply star_star1; Auto.
Intros until R0; Apply h_subst_ptoanylist.
Clear H3; Intros.
Apply h_subst_star1 with H3; Auto.
Intros until R1; Apply h_subst_ptoanylist.
Apply star_subst with (ptoanylist (plus (2) p) (minus siz (2)))
         (star (pto (minus (plus (2) p) (2)) q)
           (pto (minus (plus (2) p) (1)) siz)); Auto.
Clear H3; Intros.
Replace (plus p (minus siz siz')) with
	 (plus (plus (2) p) (minus (minus siz siz') (2))).
Apply Unfold_ptoanylist.
Replace (plus (minus (minus siz siz') (2)) siz') with (minus siz (2)); Auto.
Cut (gt siz siz').
Intro; Apply trans_eq with (minus (plus (minus siz siz') siz') (2)).
Replace (plus (minus siz siz') siz') with siz; Auto.
Apply trans_eq with (minus (plus siz siz') siz').
Replace (plus siz siz') with (plus siz' siz); Auto with arith.
Apply sym_eq; Apply plus_minus_eq_minus_plus; Auto.
Apply sym_eq; Apply plus_minus_eq_minus_plus; Auto.
Apply lt_gt.
Apply lt_O_minus_lt.
Apply lt_trans with (2); Auto.
Replace (plus (2) p) with (S (S p)); Auto.
Apply trans_eq with (plus (S p) (S (minus (minus siz siz') (2)))).
Apply plus_Snm_nSm.
Apply trans_eq with (plus p (S (S (minus (minus siz siz') (2))))).
Apply plus_Snm_nSm.
Replace (S (S (minus (minus siz siz') (2)))) with (minus siz siz'); Auto.
Apply trans_eq with (plus (2) (minus (minus siz siz') (2))); Auto.
Apply sym_eq; Apply le_plus_minus_r.
Apply lt_le_weak; Auto.
Apply star_sym.
Apply star_assoc_R; Auto.
Apply star_subst with (pto (minus (plus (2) p) (2)) q)
(star1 (minus (plus (2) p) (1)) siz
           (ptoanylist (plus (2) p) (minus siz (2)))); Auto.
Apply star1_star; Auto.
Intro; Apply star1_star.
Save.

(* For splitting *)
Lemma Write_Star_MBlk_siz_Split :
 (H:heap; R:rfile; p,q:addr; siz,siz':nat; a:assert)
 (star (MBlk (plus (2) p) q siz) a (H,R))
  -> (gt (minus siz siz') (2))
  -> (star
   (star (MBlk (plus (2) p) q (minus siz siz'))
     (ptoanylist (plus p (minus siz siz')) siz'))
      a ((hwrite H (plus (1) p) (minus siz siz')),R)).
Proof.
Intros.
Apply star_frame with (MBlk (plus (2) p) q siz); Auto.
Clear H0; Intros.
Constructor.
Apply MBlk_indom with R0 q siz; Auto.
Apply lt_trans with (2); Auto.
Apply gt_lt.
Apply MBlk_siz_gt_2 with (H0,R0) (plus (2) p) q; Auto.
Apply Write_MBlk_siz_Split; Auto.
Save.

Lemma hwrite_dom : (H:heap; i,p:addr; n:hval)
 (hindom i (hwrite H p n)) -> (hindom i H)\/(p=i).
Proof.
Intros.
Inversion_clear H0.
Unfold hwrite in H1.
Cut (false=(beq_nat p i))\/(true=(beq_nat p i)).
Intro OR; Inversion_clear OR.
Rewrite <- H0 in H1.
Left; Apply indom_some with n0; Auto.
Right; Apply beq_nat_eq; Auto.
Generalize (beq_nat p i); Induction b; Auto.
Save.

Lemma hwriteN_dom : (n:nat; H:heap; i,p:addr; l:(nat->hval))
 (hindom i (hwriteN n H p l)) -> (hindom i H)\/((le p i)/\(lt i (plus p n))).
Proof.
Induction n; Intros.
Unfold hwriteN in H0.
Left; Assumption.
Unfold hwriteN in H1; Fold hwriteN in H1.
Cut (hindom i (hwrite H0 p (l (0))))
  \/((le (S p) i)/\(lt i (plus (S p) n0))).
Intro.
Inversion_clear H2.
Cut (hindom i H0) \/ (p=i).
Intro OR; Inversion_clear OR; Auto.
Right; Rewrite H2; Constructor; Auto.
Replace (plus i (S n0)) with (S (plus i n0)); Auto with arith.
Apply hwrite_dom with (l O); Auto.
Right.
Inversion_clear H3; Constructor; Auto with arith.
Replace (plus p (S n0)) with (plus (S p) n0); Auto.
Apply plus_Snm_nSm.
Apply H with ([x:nat](l (S x))); Assumption.
Save.

Lemma heap_eq_writeN_gen : (n:nat; H0,H,H1:heap; p:addr; vals:(nat->hval))
 (heap_eq (happend H0 H) H1)
  -> (heap_eq (happend (hwriteN n H0 p vals) H) (hwriteN n H1 p vals)).
Proof.
Induction n; Intros.
Unfold hwriteN; Auto.
Unfold hwriteN; Fold hwriteN.
Apply H.
Apply heap_eq_trans with (hwrite (happend H0 H1) p (vals (0))).
Apply append_write_sym_L.
Apply h_subst_write; Auto.
Save.

Lemma heap_eq_writeN : (n:nat; H:heap; p:addr; vals:(nat->hval))
(heap_eq (happend (hwriteN n hempty p vals) H) (hwriteN n H p vals)).
Proof.
Intros.
Apply heap_eq_writeN_gen.
Apply append_emp.
Save.

Lemma hwriteN_dom_rev : (n:nat; H:heap; p,q:addr; vals:(nat->hval))
 (hnotindom p H) -> (lt p q) ->
 (hnotindom p (hwriteN n H q vals)).
Proof.
Induction n; Intros.
Unfold hwriteN; Auto.
Unfold hwriteN; Fold hwriteN.
Apply h_subst_notin with
 (happend (hwriteN n0 hempty (S q) [x:nat](vals (S x)))
   (hwrite H0 q (vals (0)))).
Apply heap_eq_writeN.
Apply append_preserve_notin.
Apply H.
Unfold hempty; Apply notindom_none; Auto.
Apply lt_trans with q; Auto.
Apply notindom_none; Inversion_clear H1.
Unfold hwrite.
Replace (beq_nat q p) with false; Auto.
Apply beq_nat_neq.
Apply lt_neq_rev; Auto.
Save.

Lemma ptoanylist_heap : (n:nat; R:rfile; p:addr; vals:(nat->hval))
 (ptoanylist p n ((hwriteN n hempty p vals),R)).
Proof.
Induction n; Intros.
Unfold hwriteN; Unfold ptoanylist.
Unfold hempty; Unfold emp; Auto.
Unfold hwriteN; Fold hwriteN; Unfold ptoanylist; Fold ptoanylist.
Exists (hwrite hempty p (vals (0))).
Exists (hwriteN n0 hempty (S p) [x:nat](vals (S x))).
Cut (hdisj (hwrite hempty p (vals (0)))
     (hwriteN n0 hempty (S p) [x:nat](vals (S x)))).
Intro; Constructor; Auto.
Constructor.
Apply heap_eq_trans with
(happend (hwriteN n0 hempty (S p) [x:nat](vals (S x)))
  (hwrite hempty p (vals (0)))).
Apply append_sym; Auto.
Apply heap_eq_writeN.
Constructor.
Exists (vals O).
Unfold hempty; Unfold hwrite; Unfold pto.
Unfold getValinH; Unfold hread; Constructor.
Replace (beq_nat p p) with true; Trivial.
Apply beq_nat_refl.
Intros.
Replace (beq_nat p j) with false; Trivial.
Apply beq_nat_neq; Auto.
Apply H.
Apply disj_notboth.
Intro.
Cut (false=(beq_nat p i)\/true=(beq_nat p i)).
Intro OR; Inversion_clear OR.
Left.
Apply notindom_none.
Unfold hempty; Unfold hwrite.
Rewrite <- H0; Trivial.
Right.
Apply hwriteN_dom_rev.
Unfold hempty; Apply notindom_none; Auto.
Replace i with p; Auto.
Apply beq_nat_eq; Auto.
Generalize (beq_nat p i); Induction b; Auto.
Save.

(* For alloc *)
Lemma star_alloc : (n:nat; H:heap; R:rfile; a:assert; p:addr; vals:(nat->int))
 (a (H,R))
  -> (gt p O)
  -> ((i:nat) (gt n i) -> (hnotindom (plus i p) H))
  -> (star (ptoanylist p n) a ((hwriteN n H p vals),R)).
Proof.
Intros.
Exists (hwriteN n hempty p vals).
Exists H.
Constructor.
Apply disj_notboth.
Intro.
Cut (hindom i (hwriteN n hempty p vals))
  \/(hnotindom i (hwriteN n hempty p vals)).
Intro OR; Inversion_clear OR; Auto.
Right.
Cut (hindom i hempty)\/((le p i)/\(lt i (plus p n))).
Intro OR; Inversion_clear OR.
Inversion_clear H4.
Unfold hempty in H5; Inversion H5.
Inversion_clear H4.
Replace i with (plus (minus i p) p).
Apply H2.
Apply lt_gt.
Replace n with (minus (plus n p) p).
Apply lt_minus_bothside; Auto.
Replace (plus n p) with (plus p n); Auto with arith.
Replace (plus n p) with (plus p n); Auto with arith.
Apply sym_eq; Replace (plus (minus i p) p) with (plus p (minus i p));
Auto with arith.
Apply hwriteN_dom with vals; Auto.
Apply indom_or_notin.
Constructor.
Apply heap_eq_writeN.
Constructor; Auto.
Apply ptoanylist_heap.
Save.
